// RMessageServerDefs.h
// 
// Copyright (c) 2004, Ingo Weinhold (bonefish@cs.tu-berlin.de)
// 
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation 
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
// THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
// DEALINGS IN THE SOFTWARE.
// 
// Except as contained in this notice, the name of a copyright holder shall
// not be used in advertising or otherwise to promote the sale, use or other
// dealings in this Software without prior written authorization of the
// copyright holder.

#ifndef RMESSAGE_SERVER_DEFS_H
#define RMESSAGE_SERVER_DEFS_H

#include <RMessageBuild.h>

// message what types
enum {
	MSG_REPLY							= 'RMre',
	MSG_CONNECT							= 'RMco',
	MSG_GET_APP_LIST					= 'RMal',
	MSG_GET_APP_INFO					= 'RMai',
	MSG_GET_MESSENGER					= 'RMgm',
	MSG_DUMP							= 'RMdu',	// debugging only

	// server to server communication
	MSG_REQUEST_REPLY					= 'RMrp',
	MSG_REQUEST_MESSAGE					= 'RMrm',

	// intra-server communication
	MSG_LOCAL_REQUEST_MESSAGE			= 'RMlm',
	MSG_LOCAL_REQUEST_CONNECTION_BROKEN	= 'RMcb',
	MSG_LOCAL_REQUEST_SCRIPTING_MESSAGE	= 'RMsm',

	// defined in <RemoteMessage.h>
	// RMESSAGE_DATA					= 'RMda'
	// RMESSAGE_DELIVERY_FAILED			= 'RMdf'
	// RMESSAGE_DELIVERY_TIMED_OUT		= 'RMto',
};

// the server signature
extern _IMPEXP_RMESSAGE const char* kRMessageServerSignature;

// the server port
enum {
	RMESSAGE_DEFAULT_SERVER_PORT	= 12345,
};

// request parameter specification (server to server communication)
enum {
	REQUEST_NO_PARAMETER		= 0,
	REQUEST_MESSAGE_PARAMETER	= 1,
	REQUEST_DATA_PARAMETER		= 2,
};


/*
Library - Server Protocols
--------------------------

standard reply message

reply:		MSG_REPLY
on error:	"error": B_INT32_TYPE
			[ "error_description": B_STRING_TYPE ]
			[ <additional fields> ]

fields:
- "error": The error code (a status_t).
- "error_description": Optional human readable description.
- <additional fields>: Request-specific fields.

-----------------------------------------------------------------------

connection request

Connects with a remote peer and returns a messenger to which to send
subsequent requests.

target:		server looper
message:	MSG_CONNECT
			"address":				B_MESSAGE_TYPE (archived BNetAddress)
reply:		standard success
			"request messenger":	B_MESSENGER_TYPE
on error:	- B_NO_REPLY (fatal)
			- standard error

message fields:
- "address": The address of the host to connect with.

reply fields:
- "request messenger": The messenger to send requests to.

error reply fields:
- "error":
  - B_BAD_VALUE: A request message field is missing or contains an
    invalid value.
  - ...

-----------------------------------------------------------------------

get app list

Gets a list of team IDs of all running apps, respectively, if a signature is
given, only the team IDs of the running apps with that signature.

target:		connection request messenger
message:	MSG_GET_APP_LIST
			[ "signature": 			B_STRING_TYPE ]
reply:		standard success
			"teams":				B_INT32_TYPE[]
on error:	- B_NO_REPLY (fatal)
			- standard error

message fields:
- "signature": The signature of the apps to return. If not specified all
  apps are returned.

reply fields:
- "teams": The list of all teams matching the request (team_id[]).

error reply fields:
- "error":
  - B_BAD_VALUE: A request message field is missing or contains an
    invalid value.
  - ...

-----------------------------------------------------------------------

get app info

Returns a app_info for a running application. The application is identified
by signature or team ID. If neither is given the app_info of the active
app is returned.

target:		connection request messenger
message:	MSG_GET_APP_INFO
			[ "signature":			B_STRING_TYPE
			  | "team": 			B_INT32_TYPE ]
reply:		standard success
			"thread":				B_INT32_TYPE
			"team":					B_INT32_TYPE
			"port":					B_INT32_TYPE
			"flags":				B_UINT32_TYPE
			"ref":					B_REF_TYPE
			"signature":			B_STRING_TYPE
on error:	- B_NO_REPLY (fatal)
			- standard error

message fields:
- "signature": The signature of the app whose info shall be returned.
- "team": The team ID of the app whose info shall be returned (team_id).

reply fields:
- "thread": The ID of the app's main thread (thread_id).
- "team": The team ID of the app (team_id).
- "port": The app's main looper message port (port_id).
- "flags": The app flags.
- "ref": A entry ref referring to the app's executable.
- "signature": The app's signature.

error reply fields:
- "error":
  - B_BAD_VALUE: A request message field is missing or contains an
    invalid value.
  - ...

-----------------------------------------------------------------------

get messenger

Gets a messenger referring to a remote application. The application is
identified by either team ID or signature.

target:		connection request messenger
message:	MSG_GET_MESSENGER
			"signature":			B_STRING_TYPE
			| "team":				B_INT32_TYPE
			[ "or name" 			B_BOOL_TYPE ]
reply:		standard success
			"messenger":			B_MESSENGER_TYPE
on error:	- B_NO_REPLY (fatal)
			- standard error

message fields:
- "signature": The signature of the app for which to return a messenger.
- "team": The team ID of the app for which to return a messenger (team_id).
- "or name": Only in conjunction with "signature". If the supplied
  signature string does not identify a running application, it is looked
  for one whose file entry name matches the string.

reply fields:
- "messenger": The messenger referring to the app.

error reply fields:
- "error":
  - B_BAD_VALUE: A request message field is missing or contains an
    invalid value.
  - ...

-----------------------------------------------------------------------

Library - Server Protocols
--------------------------

send message

Sends a message to a remote target.

target:		remote server
message:	MSG_REQUEST_MESSAGE
			"target":				B_MESSENGER_TYPE
			"reply target":			B_MESSENGER_TYPE
			| "team":				B_INT32_TYPE
reply:		none
on error:	none

message fields:
- "target": The messenger referring to the target the message shall be sent
  to. Valid on the target system.
- "reply target": Messenger referring to the target replies to the message
  shall be sent to. Valid on the source system.

parameter:
Message/data to be sent to the target.

*/

#endif	// RMESSAGE_SERVER_DEFS_H
