// RemoteMessage.cpp
// 
// Copyright (c) 2004, Ingo Weinhold (bonefish@cs.tu-berlin.de)
// 
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation 
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
// THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
// DEALINGS IN THE SOFTWARE.
// 
// Except as contained in this notice, the name of a copyright holder shall
// not be used in advertising or otherwise to promote the sale, use or other
// dealings in this Software without prior written authorization of the
// copyright holder.

#include <Message.h>

#include <RemoteMessage.h>

#include "Compatibility.h"

/*!	\namespace RemoteMessage

	\brief Features functions that simplify sending data via messages.

	Although the namespace is called RemoteMessage these functions have
	nothing to do with remote messaging directly. They just implement a
	protocol how data shall be stuffed into and extracted from an envelope
	message.
	The link to the remote messaging is that this protocol is explicitly
	supported by the RMessageServer, so that it 1) works at all, and 2)
	should be relatively efficient (the data is copied as few times as
	possible).

	SetData() finds out, which area the data lives in and at which offset and
	put those information into the envelope message.

	GetData() extracts these information and clones the original area
	read-only.

	ReleaseData() deletes the cloned area when the receiver is done with the
	data.

	There's apparently no benefit, when the amount of data is small. In fact
	in those cases the additional overhead should make it even slower than
	just adding the data to the message. The protocol may be improved later to
	do just that.
*/

// SetData
/*!	\brief Prepares an envelope for sending data via a BMessage.

	\a data and \a size specify the data to be sent, and \a parameters
	is an optional message to be sent together with the data -- usally
	providing the receiver with additional information concerning the data.

	While the contents of \a parameters can be arbitrary, \a envelope has
	a strict format and must not be manipulated by the caller afterwards.
	The method will set its \c what field to \c RMESSAGE_DATA and empty it
	of data first, and then add fields according to the given parameters.

	After successful completion of the method, \a envelope can be sent to
	the receiver via the normal messaging mechanism (\c BMessenger).

	The receiver will call GetData() on the envelope, which will clone the
	area the data live in. By convention the receiver shall not reply to the
	message until they're done with the data, and the sender shall not modify
	the data until the receiver's reply arrived.

	\param envelope the envelope message to contain the data.
	\param data the data to be sent. May be \c NULL in which case \a size must
		   be \c 0.
	\param size the number of bytes to be sent.
	\param parameters an additional message to be sent together with the data.
		   May be \c NULL.
	\return
	- \c B_OK: Everything went fine.
	- another error code otherwise.
*/
status_t
RemoteMessage::SetData(BMessage* envelope, const void* data, size_t size,
	BMessage* parameters)
{
	// check parameters
	if (!envelope || !data && size > 0)
		return B_BAD_VALUE;
	status_t error = B_OK;
	envelope->what = RMESSAGE_DATA;
	envelope->MakeEmpty();
	// add data
	if (size > 0) {
		// get area
		area_id area = area_for(const_cast<void*>(data));
		if (area < 0)
			return area;
		// get area info
		area_info info;
		error = get_area_info(area, &info);
		if (error != B_OK)
			return error;
		// add area ID and in-area offset
		error = envelope->AddInt32("area", area);
		if (error != B_OK)
			return error;
		error = envelope->AddInt32("offset",
			(const uint8*)data - (const uint8*)info.address);
		if (error != B_OK)
			return error;
	}
	// add data size
	error = envelope->AddInt32("size", (int32)size);
	if (error != B_OK)
		return error;
	// add parameter message
	if (parameters)
		error = envelope->AddMessage("parameters", parameters);
	return error;
}

// GetData
/*!	\brief Retrieves the data from a data message envelope.

	Note that the caller must invoke ReleaseData() when done with the data,
	before replying to or deleting the envelope.

	The sent data may be empty, in which case \c NULL will be returned in
	\a data and \c 0 in \a size.

	\param envelope the envelope message from which to get the data.
	\param data pointer to a variable that shall be set to point to the data.
		   Must not be \c NULL.
	\param size pointer to a variable that shall be set to the size of the
		   data.
	\param parameters pointer to a BMessage that shall be initialized to the
		   parameters contained in the envelope. May be \c NULL.
	\param hasParameters pointer to a boolean variable that shall be set to
		   \c true, if the envelope contains a parameter message, to \c false
		   otherwise. May be \c NULL. If \a parameters is \c NULL,
		   \a hasParameters will be ignored, too.
	\return
	- \c B_OK: Everything went fine.
	- another error code otherwise.
*/
status_t
RemoteMessage::GetData(const BMessage* envelope, const void** _data,
	size_t* _size, BMessage* parameters, bool* _hasParameters)
{
	// check parameters
	if (!envelope || !_data || !_size)
		return B_BAD_VALUE;
	// get parameter message
	if (parameters) {
		bool hasParameters
			= (envelope->FindMessage("parameters", parameters) == B_OK);
		if (_hasParameters)
			*_hasParameters = hasParameters;
	}
	// get data size
	int32 size;
	status_t error = envelope->FindInt32("size", &size);
	if (error != B_OK)
		return error;
	if (size < 0)
		return B_BAD_DATA;
	// get area and in-area offset
	uint8* data = NULL;
	if (size > 0) {
		area_id area;
		error = envelope->FindInt32("area", &area);
		if (error != B_OK)
			return error;
		int32 offset;
		error = envelope->FindInt32("offset", &offset);
		if (error != B_OK)
			return error;
		// get area info
		area_info info;
		error = get_area_info(area, &info);
		if (error != B_OK)
			return error;
		// check the offset
		if (offset < 0 || (uint32)offset + size > info.size)
			return B_BAD_DATA;
		// clone the area and set the result
		void* address;
		area_id clonedArea = clone_area("cloned rmessage data", &address,
			B_ANY_ADDRESS, B_READ_AREA, area);
		if (clonedArea < 0)
			return clonedArea;
		data = (uint8*)address + offset;
	}
	*_data = data;
	*_size = size;
	return error;
}

// ReleaseData
/*!	\brief Releases the data retrieved from a data message envelope.

	The area that contains the data will be released. After the function
	returns the data must not be accessed.

	There is no need to call this function for zero-sized data, though it
	does no harm.

	\param data the data.
	\param size the size of the data.
	\return
	- \c B_OK: Everything went fine.
	- another error code otherwise.
*/
status_t
RemoteMessage::ReleaseData(const void* data, size_t size)
{
	if (data) {
		area_id area = area_for(const_cast<void*>(data));
		if (area >= 0)
			delete_area(area);
	} else if (size > 0)
		return B_BAD_VALUE;
	return B_OK;
}


/*!	\var RemoteMessage::RMESSAGE_DATA
	\brief \c BMessage::what value identifying a data message.
*/

/*!	\var RemoteMessage::RMESSAGE_DELIVERY_FAILED
	\brief \c BMessage::what value identifying a message that is sent, when
		   the delivery of a remote message failed for any reason.

	The message will contain a \c "error" field of type \c B_INT32_TYPE
	with the exact error that occurred.
*/

/*!	\var RemoteMessage::RMESSAGE_DELIVERY_TIMED_OUT
	\brief \c BMessage::what value identifying a message that is sent, when
		   the delivery of a remote message timed out.
*/

