// RemoteRoster.cpp
// 
// Copyright (c) 2004, Ingo Weinhold (bonefish@cs.tu-berlin.de)
// 
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation 
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
// THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
// DEALINGS IN THE SOFTWARE.
// 
// Except as contained in this notice, the name of a copyright holder shall
// not be used in advertising or otherwise to promote the sale, use or other
// dealings in this Software without prior written authorization of the
// copyright holder.

#include <new>
#include <string.h>

#include <List.h>
#include <Roster.h>

#include <RemoteRoster.h>

#include "Debug.h"
#include "RemoteRosterImpl.h"
#include "RMessageServerDefs.h"

/*!	\class RemoteRoster

	\brief Roughly corresponds to a BRoster of a remote host.

	Only a subset of the BRoster functionality is implemented, though.
	Namely, most of the methods for querying running application.
	Additionally the GetMessenger() methods retrieve BMessengers referring
	to remote applications.
*/

// constructor
/*!	\brief Constructs a RemoteRoster referring to a host associated with the
		   supplied address.

	After construction InitCheck() shall be invoked to check whether the
	roster could be initialized properly.

	\param address the address of the host. May be \c NULL in which case the
		   roster will be uninitialized.
*/
_EXPORT
RemoteRoster::RemoteRoster(const BNetAddress* address)
	: fInitStatus(B_NO_INIT),
	  fAddress(NULL),
	  fImpl(NULL)
{
	if (address)
		SetTo(address);
}

// constructor
/*!	\brief Constructs a RemoteRoster referring to a host associated with the
		   supplied host name.

	After construction InitCheck() shall be invoked to check whether the
	roster could be initialized properly.

	\param host a string identifying the host (name or IP string). May be
		   \c NULL in which case the roster will be uninitialized.
*/
_EXPORT
RemoteRoster::RemoteRoster(const char* host)
	: fInitStatus(B_NO_INIT),
	  fAddress(NULL),
	  fImpl(NULL)
{
	if (host)
		SetTo(host);
}

// destructor
/*!	\brief Frees all resources associate with the object.
*/
_EXPORT
RemoteRoster::~RemoteRoster()
{
	Unset();
}

// SetTo
/*!	\brief Initializes a RemoteRoster to referring to a host associated with
		   the supplied address.

	\param address the address of the host.
	\return
	- \c B_OK: Everything went fine.
	- another error code otherwise.
*/
_EXPORT
status_t
RemoteRoster::SetTo(const BNetAddress* address)
{
FUNCTION_START();
	Unset();
	if (!address)
		return (fInitStatus = B_BAD_VALUE);
	// clone address
	status_t error = B_OK;
	fAddress = new(nothrow) BNetAddress(*address);
	if (!fAddress)
		RETURN_ERROR(B_NO_MEMORY);
	// get server messenger
	BMessenger serverMessenger(kRMessageServerSignature);
	if (error == B_OK && !serverMessenger.IsValid())
		SET_ERROR(error, B_ERROR);
	// ask server to connect
	if (error == B_OK) {
		// prepare request message
		BMessage request(MSG_CONNECT);
		BMessage addressArchive;
		error = fAddress->Archive(&addressArchive);
		if (error == B_OK)
			SET_ERROR(error, request.AddMessage("address", &addressArchive));
		// send request
		BMessage reply;
		if (error == B_OK)
			SET_ERROR(error, serverMessenger.SendMessage(&request, &reply));
		// check request result
		if (error == B_OK && reply.what != MSG_REPLY)
			SET_ERROR(error, B_ERROR);
		status_t requestError;
		if (error == B_OK && reply.FindInt32("error", &requestError) == B_OK)
			SET_ERROR(error, requestError);
		// get request messenger
		BMessenger requestMessenger;
		if (error == B_OK) {
			SET_ERROR(error, reply.FindMessenger("request messenger",
				&requestMessenger));
		}
		if (error == B_OK && !requestMessenger.IsValid())
			SET_ERROR(error, B_ERROR);
		// create the impl object
		if (error == B_OK) {
			fImpl = new(nothrow) RemoteRosterImpl(requestMessenger);
			if (fImpl) {
				SET_ERROR(error, fImpl->InitCheck());
			} else {
				SET_ERROR(error, B_NO_MEMORY);
			}
		}
	}
	// cleanup on error
	if (error != B_OK)
		Unset();
	RETURN_ERROR(fInitStatus = error);
}

// SetTo
/*!	\brief Initializes a RemoteRoster to referring to a host associated with
		   the supplied host name.

	\param host a string identifying the host (name or IP string).
	\return
	- \c B_OK: Everything went fine.
	- another error code otherwise.
*/
_EXPORT
status_t
RemoteRoster::SetTo(const char* host)
{
	Unset();
	if (!host)
		return (fInitStatus = B_BAD_VALUE);
	BNetAddress address(host);
	if (address.InitCheck() != B_OK)
		RETURN_ERROR(fInitStatus = address.InitCheck());
	return SetTo(&address);
}

// Unset
/*!	\brief Uninitializes the object.
*/
_EXPORT
void
RemoteRoster::Unset()
{
	delete fAddress;
	fAddress = NULL;
	delete fImpl;
	fImpl = NULL;
	fInitStatus = B_NO_INIT;
}

// InitCheck
/*!	\brief Returns the status of the last initialization.

	\return
	- \c B_OK: The object is properly initialized.
	- \c B_NO_INIT: The object is not initialized.
	- another error code.
*/
_EXPORT
status_t
RemoteRoster::InitCheck() const
{
	return fInitStatus;
}

// GetAppList
/*!	\brief Returns a list of the applications running on the remote host.

	Corresponds to BRoster::GetAppList(BList*). The returned team_ids are, of
	course, not valid on this machine. They can be used as parameter for
	GetRunningAppInfo() or GetMessenger() though.

	\param teams BList to which the team_ids of the applications running on
		   the remote machine shall be added.
	\return
	- \c B_OK: Everything went fine.
	- another error code otherwise.
*/
_EXPORT
status_t
RemoteRoster::GetAppList(BList* teams) const
{
	return (InitCheck() == B_OK ? fImpl->GetAppList(teams) : InitCheck());
}

// GetAppList
/*!	\brief Returns a list of the applications running on the remote host that
		   have the given signature.

	Corresponds to BRoster::GetAppList(const char*, BList*). The returned
	team_ids are, of course, not valid on this machine. They can be used as
	parameter for GetRunningAppInfo() or GetMessenger() though.

	\param teams BList to which the team_ids of the applications running on
		   the remote machine shall be added.
	\param signature signature of the apps whose team_ids shall be returned.
		   May be \c NULL, in which case the team_ids of all running
		   applications are returned.
	\return
	- \c B_OK: Everything went fine.
	- another error code otherwise.
*/
_EXPORT
status_t
RemoteRoster::GetAppList(const char* signature, BList* teams) const
{
	return (InitCheck() == B_OK ? fImpl->GetAppList(signature, teams)
								: InitCheck());
}

// TeamFor
/*!	\brief Returns the team_id of the application running on the remote host
		   that has the given signature.

	Corresponds to BRoster::TeamFor(const char*). The returned
	team_id is, of course, not valid on this machine. It can be used as
	parameter for GetRunningAppInfo() or GetMessenger() though.

	If there are more than one running applications with the given signature,
	it is undefined whose team_id will be returned.

	\param signature signature of the app whose team_id shall be returned.
	\return
	- \c B_OK: Everything went fine.
	- another error code otherwise.

*/
_EXPORT
team_id
RemoteRoster::TeamFor(const char* signature) const
{
	return (InitCheck() == B_OK ? fImpl->TeamFor(signature) : InitCheck());
}

// GetRunningAppInfo
/*!	\brief Returns an info of the application running on the remote host
		   that has the given signature.

	If there are more than one running applications with the given signature,
	it is undefined whose info will be returned.

	The method is equivalent to
	\code GetRunningAppInfo(TeamFor(signature), info)
	\endcode.

	\param signature signature of the app whose info shall be returned.
	\param info pointer to an app_info which shall be filled with the info
		   for the specified application.
	\return
	- \c B_OK: Everything went fine.
	- another error code otherwise.

*/
_EXPORT
status_t
RemoteRoster::GetRunningAppInfo(const char* signature, app_info* info) const
{
	return (InitCheck() == B_OK ? fImpl->GetRunningAppInfo(signature, info)
								: InitCheck());
}

// GetRunningAppInfo
/*!	\brief Returns an info of the application running on the remote host
		   that has the given team_id.

	Corresponds to BRoster::GetRunningAppInfo(team_id, app_info*).

	\param signature signature of the app whose info shall be returned.
	\param info pointer to an app_info which shall be filled with the info
		   for the specified application.
	\return
	- \c B_OK: Everything went fine.
	- \c B_BAD_TEAM_ID: The given team_id does not refer to an existing team.
	- another error code otherwise.

*/
_EXPORT
status_t
RemoteRoster::GetRunningAppInfo(team_id team, app_info* info) const
{
	return (InitCheck() == B_OK ? fImpl->GetRunningAppInfo(team, info)
								: InitCheck());
}

// GetActiveAppInfo
/*!	\brief Returns an info of the application active on the remote host.

	Corresponds to BRoster::GetActiveAppInfo(app_info*).

	\param info pointer to an app_info which shall be filled with the info
		   for the active application.
	\return
	- \c B_OK: Everything went fine.
	- another error code otherwise.

*/
_EXPORT
status_t
RemoteRoster::GetActiveAppInfo(app_info* info) const
{
	return (InitCheck() == B_OK ? fImpl->GetActiveAppInfo(info) : InitCheck());
}

// GetMessenger
/*!	\brief Returns a messenger for the application running on the remote host
		   that has the given signature.

	If no application with the given signature could be found and \a byName
	is \c true, \a signature will be interpreted as a file name and a messenger
	for the first running application with that file name will be returned.

	\param signature signature of the app for which to return a messenger.
	\param messenger pointer to the BMessenger to be initialized to refer to
		   the remote application.
	\param byName if \c true, and \a signature is not a signature of a running
		   application \a signature will be interpreted as a file name.
	\return
	- \c B_OK: Everything went fine.
	- anther error code otherwise.
*/
_EXPORT
status_t
RemoteRoster::GetMessenger(const char* signature, BMessenger* messenger,
	bool byName) const
{
	return (InitCheck() == B_OK
		? fImpl->GetMessenger(signature, messenger, byName) : InitCheck());
}

// GetMessenger
/*!	\brief Returns a messenger for the application running on the remote host
		   that has the given team_id.

	\param team the team_id identifying the application for which to return
		   a BMessenger.
	\param messenger pointer to the BMessenger to be initialized to refer to
		   the remote application.
	\return
	- \c B_OK: Everything went fine.
	- \c B_BAD_TEAM_ID: The given team_id does not refer to an existing team.
	- anther error code otherwise.
*/
_EXPORT
status_t
RemoteRoster::GetMessenger(team_id team, BMessenger* messenger) const
{
	return (InitCheck() == B_OK ? fImpl->GetMessenger(team, messenger)
								: InitCheck());
}
