// RemoteRosterImpl.cpp
// 
// Copyright (c) 2004, Ingo Weinhold (bonefish@cs.tu-berlin.de)
// 
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation 
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
// THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
// DEALINGS IN THE SOFTWARE.
// 
// Except as contained in this notice, the name of a copyright holder shall
// not be used in advertising or otherwise to promote the sale, use or other
// dealings in this Software without prior written authorization of the
// copyright holder.

#include <new>
#include <string.h>

#include <List.h>
#include <Roster.h>

#include "Debug.h"
#include "RemoteRosterImpl.h"
#include "RMessageServerDefs.h"

// constructor
_EXPORT
RemoteRosterImpl::RemoteRosterImpl()
	: fInitStatus(B_NO_INIT),
	  fRequestMessenger()
{
}

// constructor
_EXPORT
RemoteRosterImpl::RemoteRosterImpl(BMessenger requestMessenger)
	: fInitStatus(B_NO_INIT),
	  fRequestMessenger()
{
	SetTo(requestMessenger);
}

// destructor
_EXPORT
RemoteRosterImpl::~RemoteRosterImpl()
{
	Unset();
}

// SetTo
_EXPORT
status_t
RemoteRosterImpl::SetTo(BMessenger requestMessenger)
{
	fRequestMessenger = requestMessenger;
	if (fRequestMessenger.IsValid())
		fInitStatus = B_OK;
	else
		fInitStatus = B_NO_INIT;
	return fInitStatus;
}

// Unset
_EXPORT
void
RemoteRosterImpl::Unset()
{
	fRequestMessenger = BMessenger();
	fInitStatus = B_NO_INIT;
}

// InitCheck
_EXPORT
status_t
RemoteRosterImpl::InitCheck() const
{
	return fInitStatus;
}

// GetAppList
_EXPORT
status_t
RemoteRosterImpl::GetAppList(BList* teams) const
{
	return GetAppList(NULL, teams);
}

// GetAppList
_EXPORT
status_t
RemoteRosterImpl::GetAppList(const char* signature, BList* teams) const
{
	// check initialization and parameters
	if (InitCheck() != B_OK)
		return InitCheck();
	if (!teams)
		return B_BAD_VALUE;
	// prepare request message
	status_t error = B_OK;
	BMessage request(MSG_GET_APP_LIST);
	if (error == B_OK && signature)
		error = request.AddString("signature", signature);
	// send request
	BMessage reply;
	if (error == B_OK)
		error = _SendRequest(&request, &reply);
	// get the teams
	if (error == B_OK) {
		team_id team;
		for (int32 i = 0;
			 error == B_OK && reply.FindInt32("teams", i, &team) == B_OK;
			 i++) {
			if (!teams->AddItem((void*)team))
				error = B_NO_MEMORY;
		}
	}
	return error;
}

// TeamFor
_EXPORT
team_id
RemoteRosterImpl::TeamFor(const char* signature) const
{
	app_info info;
	status_t error = GetRunningAppInfo(signature, &info);
	return (error == B_OK ? info.team : error);
}

// GetRunningAppInfo
_EXPORT
status_t
RemoteRosterImpl::GetRunningAppInfo(const char* signature, app_info* info) const
{
	if (!signature)
		return B_BAD_VALUE;
	// prepare the request and delegate the real work to the worker method
	BMessage request(MSG_GET_APP_INFO);
	status_t error = request.AddString("signature", signature);
	if (error  == B_OK)
		error = _GetAppInfo(&request, info);
	return error;
}

// GetRunningAppInfo
_EXPORT
status_t
RemoteRosterImpl::GetRunningAppInfo(team_id team, app_info* info) const
{
	// prepare the request and delegate the real work to the worker method
	BMessage request(MSG_GET_APP_INFO);
	status_t error = request.AddInt32("team", team);
	if (error  == B_OK)
		error = _GetAppInfo(&request, info);
	return error;
}

// GetActiveAppInfo
_EXPORT
status_t
RemoteRosterImpl::GetActiveAppInfo(app_info* info) const
{
	// prepare the request and delegate the real work to the worker method
	BMessage request(MSG_GET_APP_INFO);
	return _GetAppInfo(&request, info);
}

// GetMessenger
_EXPORT
status_t
RemoteRosterImpl::GetMessenger(const char* signature, BMessenger* messenger,
	bool byName) const
{
	if (!signature)
		return B_BAD_VALUE;
	// prepare the request and delegate the real work to the worker method
	BMessage request(MSG_GET_MESSENGER);
	status_t error = request.AddString("signature", signature);
	if (error == B_OK && byName)
		error = request.AddBool("or name", byName);
	if (error  == B_OK)
		error = _GetMessenger(&request, messenger);
	return error;
}

// GetMessenger
_EXPORT
status_t
RemoteRosterImpl::GetMessenger(team_id team, BMessenger* messenger) const
{
	// prepare the request and delegate the real work to the worker method
	BMessage request(MSG_GET_MESSENGER);
	status_t error = request.AddInt32("team", team);
	if (error  == B_OK)
		error = _GetMessenger(&request, messenger);
	return error;
}

// _SendRequest
status_t
RemoteRosterImpl::_SendRequest(BMessage* request, BMessage* reply) const
{
	status_t error = fRequestMessenger.SendMessage(request, reply);
	// check request result
	if (error == B_OK && reply->what != MSG_REPLY)
		error = B_ERROR;
	status_t requestError;
	if (error == B_OK && reply->FindInt32("error", &requestError) == B_OK)
		error = requestError;
	return error;
}

// _GetAppInfo
status_t
RemoteRosterImpl::_GetAppInfo(BMessage* request, app_info* info) const
{
	// check initialization and parameters
	if (InitCheck() != B_OK)
		return InitCheck();
	if (!info)
		return B_BAD_VALUE;
	// send request
	BMessage reply;
	status_t error = _SendRequest(request, &reply);
	// get the info
	if (error == B_OK)
		 error = reply.FindInt32("thread", &info->thread);
	if (error == B_OK)
		 error = reply.FindInt32("team", &info->team);
	if (error == B_OK)
		 error = reply.FindInt32("port", &info->port);
	if (error == B_OK)
		 error = reply.FindInt32("flags", (int32*)&info->flags);
	if (error == B_OK)
		 error = reply.FindRef("ref", &info->ref);
	const char* signature;
	if (error == B_OK)
		 error = reply.FindString("signature", &signature);
	if (error == B_OK) {
		strncpy(info->signature, signature, B_MIME_TYPE_LENGTH);
		info->signature[B_MIME_TYPE_LENGTH - 1] = '\0';
	}
	return error;
}

// _GetMessenger
status_t
RemoteRosterImpl::_GetMessenger(BMessage* request, BMessenger* messenger) const
{
	// check initialization and parameters
	if (InitCheck() != B_OK)
		return InitCheck();
	if (!messenger)
		return B_BAD_VALUE;
	// send request
	BMessage reply;
	status_t error = _SendRequest(request, &reply);
	// get the messenger
	if (error == B_OK)
		 error = reply.FindMessenger("messenger", messenger);
	return error;
}

