// Connection.cpp
// 
// Copyright (c) 2004, Ingo Weinhold (bonefish@cs.tu-berlin.de)
// 
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation 
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
// THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
// DEALINGS IN THE SOFTWARE.
// 
// Except as contained in this notice, the name of a copyright holder shall
// not be used in advertising or otherwise to promote the sale, use or other
// dealings in this Software without prior written authorization of the
// copyright holder.

#include <new>

#include <Autolock.h>
#include <Message.h>
#include <String.h>

#include "AutoDeleter.h"
#include "Compatibility.h"
#include "Connection.h"
#include "Debug.h"
#include "HostAddress.h"
#include "Request.h"
#include "RequestHandler.h"
#include "RMessageServerDefs.h"

static const int32 kMaxInterruptedTxCount = 20;

// hexdump
/*static
void
hexdump(const void* _buffer, int32 size)
{
	uint8* buffer = (uint8*)_buffer;
	for (int32 i = 0; i < size; i++) {
		printf("%02x", buffer[i]);
	}
	printf("\n");
}
*/


// constructor
Connection::Connection(const BNetAddress& address)
	: fRemoteAddress(address),
	  fNetEndPoint(NULL),
	  fRequestHandler(NULL),
	  fReceiver(-1),
	  fTerminating(false),
	  fValid(false),
	  fReceivingConnection(false)
{
}

// destructor
Connection::~Connection()
{
PRINT(("Connection::~Connection()\n"));
	fTerminating = true;
	fSendLock.Lock();
	if (fNetEndPoint)
		fNetEndPoint->Close();
	if (fReceiver >= 0) {
		int32 result;
		wait_for_thread(fReceiver, &result);
	}
	delete fNetEndPoint;
PRINT(("Connection::~Connection() done\n"));
}

// Init
status_t
Connection::Init(BNetEndpoint* endPoint)
{
	status_t error = B_OK;
	fReceivingConnection = endPoint;
	if (fReceivingConnection) {
		// check end point
		error = endPoint->InitCheck();
		if (error != B_OK)
			return error;
		// spawn receiver thread
		if (fReceivingConnection) {
			HostAddress address(fRemoteAddress);
			fReceiver = spawn_thread(_ReceiverEntry,
				(BString("receiver ") += address.ToString()).String(),
				B_NORMAL_PRIORITY,
				this);
			if (fReceiver < 0)
				return fReceiver;
		}
		fNetEndPoint = endPoint;
	} else {
		// create endpoint
		fNetEndPoint = new(nothrow) BNetEndpoint;
		if (!fNetEndPoint)
			return B_NO_MEMORY;
		// check end point
		error = fNetEndPoint->InitCheck();
		if (error != B_OK)
			return error;
		// connect
		error = fNetEndPoint->Connect(GetRemoteAddress());
		if (error != B_OK)
			return error;
	}
	// everything went fine
	fValid = true;
	return B_OK;
}

// IsValid
bool
Connection::IsValid() const
{
	return (!fTerminating && fValid);
}

// GetRemoteAddress
const BNetAddress&
Connection::GetRemoteAddress() const
{
	return fRemoteAddress;
}

// SetRequestHandler
void
Connection::SetRequestHandler(RequestHandler* requestHandler)
{
	fRequestHandler = requestHandler;
	// resume the thread
	if (fReceivingConnection)
		resume_thread(fReceiver);
}

// SendRequest
status_t
Connection::SendRequest(const Request* request)
{
PRINT(("Connection::SendRequest()\n"));
	BAutolock sendLocker(fSendLock);
	if (!sendLocker.IsLocked())
		RETURN_ERROR(B_ERROR);
	// prepare send buffer
	RequestHeader header;
	request->GetHeader(&header);
PRINT(("SendRequest(): header: overall size: %ld, request size: %ld, "
"parameter type: %lu, message size: %ld, data size: %ld\n",
header.GetOverallSize(), header.requestSize, header.parameterType,
header.messageSize, header.dataSize));
	// request message
	BNetBuffer buffer(header.GetOverallSize());
	status_t error = header.Write(buffer);
	if (error != B_OK)
		RETURN_ERROR(error);
PRINT(("  buffer size: %lu\n", buffer.Size()));
	error = _AddMessage(buffer, request->GetRequestMessage());
	if (error != B_OK)
		RETURN_ERROR(error);
PRINT(("  buffer size: %lu\n", buffer.Size()));
	// parameter
	const uint8* data = NULL;
	int32 dataSize = 0;
	const Message* parameter = request->GetParameter();
	if (parameter && !parameter->IsEmpty()) {
		// message
		if (parameter->GetMessage())
			SET_ERROR(error, _AddMessage(buffer, parameter->GetMessage()));
		if (error != B_OK)
			RETURN_ERROR(error);
PRINT(("  buffer size: %lu\n", buffer.Size()));
		// data
		if (parameter->GetDataSize() > 0) {
			SET_ERROR(error, buffer.AppendData(parameter->GetData(),
				parameter->GetDataSize()));
		}
		if (error != B_OK)
			RETURN_ERROR(error);
	}
	// send package
PRINT(("  sending buffer: size: %lu + %ld\n", buffer.Size(), dataSize));
	if ((int32)buffer.Size() + dataSize != header.GetOverallSize()) {
		FATAL(("Buffer size is not overall size: %lu + %ld vs. %ld\n",
			buffer.Size(), dataSize, header.GetOverallSize()));
		RETURN_ERROR(B_ERROR);
	}
	error = Send(buffer);
	if (error != B_OK)
		RETURN_ERROR(error);
	// send data
	if (dataSize > 0) {
		error = Send(data, dataSize);
		if (error != B_OK)
			RETURN_ERROR(error);
	}
PRINT(("Connection::SendRequest() done\n"));
	return B_OK;
}

// SendRequest
status_t
Connection::SendRequest(BMessage* requestMessage, Message* parameter)
{
	Request request(requestMessage, parameter);
	status_t result = SendRequest(&request);
	// unset the request message -- the caller retains ownership
	request.SetRequestMessage(NULL);
	return result;
}

// Receive
status_t
Connection::Receive(BNetBuffer& buffer, size_t size)
{
	int32 remainingRetries = kMaxInterruptedTxCount;
	while (size > 0) {
		int32 bytesRead = fNetEndPoint->Receive(buffer, size);
		if (bytesRead < 0) {
			if (fTerminating || fNetEndPoint->Error() != B_INTERRUPTED
				|| --remainingRetries < 0) {
				return fNetEndPoint->Error();
			}
PRINT(("Connection::Receive(): B_INTERRUPTED ignored\n"));
		} else if (bytesRead == 0) {
			// This seems to indicate that the remote peer closed the
			// connection.
			PRINT(("Connection::Receive(): received 0 bytes. Assuming "
				"that the remote peer closed the connection\n"));
PRINT(("end point error: %lx\n", (int32)fNetEndPoint->Error()));
			return B_ERROR;
		} else {
			size -= bytesRead;
			remainingRetries = kMaxInterruptedTxCount;
		}
	}
	return B_OK;
}

// Receive
status_t
Connection::Receive(uint8* buffer, size_t size)
{
	int32 remainingRetries = kMaxInterruptedTxCount;
	while (size > 0) {
		int32 bytesRead = fNetEndPoint->Receive(buffer, size);
		if (bytesRead < 0) {
			if (fTerminating || fNetEndPoint->Error() != B_INTERRUPTED
				|| --remainingRetries < 0) {
				return fNetEndPoint->Error();
			}
PRINT(("Connection::Receive(): B_INTERRUPTED ignored\n"));
		} else if (bytesRead == 0) {
			// This seems to indicate that the remote peer closed the
			// connection.
			PRINT(("Connection::Receive(): received 0 bytes. Assuming "
				"that the remote peer closed the connection\n"));
			return B_ERROR;
		} else {
			size -= bytesRead;
			buffer += bytesRead;
			remainingRetries = kMaxInterruptedTxCount;
PRINT(("Connection::Receive(): received %ld bytes, still to read: %lu\n", bytesRead, size));
		}
	}
	return B_OK;
}

// Send
status_t
Connection::Send(BNetBuffer& buffer)
{
	return Send((uint8*)buffer.Data(), buffer.Size());
}

// Send
status_t
Connection::Send(const uint8* buffer, size_t size)
{
	int32 remainingRetries = kMaxInterruptedTxCount;
	while (size > 0) {
		int32 bytesSent = fNetEndPoint->Send(buffer, size);
		if (bytesSent < 0) {
			if (fTerminating || fNetEndPoint->Error() != B_INTERRUPTED
				|| --remainingRetries < 0) {
				return fNetEndPoint->Error();
			}
PRINT(("Connection::Send(): B_INTERRUPTED ignored, remaining retries: %ld\n",
remainingRetries));
			bytesSent = 0;
		} else if (bytesSent == 0) {
			// This seems to indicate that the remote peer closed the
			// connection.
			PRINT(("Connection::Send(): sent 0 bytes. Assuming "
				"that the remote peer closed the connection\n"));
			return B_ERROR;
		} else {
			size -= bytesSent;
			buffer += bytesSent;
			remainingRetries = kMaxInterruptedTxCount;
PRINT(("Connection::Send(): sent %ld bytes, still to sent: %lu\n", bytesSent,
size));
		}
	}
	return B_OK;
}

// ReceiveRequestHeader
status_t
Connection::ReceiveRequestHeader(BNetBuffer& buffer,
	RequestHeader& header)
{
	status_t error = Receive(buffer, RequestHeader::GetHeaderSize());
	if (error != B_OK)
		return error;
	return header.Read(buffer);
}

// _ReceiveMessage
status_t
Connection::_ReceiveMessage(int32 size, BMessage** _message)
{
	BNetBuffer messageBuffer(size);
PRINT(("waiting for message...\n"));
	// receive the buffer
	status_t error = Receive(messageBuffer, size);
	if (error != B_OK)
		RETURN_ERROR(error);
PRINT(("successfully received message\n"));
	// allocate a message
	BMessage* message = new BMessage;
	if (!message)
		RETURN_ERROR(B_NO_MEMORY);
	ObjectDeleter<BMessage> deleter(message);
	// remove the message from the buffer
	error = _RemoveMessage(messageBuffer, message, size);
	if (error != B_OK)
		RETURN_ERROR(error);
	if (messageBuffer.BytesRemaining() != 0)
		RETURN_ERROR(B_BAD_DATA);
	// set result
	deleter.Detach();
	*_message = message;
	return error;
}

// ReceiveRequest
status_t
Connection::ReceiveRequest(Request& request)
{
PRINT(("Connection::ReceiveRequest()\n"));
PRINT(("waiting for request...\n"));
	// get header
	BNetBuffer headerBuffer(RequestHeader::GetHeaderSize());
//BNetBuffer headerBuffer(1024);
	RequestHeader header;
	status_t error = ReceiveRequestHeader(headerBuffer, header);
	if (error != B_OK)
		RETURN_ERROR(error);
PRINT(("successfully received request header\n"));
PRINT(("ReceiveRequest(): header: overall size: %ld, request size: %ld, "
"parameter type: %lu, message size: %ld, data size: %ld\n",
header.GetOverallSize(), header.requestSize, header.parameterType,
header.messageSize, header.dataSize));
	// receive request message
PRINT(("waiting for request message (%ld bytes)...\n", header.requestSize));
	BMessage* requestMessage = NULL;
	error = _ReceiveMessage(header.requestSize, &requestMessage);
	if (error != B_OK)
		RETURN_ERROR(error);
PRINT(("successfully received request message\n"));
	request.SetRequestMessage(requestMessage);
	// get parameter
	switch (header.parameterType) {
		case REQUEST_NO_PARAMETER:
			break;
		case REQUEST_MESSAGE_PARAMETER:
		{
			Message* parameter = new(nothrow) Message;
			if (!parameter)
				RETURN_ERROR(B_NO_MEMORY);
			MessageReference reference(parameter, false);
			BMessage* message = NULL;
PRINT(("waiting for parameter message (%ld bytes)...\n", header.messageSize));
			error = _ReceiveMessage(header.messageSize, &message);
			if (error != B_OK)
				RETURN_ERROR(error);
PRINT(("successfully received parameter message\n"));
			parameter->SetTo(message);
			request.SetParameter(parameter);
			break;
		}
		case REQUEST_DATA_PARAMETER:
		{
			Message* parameter = new(nothrow) Message;
			if (!parameter)
				RETURN_ERROR(B_NO_MEMORY);
			MessageReference parameterReference(parameter, false);
			// message
			BMessage* message = NULL;
			if (header.messageSize > 0) {
PRINT(("waiting for parameter message (%ld bytes)...\n", header.messageSize));
				error = _ReceiveMessage(header.messageSize, &message);
				if (error != B_OK)
					RETURN_ERROR(error);
PRINT(("successfully received parameter message\n"));
			}
			ObjectDeleter<BMessage> deleter(message);
			// allocate the data
			RemoteData* parameterData = NULL;
			error = RemoteData::Create(header.dataSize, &parameterData);
			if (error != B_OK)
				RETURN_ERROR(error);
			// set the request parameter
			parameter->SetTo(message, parameterData);
			parameterData->RemoveReference();
			deleter.Detach();
			request.SetParameter(parameter);
			// receive the data
PRINT(("waiting for parameter data (%ld bytes)...\n", parameterData->GetSize()));
			error = Receive(parameterData->GetData(), parameterData->GetSize());
			if (error != B_OK)
				RETURN_ERROR(error);
PRINT(("successfully received parameter data\n"));
			break;
		}
		default:
			RETURN_ERROR(B_BAD_DATA);
	}
PRINT(("Connection::ReceiveRequest() done\n"));
	return error;
}

// _AddMessage
status_t
Connection::_AddMessage(BNetBuffer& buffer, const BMessage* message)
{
	int32 size = message->FlattenedSize();
	uint8* tmpBuffer = new(nothrow) uint8[size];
	if (!tmpBuffer)
		return B_NO_MEMORY;
	ArrayDeleter<uint8> _(tmpBuffer);
	status_t error = message->Flatten((char*)tmpBuffer, size);
	if (error != B_OK)
		return error;
	return buffer.AppendData(tmpBuffer, size);
}

// _RemoveMessage
status_t
Connection::_RemoveMessage(BNetBuffer& buffer, BMessage* message,
	size_t size)
{
	uint8* tmpBuffer = new(nothrow) uint8[size];
	if (!tmpBuffer)
		return B_NO_MEMORY;
	ArrayDeleter<uint8> _(tmpBuffer);
	status_t error = buffer.RemoveData(tmpBuffer, size);
	if (error != B_OK)
		return error;
	BMemoryIO io(tmpBuffer, size);
	return message->Unflatten(&io);
}

// _ReceiverEntry
int32
Connection::_ReceiverEntry(void* data)
{
	return ((Connection*)data)->_Receiver();
}

// _Receiver
int32
Connection::_Receiver()
{
	status_t error = B_OK;
	while (!fTerminating && error == B_OK) {
		// create a request
		Request* request = new(nothrow) Request;
		// receive the request
		if (request) {
			RequestReference reference(request, false);
			error = ReceiveRequest(*request);
			if (error == B_OK) {
				if (fRequestHandler)
					fRequestHandler->HandleRequest(this, request);
			}
		} else
			error = B_NO_MEMORY;
		// handle errors
		if (error != B_OK && !fTerminating) {
			FATAL(("Connection::_Receiver(): error receiving "
				"request: %s\n", strerror(error)));
			if (fRequestHandler) {
				if (fRequestHandler->ErrorOccurred(this, error))
					error = B_OK;
				else
					fValid = false;
			}
		}
	}
	return 0;
}

