// LocalMessageRequest.cpp
// 
// Copyright (c) 2004, Ingo Weinhold (bonefish@cs.tu-berlin.de)
// 
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation 
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
// THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
// DEALINGS IN THE SOFTWARE.
// 
// Except as contained in this notice, the name of a copyright holder shall
// not be used in advertising or otherwise to promote the sale, use or other
// dealings in this Software without prior written authorization of the
// copyright holder.

#include <new>

#include <String.h>

#include "Debug.h"
#include "LocalMessageDeliverer.h"

// number of threads delivering messages
static const int32 kDeliveryThreadCount = 5;

// LocalMessageRequest

// constructor
LocalMessageRequest::LocalMessageRequest()
	: fTarget(),
	  fReplyToMessage(NULL),
	  fMessage(NULL),
	  fReplyMessage(NULL)
{
}

// constructor
LocalMessageRequest::LocalMessageRequest(BMessenger target, BMessage* message,
	BMessage* replyMessage)
	: fTarget(target),
	  fReplyToMessage(NULL),
	  fMessage(message),
	  fReplyMessage(replyMessage)
{
}

// constructor
LocalMessageRequest::LocalMessageRequest(BMessage* replyToMessage,
	BMessage* message, BMessage* replyMessage)
	: fTarget(),
	  fReplyToMessage(replyToMessage),
	  fMessage(message),
	  fReplyMessage(replyMessage)
{
}

// destructor
LocalMessageRequest::~LocalMessageRequest()
{
}

// GetTarget
BMessenger
LocalMessageRequest::GetTarget()
{
	return fTarget;
}

// GetReplyToMessage
BMessage*
LocalMessageRequest::GetReplyToMessage()
{
	return fReplyToMessage;
}

// GetMessage
BMessage*
LocalMessageRequest::GetMessage()
{
	return fMessage;
}

// GetReplyMessage
BMessage*
LocalMessageRequest::GetReplyMessage()
{
	return fReplyMessage;
}

// HandleReply
void
LocalMessageRequest::HandleReply(BMessage* reply)
{
}

// DeliveryFailed
void
LocalMessageRequest::DeliveryFailed(status_t error)
{
}


// DeliveryThreadInfo
class LocalMessageDeliverer::DeliveryThreadInfo : public BLocker {
public:
	DeliveryThreadInfo()
		: fThread(-1),
		  fDelivering(false)
	{
	}

	void SetThread(thread_id thread)
	{
		fThread = thread;
	}

	thread_id GetThread() const
	{
		return fThread;
	}

	void SetDelivering(bool delivering)
	{
		fDelivering =  delivering;
	}

	bool IsDelivering() const
	{
		return fDelivering;
	}

private:
	thread_id	fThread;
	bool		fDelivering;
};


// LocalMessageDeliverer

// constructor
LocalMessageDeliverer::LocalMessageDeliverer()
	: fDeliveryThreads(NULL),
	  fPendingRequests(),
	  fTerminating(false)
{
}

// destructor
LocalMessageDeliverer::~LocalMessageDeliverer()
{
	Uninit();
}

// Init
status_t
LocalMessageDeliverer::Init()
{
	// pending requests
	status_t error = fPendingRequests.InitCheck();
	if (error != B_OK)
		return error;
	// allocate delivery thread infos
	fDeliveryThreads = new(nothrow) DeliveryThreadInfo[kDeliveryThreadCount];
	if (!fDeliveryThreads)
		return B_NO_MEMORY;
	// spawn delivery threads
	for (int32 i = 0; i < kDeliveryThreadCount; i++) {
		DeliveryThreadInfo& info = fDeliveryThreads[i];
		thread_id thread = spawn_thread(_DeliveryThreadEntry,
			(BString("message delivery ") << i).String(),
			B_NORMAL_PRIORITY, this);
		if (thread < 0)
			return thread;
		info.SetThread(thread);
	}
	// resume the threads
	for (int32 i = 0; i < kDeliveryThreadCount; i++)
		resume_thread(fDeliveryThreads[i].GetThread());
	return B_OK;
}

// Uninit
void
LocalMessageDeliverer::Uninit()
{
	if (fTerminating)
		return;
	fTerminating = false;
	// unblock waiting delivery threads
	const BList* requests = NULL;
	fPendingRequests.Close(false, &requests);
	// terminate the threads
	if (fDeliveryThreads) {
		for (int32 i = 0; i < kDeliveryThreadCount; i++) {
			DeliveryThreadInfo& info = fDeliveryThreads[i];
			// if just delivering, unblock the thread
			info.Lock();
			if (info.IsDelivering())
				suspend_thread(info.GetThread());
			info.Unlock();
			int32 result;
			wait_for_thread(info.GetThread(), &result);
		}
		delete[] fDeliveryThreads;
		fDeliveryThreads = NULL;
	}
	// delete the pending requests
	for (int32 i = 0;
		 LocalMessageRequest* request
		 	= (LocalMessageRequest*)requests->ItemAt(i);
		 i++) {
		request->DeliveryFailed(B_ERROR);
		delete request;
	}
}

// PushRequest
status_t
LocalMessageDeliverer::PushRequest(LocalMessageRequest* request)
{
	if (!request)
		return B_BAD_VALUE;
	return fPendingRequests.Push(request);
}

// Dump
void
LocalMessageDeliverer::Dump()
{
	PRINT(("DUMP: LocalMessageDeliverer\n"));
	PRINT(("DUMP: terminating: %d\n", fTerminating));
	PRINT(("DUMP: threads:\n"));
	if (fDeliveryThreads) {
		for (int32 i = 0; i < kDeliveryThreadCount; i++) {
			DeliveryThreadInfo& info = fDeliveryThreads[i];
			TOUCH(info);
			PRINT(("DUMP:   thread %ld: %ld, delivering: %d\n", i,
				info.GetThread(), info.IsDelivering()));
		}
	}
	fPendingRequests.Lock();
	PRINT(("DUMP: requests: %ld\n", fPendingRequests.Size()));
	fPendingRequests.Unlock();
}

// _RequestProcessorEntry
int32
LocalMessageDeliverer::_DeliveryThreadEntry(void* data)
{
	return ((LocalMessageDeliverer*)data)->_DeliveryThread();
}

// _RequestProcessor
int32
LocalMessageDeliverer::_DeliveryThread()
{
	// get delivery thread info
	DeliveryThreadInfo* info = NULL;
	thread_id me = find_thread(NULL);
	for (int32 i = 0; i < kDeliveryThreadCount; i++) {
		DeliveryThreadInfo& threadInfo = fDeliveryThreads[i];
		if (threadInfo.GetThread() == me) {
			info = &threadInfo;
			break;
		}
	}
	if (!info) {
		FATAL(("LocalMessageDeliverer::_DeliveryThread(): failed to find "
			"delivery thread info!\n"));
		return B_ERROR;
	}
	// process requests
	while (!fTerminating) {
		LocalMessageRequest* request = NULL;
		status_t error = fPendingRequests.Pop(&request);
		if (error != B_OK || !request)
			continue;
		// handle request
		// get message and reply message
		BMessage* message = request->GetMessage();
		BMessage* replyMessage = request->GetReplyMessage();
		BMessage _replyMessage;
		if (!replyMessage)
			replyMessage = &_replyMessage;
		// prepare for sending
		error = (message ? B_OK : B_BAD_VALUE);
		if (error == B_OK) {
			info->Lock();
			if (fTerminating)
				error = B_ERROR;
			else
				info->SetDelivering(true);
			info->Unlock();
		}
		// send the message
		if (error == B_OK) {
PRINT(("LocalMessageDeliverer::_DeliveryThread(): delivering message:\n"));
D(message->PrintToStream());
			BMessage* replyToMessage = request->GetReplyToMessage();
			if (replyToMessage)
				error = replyToMessage->SendReply(message, replyMessage);
			else
				error = request->GetTarget().SendMessage(message, replyMessage);
		}
		info->Lock();
		info->SetDelivering(false);
		info->Unlock();
		// notify the request of the result
		if (error == B_OK)
			request->HandleReply(replyMessage);
		else
			request->DeliveryFailed(error);
		delete request;
	}
	return B_OK;
}

