// RMessageServer.cpp
// 
// Copyright (c) 2004, Ingo Weinhold (bonefish@cs.tu-berlin.de)
// 
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation 
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
// THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
// DEALINGS IN THE SOFTWARE.
// 
// Except as contained in this notice, the name of a copyright holder shall
// not be used in advertising or otherwise to promote the sale, use or other
// dealings in this Software without prior written authorization of the
// copyright holder.

#include <NetEndpoint.h>
#include <OS.h>
#include <PropertyInfo.h>
#include <Roster.h>

#include "AutoDeleter.h"
#include "Connection.h"
#include "ConnectionFactory.h"
#include "Debug.h"
#include "DebugSupport.h"
#include "EventQueue.h"
#include "LocalMessageDeliverer.h"
#include "RemoteRosterImpl.h"
#include "RemoteServer.h"
#include "RMessageServer.h"
#include "RMessageServerDefs.h"

// usage
const char* kUsage =
"Usage: RMessageServer <options>\n"
"Options:\n"
"  -h, --help     - print this text\n"
"  --port <port>  - the port the server shall use\n"
"\n";

// server property infos
static property_info kRMessageServerProperties[] = {
	{ "Host", { B_GET_PROPERTY, 0 }, { B_DIRECT_SPECIFIER, 0 },
		"get connected hosts", 0 },
	{ "Host", { 0 }, { B_NAME_SPECIFIER, 0 },
		"specify host by address/name", 0 },
	{ 0, {}, {}, 0, 0 }
};

// "Host" property infos
static property_info kHostProperties[] = {
	{ "Application", { B_GET_PROPERTY, 0 }, { B_DIRECT_SPECIFIER, 0 },
		"get running applications", 0 },
	{ "Application", { 0 }, { B_NAME_SPECIFIER, 0 },
		"specify application by signature/name", 0 },
	{ "Suites", { B_GET_PROPERTY, 0 }, { B_DIRECT_SPECIFIER, 0 },
		"get supported suites", 0 },
	{ 0, {}, {}, 0, 0 }
};

// constructor
RMessageServer::RMessageServer(int port)
	: BApplication(kRMessageServerSignature),
	  fPort(port),
	  fConnectionFactory(port),
	  fMessageDeliverer(),
	  fEventQueue(),
	  fRemoteServers(),
	  fPendingRequests(),
	  fPendingScriptingRequests(),
	  fRequestProcessor(-1),
	  fTerminating(false)
{
}

// destructor
RMessageServer::~RMessageServer()
{
	fTerminating = true;
	// unblock the connection listener and the request processor, and wait for
	// them to terminate
	fIncomingConnections.Close();
	fPendingRequests.Close(true);
	fPendingScriptingRequests.Close(true);
	fEventQueue.Die();
	if (fConnectionListener >= 0) {
		int32 result;
		wait_for_thread(fConnectionListener, &result);
	}
	fMessageDeliverer.Uninit();
	if (fScriptingProcessor >= 0) {
		int32 result;
		wait_for_thread(fScriptingProcessor, &result);
	}
 	if (fRequestProcessor >= 0) {
		int32 result;
		wait_for_thread(fRequestProcessor, &result);
	}
	// now close the connections to the remote servers
	for (RemoteServerMap::Iterator it = fRemoteServers.GetIterator();
		it.HasNext();) {
		RemoteServer* server = it.Next().value;
		server->Uninit();
		server->Lock();
		server->Quit();
	}
}

// Init
status_t
RMessageServer::Init()
{
	status_t error = fPendingRequests.InitCheck();
	if (error != B_OK)
		return error;
	error = fPendingScriptingRequests.InitCheck();
	if (error != B_OK)
		return error;
	error = fEventQueue.InitCheck();
	if (error != B_OK)
		return error;
	error = fMessageDeliverer.Init();
	if (error != B_OK)
		return error;
	// init incoming connection end point
	error = fIncomingConnections.InitCheck();
	if (error != B_OK)
		return error;
	error = fIncomingConnections.Bind(fPort);
	if (error != B_OK)
		return error;
	error = fIncomingConnections.Listen();
	if (error != B_OK)
		return error;
	// spawn request processor thread
	fRequestProcessor = spawn_thread(_RequestProcessorEntry,
		"main request processor", B_NORMAL_PRIORITY, this);
	if (fRequestProcessor < 0)
		return fRequestProcessor;
	// spawn scripting request processor thread
	fScriptingProcessor = spawn_thread(_ScriptingProcessorEntry,
		"scripting processor", B_NORMAL_PRIORITY, this);
	if (fScriptingProcessor < 0)
		return fScriptingProcessor;
	// spawn connection listener thread
	fConnectionListener = spawn_thread(_ConnectionListenerEntry,
		"connection listener", B_NORMAL_PRIORITY, this);
	if (fConnectionListener < 0)
		return fConnectionListener;
	// resume the threads
	resume_thread(fRequestProcessor);
	resume_thread(fScriptingProcessor);
	resume_thread(fConnectionListener);
	return B_OK;
}

// ReadyToRun
void
RMessageServer::ReadyToRun()
{
}

// MessageReceived
void
RMessageServer::MessageReceived(BMessage* message)
{
char dbgBuffer[5];
TOUCH(dbgBuffer);
PRINT(("RMessageServer::MessageReceived(%s)\n",
message_what(message, dbgBuffer)));
	// handle remote scripting messages
	int32 index;
	BMessage specifier;
	int32 form;
	const char* property;
	const char* name;
	if (message->GetCurrentSpecifier(&index, &specifier, &form, &property)
		== B_OK && !strcmp(property, "Host")) {
PRINT(("  scripting message: Host property\n"));
		if (form == B_NAME_SPECIFIER
			&& specifier.FindString("name", &name) == B_OK) {
PRINT(("  name specifier\n"));
			DetachCurrentMessage();
			if (fPendingScriptingRequests.Push(message) != B_OK)
				delete message;
			return;
		} else if (form == B_DIRECT_SPECIFIER
			&& message->what == B_GET_PROPERTY) {
PRINT(("  direct specifier and B_GET_PROPERTY\n"));
			BMessage reply(B_REPLY);
			reply.AddInt32("error", B_OK);
			BAutolock _(fRemoteServers);
			for (RemoteServerMap::Iterator it = fRemoteServers.GetIterator();
				it.HasNext();) {
				BString address(it.Next().key.ToString());
				if (address.Length() > 0)
					reply.AddString("result", address.String());
			}
			message->SendReply(&reply);
			return;
		}
	}
	// handle other messages
	switch (message->what) {
		case MSG_CONNECT:
			PRINT(("RMessageServer::MessageReceived(): MSG_CONNECT\n"));
			DetachCurrentMessage();
			if (fPendingRequests.Push(message) != B_OK) {
				FATAL(("  Failed to push request message\n"));
				delete message;
			}
			break;
		case MSG_DUMP:
			_Dump();
			break;
		default:
			BApplication::MessageReceived(message);
			break;
	}
}

// GetSupportedSuites
status_t
RMessageServer::GetSupportedSuites(BMessage* message)
{
	// add suite name
	status_t error
		= message->AddString("suites", "suite/vnd.bonefish.rmessage_server");
	if (error != B_OK)
		return error;
	// add property info
	BPropertyInfo info(kRMessageServerProperties);
	error = message->AddFlat("messages", &info);
	if (error != B_OK)
		return error;
	return BApplication::GetSupportedSuites(message);
}

// ResolveSpecifier
BHandler*
RMessageServer::ResolveSpecifier(BMessage* message, int32 index,
	BMessage* specifier, int32 form, const char* property)
{
PRINT(("RMessageServer::ResolveSpecifier()\n"));
	const char* name;
	if (!strcmp(property, "Host")) {
		if (form == B_NAME_SPECIFIER
			&& specifier->FindString("name", &name) == B_OK) {
			return this;
		} else if (form == B_DIRECT_SPECIFIER && index == 0
			&& message->what == B_GET_PROPERTY) {
			return this;
		}
	}
	return BApplication::ResolveSpecifier(message, index, specifier, form,
		property);
}

// _GetRemoteServer
status_t
RMessageServer::_GetRemoteServer(const BNetAddress& _address,
	Connection* receivingConnection, RemoteServer** remoteServer)
{
	HostAddress address(_address);
	status_t error = InitCheck();
	if (error != B_OK)
		return error;
	fRemoteServers.Lock();
PRINT(("  REMOTE SERVERS LOCK\n"));
	RemoteServer* server = fRemoteServers.Get(address);
	if (server) {
		if (receivingConnection)
			error = server->AttachReceivingConnection(receivingConnection);
	} else {
		// no remote server instance yet: create one
		// create a new sending connection
		fRemoteServers.Unlock();	// deadlock prevention
		Connection* sendingConnection = NULL;
		error = fConnectionFactory.CreateConnection(address,
			&sendingConnection);
		fRemoteServers.Lock();
		if (error == B_OK) {
			// check whether there's still no remote server
			server = fRemoteServers.Get(address);
			if (server) {
				// oh, there is one now: just attach the sending connection
				server->AttachSendingConnection(sendingConnection);
			} else {
				// create the remote server
				server = new(nothrow) RemoteServer(address, &fConnectionFactory,
					&fMessageDeliverer, &fEventQueue);
				if (server) {
					error = server->Init(receivingConnection,
						sendingConnection);
					if (error == B_OK)
						error = fRemoteServers.Put(address, server);
					if (error != B_OK) {
						delete server;
						server = NULL;
					}
				} else
					error = B_NO_MEMORY;
				if (error != B_OK)
					delete sendingConnection;
			}
		}
	}
PRINT(("  REMOTE SERVERS UNLOCK\n"));
	fRemoteServers.Unlock();
	if (error == B_OK)
		*remoteServer = server;
	return error;
}

// _HandleConnectRequest
void
RMessageServer::_HandleConnectRequest(BMessage* request)
{
PRINT(("RMessageServer::_HandleConnectRequest()\n"));
	// get address
	BMessage addressArchive;
	status_t error = request->FindMessage("address", &addressArchive);
	BNetAddress address(&addressArchive);
	if (error == B_OK)
		SET_ERROR(error, address.InitCheck());
	// get a remote server for the address
	RemoteServer* remoteServer = NULL;
	if (error == B_OK)
		SET_ERROR(error, _GetRemoteServer(address, NULL, &remoteServer));
	// prepare reply
	BMessage reply(MSG_REPLY);
	if (error == B_OK)
		reply.AddMessenger("request messenger", BMessenger(remoteServer));
	else
		reply.AddInt32("error", error);
	// send reply
	request->SendReply(&reply, (BHandler*)NULL, 0LL);
PRINT(("RMessageServer::_HandleConnectRequest() done\n"));
}

// _HandleScriptingRequest
/*void
RMessageServer::_HandleScriptingRequest(BMessage* request)
{
PRINT(("RMessageServer::_HandleScriptingRequest()\n"));
	// get message
	BMessage* message;
	if (request->FindPointer("message", (void**)&message) != B_OK) {
PRINT(("RMessageServer::_HandleScriptingRequest() done 1\n"));
		return;
	}
	ObjectDeleter<BMessage> messageDeleter(message);
	// get remote host
	const char* host;
	if (request->FindString("host", &host) != B_OK) {
PRINT(("RMessageServer::_HandleScriptingRequest() done 2\n"));
		return;
	}
	// get its address
	BNetAddress address(host);
	if (address.InitCheck() != B_OK) {
PRINT(("RMessageServer::_HandleScriptingRequest() done 3\n"));
		return;
	}
	// get a remote server for the address
	RemoteServer* remoteServer = NULL;
	if (_GetRemoteServer(address, NULL, &remoteServer) != B_OK) {
PRINT(("RMessageServer::_HandleScriptingRequest() done 4\n"));
		return;
	}
	// let the remote server do the rest
	remoteServer->HandleScriptingMessage(message);
	messageDeleter.Detach();
PRINT(("RMessageServer::_HandleScriptingRequest() done\n"));
}
*/

// _GetSupportedHostSuites
void
RMessageServer::_GetSupportedHostSuites(BMessage* request)
{
	BMessage reply(B_REPLY);
	// add suite name
	status_t error
		= reply.AddString("suites", "suite/vnd.bonefish.rmessage_host");
	// add property info
	BPropertyInfo info(kHostProperties);
	if (error == B_OK)
		error = reply.AddFlat("messages", &info);
	reply.AddInt32("error", error);
	request->SendReply(&reply);
}

// _Dump
void
RMessageServer::_Dump()
{
	PRINT(("DUMP: RMessageServer\n"));
	PRINT(("DUMP: terminating: %d\n", fTerminating));
	PRINT(("DUMP: request processor: %ld\n", fRequestProcessor));
	PRINT(("DUMP: scripting processor: %ld\n", fScriptingProcessor));
	PRINT(("DUMP: connection listener: %ld\n", fConnectionListener));
	PRINT(("DUMP: incoming connection: %s\n",
		strerror(fIncomingConnections.InitCheck())));
	fPendingRequests.Lock();
	PRINT(("DUMP: pending requests: %ld\n", fPendingRequests.Size()));
	fPendingRequests.Unlock();
	fPendingScriptingRequests.Lock();
	PRINT(("DUMP: pending scripting requests: %ld\n",
		fPendingScriptingRequests.Size()));
	fPendingScriptingRequests.Unlock();
	fMessageDeliverer.Dump();
	fRemoteServers.Lock();
	for (RemoteServerMap::Iterator it = fRemoteServers.GetIterator();
		it.HasNext();) {
		it.Next().value->Dump();
	}
	fRemoteServers.Unlock();
//EventQueue			fEventQueue;
}

// _RequestProcessorEntry
int32
RMessageServer::_RequestProcessorEntry(void* data)
{
	return ((RMessageServer*)data)->_RequestProcessor();
}

// _RequestProcessor
int32
RMessageServer::_RequestProcessor()
{
	while (!fTerminating) {
		BMessage* request = NULL;
		status_t error = fPendingRequests.Pop(&request);
		if (error != B_OK || !request)
			continue;
		PRINT(("RMessageServer::_RequestProcessor(): Got request\n"));
		// handle request
		switch (request->what) {
			case MSG_CONNECT:
				_HandleConnectRequest(request);
				break;
			default:
			{
				char dbgBuffer[5];
				TOUCH(dbgBuffer);
				FATAL(("Unknown request: %s\n",
					message_what(request, dbgBuffer)));
				break;
			}
		}
		delete request;
	}
	return B_OK;
}

// _ScriptingProcessorEntry
int32
RMessageServer::_ScriptingProcessorEntry(void* data)
{
	return ((RMessageServer*)data)->_ScriptingProcessor();
}

// _ScriptingProcessor
int32
RMessageServer::_ScriptingProcessor()
{
	PRINT(("RMessageServer::_ScriptingProcessor()\n"));
	while (!fTerminating) {
		BMessage* request = NULL;
		status_t error = fPendingScriptingRequests.Pop(&request);
		if (error != B_OK || !request)
			continue;
		PRINT(("RMessageServer::_ScriptingProcessor(): Got request\n"));
		ObjectDeleter<BMessage> deleter(request);
		// handle request
		// get host
		int32 index;
		BMessage specifier;
		int32 form;
		const char* property;
		const char* host;
		if (request->GetCurrentSpecifier(&index, &specifier, &form, &property)
			!= B_OK || strcmp(property, "Host") != 0) {
			PRINT(("  Failed to get specifier or property is not Host.\n"));
			continue;
		}
		if (form != B_NAME_SPECIFIER
			|| specifier.FindString("name", &host) != B_OK) {
			PRINT(("  No NAME specifier.\n"));
			continue;
		}
		request->PopSpecifier();
		// get remote host address
		BNetAddress address(host);
		if (address.InitCheck() != B_OK) {
			PRINT(("  Failed to get host address.\n"));
			continue;
		}
		// get a remote server for the address
		RemoteServer* remoteServer = NULL;
		if (_GetRemoteServer(address, NULL, &remoteServer) != B_OK) {
			PRINT(("  Failed to get remote server.\n"));
			continue;
		}
		// get the sub-property
		if (index == 0) {
			// no sub-property: only B_GET_SUPPORTED_SUITES is allowed
			if (request->what == B_GET_SUPPORTED_SUITES) {
				_GetSupportedHostSuites(request);
				continue;
			}
			PRINT(("  Unknown scripting message.\n"));
			continue;
		}
		if (request->GetCurrentSpecifier(&index, &specifier, &form, &property)
			!= B_OK) {
			PRINT(("  Failed to get specifier for sub-Host property.\n"));
			continue;
		}
		if (strcmp(property, "Suites") == 0 && index == 0
			&& request->what == B_GET_PROPERTY) {
			_GetSupportedHostSuites(request);
			continue;
		} else if (strcmp(property, "Application") != 0) {
			PRINT(("  Unknown sub-Host property: %s.\n", property));
			continue;
		}
		// it's the Application property from here on
		// init a remote roster we'll use to get the info we need
		RemoteRosterImpl roster;
		if (roster.SetTo(BMessenger(remoteServer)) != B_OK) {
			PRINT(("  Failed to init remote roster.\n"));
			continue;
		}
		// now let's see what kind of message we got
		if (form == B_NAME_SPECIFIER) {
			// a scripting message that shall be forwarded to the remote
			// application
			PRINT(("  NAME specifier.\n"));
			const char* application;
			if (specifier.FindString("name", &application) != B_OK) {
				PRINT(("  Failed to get application.\n"));
				continue;
			}
			request->PopSpecifier();
			// get a messenger for the remote app and forward the message
			BMessenger messenger;
			if (roster.GetMessenger(application, &messenger, true) != B_OK) {
				PRINT(("  Failed to send message.\n"));
				continue;
			}
			messenger.SendMessage(request);
		} else if (form == B_DIRECT_SPECIFIER && index == 0
			&& request->what == B_GET_PROPERTY) {
			// a simple GET on the Application property: we retrieve
			// a list of applications
			PRINT(("  direct specifier and B_GET_PROPERTY.\n"));
			// get teams
			BList teams;
			if (roster.GetAppList(&teams) != B_OK) {
				PRINT(("  Failed to get team list.\n"));
				continue;
			}
			// get app infos
			BMessage reply(B_REPLY);
			int32 count = teams.CountItems();
			for (int32 i = 0; i < count; i++) {
				team_id team = (team_id)teams.ItemAt(i);
				app_info info;
				if (roster.GetRunningAppInfo(team, &info) == B_OK) {
					reply.AddString("result", info.signature);
//					reply.AddData("result",  B_MIME_STRING_TYPE, info.signature,
//						strlen(info.signature) + 1, false, 1);
//					reply.AddString("result", info.ref.name);
				}
			}
			// and off goes the reply
			reply.AddInt32("error", B_OK);
			request->SendReply(&reply);
		} else {
			PRINT(("  Unsupported specifier.\n"));
			BMessage reply(B_REPLY);
			reply.AddInt32("error", B_BAD_SCRIPT_SYNTAX);
			request->SendReply(&reply);
		}
	}
	PRINT(("RMessageServer::_ScriptingProcessor() done\n"));
	return B_OK;
}

// _ConnectionListenerEntry
int32
RMessageServer::_ConnectionListenerEntry(void* data)
{
	return ((RMessageServer*)data)->_ConnectionListener();
}

// _ConnectionListener
int32
RMessageServer::_ConnectionListener()
{
	while (!fTerminating) {
		if (BNetEndpoint* endPoint = fIncomingConnections.Accept(100000LL)) {
FUNCTION(("Incoming connection\n"));
			const BNetAddress& address = endPoint->RemoteAddr();
			Connection* connection = NULL;
			status_t error = fConnectionFactory.CreateConnection(endPoint,
				&connection);
			if (error == B_OK) {
				RemoteServer* server;
				error = _GetRemoteServer(address, connection, &server);
			}
			if (error != B_OK) {
				if (connection)
					delete connection;
				else
					delete endPoint;
			}
		}
	}
	return B_OK;
}


// print_usage
static
void
print_usage(bool error)
{
	fprintf((error ? stderr : stdout), kUsage);
}

// main
int
main(int argc, char** argv)
{
	init_debugging();
	FUNCTION_START();
	// parse parameters
	const char* portString = NULL;
	int port = RMESSAGE_DEFAULT_SERVER_PORT;
	int argi = 1;
	for (;argi < argc; argi++) {
		const char* arg = argv[argi];
		int argLen = strlen(arg);
		if (argLen > 0 && arg[0] == '-') {
			// option
			const char* option = arg + 1;
			// help
			if (strcmp(option, "h") == 0 || strcmp(option, "-help") == 0) {
				print_usage(false);
				return 0;
			// port
			} else if (strcmp(option, "-port") == 0) {
				argi++;
				if (argi >= argc) {
					fprintf(stderr, "--port requires a parameter\n");
					print_usage(true);
					return 1;
				}
				portString = argv[argi];
			// unknown
			} else {
				fprintf(stderr, "Illegal option: `-%s'\n", option);
				print_usage(true);
				return 1;
			}
		} else
			break;
	}
	if (argi != argc) {
		print_usage(true);
		return 1;
	}
	// if a port was given, analyze the string
	if (portString) {
		if (sscanf(portString, "%d", &port) < 1 || port < 1) {
			fprintf(stderr, "`%s' does not specify a valid port.\n",
				portString);
			return 1;
		}
	}
	// start application
	RMessageServer app(port);
	status_t error = app.Init();
	if (error == B_OK)
		app.Run();
	else
		fprintf(stderr, "Initialization failed: %s\n", strerror(error));
	FUNCTION_END();
	exit_debugging();
	return 0;
}
