// RemoteServer.cpp
// 
// Copyright (c) 2004, Ingo Weinhold (bonefish@cs.tu-berlin.de)
// 
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation 
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
// THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
// DEALINGS IN THE SOFTWARE.
// 
// Except as contained in this notice, the name of a copyright holder shall
// not be used in advertising or otherwise to promote the sale, use or other
// dealings in this Software without prior written authorization of the
// copyright holder.

#include <new>

#include <Autolock.h>
#include <Roster.h>

#include <RemoteMessage.h>

#include "AutoDeleter.h"
#include "Connection.h"
#include "ConnectionFactory.h"
#include "Debug.h"
#include "DebugSupport.h"
#include "Event.h"
#include "EventQueue.h"
#include "LocalMessageDeliverer.h"
#include "RemoteServer.h"
#include "Request.h"

#include "HashMap.h"
#include "Messenger.h"
#include "RMessageServerDefs.h"

// The default time after which a message that is sent to the remote host is
// auto-replied with RMESSAGE_DELIVERY_TIMED_OUT
static const bigtime_t kDefaultDeliveryTimeout = 30000000; // 30s

// LocalRequest
class RemoteServer::LocalRequest {
public:
								LocalRequest();
	virtual						~LocalRequest();

			void				SetServer(RemoteServer* server);

			bool				IsDone();

	virtual	void				Process() = 0;

protected:
			RemoteServer*		fServer;
			bool				fDone;
};


// UserRequest
class RemoteServer::UserRequest : public RemoteServer::LocalRequest {
public:
								UserRequest(BMessage* requestMessage);
	virtual						~UserRequest();

			void				SetID(int32 id);
			int32				GetID() const;

	virtual	void				Process();
	virtual	void				ProcessReply(BMessage* reply);

protected:
			int32				fID;
			BMessage*			fRequestMessage;
};


// GetMessengerRequest
class RemoteServer::GetMessengerRequest : public RemoteServer::UserRequest {
public:
								GetMessengerRequest(BMessage* requestMessage)
									: UserRequest(requestMessage) {}

	virtual	void				ProcessReply(BMessage* reply);
};


// MessageRequest
class RemoteServer::MessageRequest : public RemoteServer::LocalRequest {
public:
								MessageRequest(BMessage* message,
									BMessenger target, BMessenger replyTarget,
									int32 originalMessageID)
									: LocalRequest(),
									  fMessage(message),
									  fTarget(target),
									  fReplyTarget(replyTarget),
									  fOriginalMessageID(originalMessageID) {}
	virtual						~MessageRequest();

	virtual	void				Process();

private:
			BMessage*			fMessage;
			BMessenger			fTarget;
			BMessenger			fReplyTarget;
			int32				fOriginalMessageID;
};


// ConnectionBrokenRequest
class RemoteServer::ConnectionBrokenRequest : public RemoteServer::LocalRequest {
public:
								ConnectionBrokenRequest(Connection* connection)
									: LocalRequest(),
									  fConnection(connection) {}

	virtual	void				Process();

private:
			Connection*			fConnection;
};


// DeliveryTimeoutEvent
class RemoteServer::DeliveryTimeoutEvent : public Event,
	public RemoteServer::LocalRequest {
public:
	DeliveryTimeoutEvent(RemoteServer* remoteServer, int32 messageID)
		: Event(false),
		  fRemoteServer(remoteServer),
		  fMessageID(messageID)
	{
	}

	~DeliveryTimeoutEvent()
	{
PRINT(("RemoteServer::DeliveryTimeoutEvent::~DeliveryTimeoutEvent()\n"));
	}

	virtual bool Do(EventQueue* queue)
	{
PRINT(("RemoteServer::DeliveryTimeoutEvent::Do()\n"));
		return (fRemoteServer->PushLocalRequest(this) != B_OK);
	}

	virtual	void				Process()
	{
PRINT(("RemoteServer::DeliveryTimeoutEvent::Process()\n"));
		BMessage* message
			= fRemoteServer->fUnrepliedMessages.Remove(fMessageID);
		if (message)
			message->SendReply(RMESSAGE_DELIVERY_TIMED_OUT);
		fDone = true;
PRINT(("RemoteServer::DeliveryTimeoutEvent::Process() done\n"));
	}

private:
	RemoteServer*	fRemoteServer;
	int32			fMessageID;
};


// MessageDeliveryRequest
class RemoteServer::MessageDeliveryRequest : public LocalMessageRequest {
public:
	MessageDeliveryRequest(RemoteServer* remoteServer, BMessenger target,
		BMessenger replyTarget, int32 originalMessageID)
		: LocalMessageRequest(target, NULL, NULL),
		  fRemoteServer(remoteServer),
		  fReplyTarget(replyTarget),
		  fMessageToSend(NULL),
		  fEnvelope(RMESSAGE_DATA),
		  fMessageID(originalMessageID)
	{
	}

	~MessageDeliveryRequest()
	{
		delete fReplyToMessage;
		delete fReplyMessage;
	}

	status_t InitCheck() const
	{
		return (fReplyMessage ? B_OK : B_NO_MEMORY);
	}

	status_t Init(Message* message, BMessage* replyToMessage)
	{
PRINT(("RemoteServer::MessageDeliveryRequest::Init()\n"));
		fReplyToMessage = replyToMessage;
		fReplyMessage = new BMessage;
		if (!fReplyMessage)
			RETURN_ERROR(B_NO_MEMORY);
		// get the message to be sent
		if (message->IsDataMessage()) {
			// a data message: prepare the envelope
			RemoteData* data = message->GetRemoteData();
			status_t error = RemoteMessage::SetData(&fEnvelope,
				data->GetData(), data->GetSize(), message->GetMessage());
			if (error != B_OK)
				RETURN_ERROR(error);
			fMessage = &fEnvelope;
		} else {
			// a standard message
			fMessage = message->GetMessage();
		}
		fMessageToSend.SetTo(message);
PRINT(("RemoteServer::MessageDeliveryRequest::Init() done\n"));
		return B_OK;
	}

	virtual void HandleReply(BMessage* reply)
	{
PRINT(("RemoteServer::MessageDeliveryRequest::HandleReply() for %ld\n",
fMessageID));
		// forward the reply only, if the original message wasn't a B_NO_REPLY
		bool noReply = !_NeedsReply(fMessage) || (fMessageID < 0);
PRINT(("  no reply needed: %d\n", noReply));
		fMessageToSend.Unset();
		if (!noReply && fRemoteServer->Lock()) {
			fRemoteServer->ProxyMessageReceived(fReplyTarget, reply,
				fMessageID);
			fReplyMessage = NULL; // ownership transferred
			fRemoteServer->Unlock();
		}
PRINT(("RemoteServer::MessageDeliveryRequest::HandleReply() done\n"));
	}

	virtual void DeliveryFailed(status_t error)
	{
PRINT(("RemoteServer::MessageDeliveryRequest::DeliveryFailed(%s) for %ld\n",
strerror(error), fMessageID));
		fRemoteServer->_SendMessageRequestFailedReply(fMessage,
			fMessageID, fReplyTarget, error, false);
		fMessageToSend.Unset();
	}

private:
	RemoteServer*		fRemoteServer;
	BMessenger			fReplyTarget;
	MessageReference	fMessageToSend;
	BMessage			fEnvelope;
	int32				fMessageID;
};


// ProxyHandler

class RemoteServer::ProxyHandler : public BHandler {
public:
								ProxyHandler(BMessenger messenger);
	virtual						~ProxyHandler();

	virtual	void				MessageReceived(BMessage* message);
	virtual	BHandler*			ResolveSpecifier(BMessage* message, int32 index,
									BMessage* specifier, int32 form,
									const char* property);

			BMessenger 			GetMessenger() const;

private:
			BMessenger			fMessenger;
};

// constructor
RemoteServer::ProxyHandler::ProxyHandler(BMessenger messenger)
	: fMessenger(messenger)
{
}

// destructor
RemoteServer::ProxyHandler::~ProxyHandler()
{
}

// MessageReceived
void
RemoteServer::ProxyHandler::MessageReceived(BMessage* message)
{
PRINT(("ProxyHandler::MessageReceived()\n"));
	if (RemoteServer* remoteServer = dynamic_cast<RemoteServer*>(Looper())) {
		remoteServer->DetachCurrentMessage();
		remoteServer->ProxyMessageReceived(fMessenger, message, -1);
	} else {
		BHandler::MessageReceived(message);
PRINT(("ProxyHandler::MessageReceived(): no remote server!\n"));
	}
}

// ResolveSpecifier
BHandler*
RemoteServer::ProxyHandler::ResolveSpecifier(BMessage* message, int32 index,
	BMessage* specifier, int32 form, const char* property)
{
	return this;
}

// GetMessenger
BMessenger
RemoteServer::ProxyHandler::GetMessenger() const
{
	return fMessenger;
}


// RemoteServer

// constructor
RemoteServer::RemoteServer(const HostAddress& address,
	ConnectionFactory* connectionFactory,
	LocalMessageDeliverer* messageDeliverer, EventQueue* eventQueue)
	: BLooper((BString("remove server ") += address.ToString()).String()),
	  RequestHandler(),
	  fAddress(address),
	  fConnectionFactory(connectionFactory),
	  fMessageDeliverer(messageDeliverer),
	  fEventQueue(eventQueue),
	  fProxyHandlers(),
	  fUnrepliedRequests(),
	  fUnrepliedMessages(),
	  fReceivingConnections(),
	  fFreeConnections(),
	  fConnectionCount(0),
	  fPendingRequests("pending requests"),
	  fPendingLocalRequests("pending local requests"),
	  fRequestProcessor(-1),
	  fLocalRequestProcessor(-1),
	  fNextRequestID(0),
	  fNextMessageID(0),
	  fTerminating(false)
{
}

// destructor
RemoteServer::~RemoteServer()
{
	Uninit();
}

// Init
status_t
RemoteServer::Init(Connection* receivingConnection,
	Connection* sendingConnection)
{
	if (!sendingConnection)
		return B_BAD_VALUE;
	// pending requests
	status_t error = fPendingRequests.InitCheck();
	if (error != B_OK)
		return error;
	// free connections
	error = fFreeConnections.InitCheck();
	if (error != B_OK)
		return error;
	// spawn request processor thread
	fRequestProcessor = spawn_thread(_RequestProcessorEntry,
		(BString("req. processor ") += fAddress.ToString()).String(),
		B_NORMAL_PRIORITY, this);
	if (fRequestProcessor < 0)
		return fRequestProcessor;
	// spawn local request processor thread
	fLocalRequestProcessor = spawn_thread(_LocalRequestProcessorEntry,
		(BString("l. req. processor ") += fAddress.ToString()).String(),
		B_NORMAL_PRIORITY, this);
	if (fLocalRequestProcessor < 0)
		return fLocalRequestProcessor;
	// resume the threads
	resume_thread(fRequestProcessor);
	resume_thread(fLocalRequestProcessor);
	// add the supplied receiving connection
	if (error == B_OK) {
		if (receivingConnection) {
			error = AttachReceivingConnection(receivingConnection);
			if (error != B_OK)
				return error;
		}
	}
	// add the supplied sending connection
	if (error == B_OK) {
		AttachSendingConnection(sendingConnection);
		Run();
	}
	return error;
}

// Uninit
void
RemoteServer::Uninit()
{
PRINT(("RemoteServer::Uninit()\n"));
	if (fTerminating)
		return;
	fTerminating = true;
	Lock();
	// unblock threads waiting for free connections
	const BList* connections = NULL;
	const BList* receivingConnections = NULL;
	fFreeConnections.Close(false, &connections);
	fReceivingConnections.Close(false, &receivingConnections);
	// unblock the request processors, if waiting for requests, and terminate
	// them
	fPendingRequests.Close(true);
	fPendingLocalRequests.Close(true);
	if (fRequestProcessor >= 0) {
		int32 result;
		wait_for_thread(fRequestProcessor, &result);
	}
	if (fLocalRequestProcessor >= 0) {
		int32 result;
		wait_for_thread(fLocalRequestProcessor, &result);
	}
	// Now the only threads that may execute our code are the connection
	// receiver threads. But all they do is pushing requests onto
	// fPendingRequests, which is closed anyway. That leaves the looper
	// thread, which can only enter some MessageReceived() (ours or that
	// of some proxy handler) and hence would have to Lock() before, but
	// at this point we already have the lock. So, we can safely blow away
	// the connections one after the other, without anyone interfering.
	for (int32 i = 0;
		 Connection* connection = (Connection*)connections->ItemAt(i);
		 i++) {
		delete connection;
	}
 	for (int32 i = 0;
		 Connection* connection = (Connection*)receivingConnections->ItemAt(i);
		 i++) {
		delete connection;
	}
	// now blow away all data that remain
	// unreplied requests
	for (IDRequestMap::Iterator it = fUnrepliedRequests.GetIterator();
		 it.HasNext();) {
		delete it.Next().value;
		it.Remove();
	}
	// pending remote and local requests
	fPendingRequests.Close(true);
	fPendingLocalRequests.Close(true);
	// delivery timeout events
	fDeliveryTimeoutEvents.Lock();
	for (DeliveryTimeoutEventMap::Iterator it
		 = fDeliveryTimeoutEvents.GetIterator();
		it.HasNext();) {
		DeliveryTimeoutEvent* timeoutEvent = it.Next().value;
		if (fEventQueue->RemoveEvent(timeoutEvent))
			delete timeoutEvent;
		it.Remove();
	}
	fDeliveryTimeoutEvents.Unlock();
	// unreplied messages
	fUnrepliedMessages.Lock();
	for (UnrepliedMessageMap::Iterator it = fUnrepliedMessages.GetIterator();
		it.HasNext();) {
		BMessage* message = it.Next().value;
		delete message;
		it.Remove();
	}
	fUnrepliedMessages.Unlock();
	// proxy handlers
	for (MessengerProxyMap::Iterator it = fProxyHandlers.GetIterator();
		it.HasNext();) {
		ProxyHandler* handler = it.Next().value;
		RemoveHandler(handler);
		delete handler;
		it.Remove();
	}
	// and unlock
	Unlock();
PRINT(("RemoteServer::Uninit() done\n"));
}

// AttachReceivingConnection
status_t
RemoteServer::AttachReceivingConnection(Connection* connection)
{
	if (!connection)
		return B_BAD_VALUE;
	status_t error = fReceivingConnections.Push(connection);
	if (error == B_OK)
		connection->SetRequestHandler(this);
	return error;
}

// AttachSendingConnection
void
RemoteServer::AttachSendingConnection(Connection* connection)
{
	if (!connection)
		return;
	atomic_add(&fConnectionCount, 1);
	_ReleaseConnection(connection);
}

// MessageReceived
void
RemoteServer::MessageReceived(BMessage* message)
{
	switch (message->what) {
		case MSG_GET_APP_LIST:
		case MSG_GET_APP_INFO:
		case MSG_GET_MESSENGER:
		{
			DetachCurrentMessage();
			LocalRequest* request = NULL;
			switch (message->what) {
				case MSG_GET_APP_LIST:
				case MSG_GET_APP_INFO:
					request = new(nothrow) UserRequest(message);
					break;
				case MSG_GET_MESSENGER:
					request = new(nothrow) GetMessengerRequest(message);
					break;
			}
			if (request) {
				if (PushLocalRequest(request) != B_OK)
					delete request;
			} else {
				delete message;
				return;
			}			break;
		}
		default:
			BLooper::MessageReceived(message);
			break;
	}
}

// GetEventQueue
EventQueue*
RemoteServer::GetEventQueue() const
{
	return fEventQueue;
}

// Dump
void
RemoteServer::Dump()
{
	PRINT(("DUMP: RemoteServer(%s)\n", fAddress.ToString().String()));
	PRINT(("DUMP: terminating: %d\n", fTerminating));
	PRINT(("DUMP: request processor: %ld\n", fRequestProcessor));
	PRINT(("DUMP: local request processor: %ld\n", fLocalRequestProcessor));
	// connections
	fFreeConnections.Lock();
	PRINT(("DUMP: connections: %ld, free: %ld\n", fConnectionCount,
		fFreeConnections.Size()));
	fFreeConnections.Unlock();
	// pending requests
	fPendingRequests.Lock();
	PRINT(("DUMP: requests: %ld\n", fPendingRequests.Size()));
	fPendingRequests.Unlock();
	// pending local requests
	fPendingLocalRequests.Lock();
	PRINT(("DUMP: local requests: %ld\n", fPendingLocalRequests.Size()));
	fPendingLocalRequests.Unlock();
	// proxy handlers
	fProxyHandlers.Lock();
	PRINT(("DUMP: proxy handlers: %ld\n", fProxyHandlers.Size()));
	fProxyHandlers.Unlock();
	// unreplied requests
	fUnrepliedRequests.Lock();
	PRINT(("DUMP: unreplied requests: %ld\n", fUnrepliedRequests.Size()));
	fUnrepliedRequests.Unlock();
	// unreplied messages
	fUnrepliedMessages.Lock();
	PRINT(("DUMP: unreplied messages: %ld\n", fUnrepliedMessages.Size()));
	fUnrepliedMessages.Unlock();
}

// PushLocalRequest
status_t
RemoteServer::PushLocalRequest(LocalRequest* request)
{
	if (!request)
		return B_BAD_VALUE;
	request->SetServer(this);
	return fPendingLocalRequests.Push(request);
}

// ProxyMessageReceived
void
RemoteServer::ProxyMessageReceived(BMessenger messenger, BMessage* message,
	int32 originalMessageID)
{
	// takes over message ownership
PRINT(("RemoteServer::ProxyMessageReceived()\n"));
D(
	char target[128];
	char replyTarget[128];
	messenger_to_string(messenger, target);
	messenger_to_string(message->ReturnAddress(), replyTarget);
	PRINT(("  target: %s, reply target: %s\n", target, replyTarget));
)
	// create a local request
	LocalRequest* localRequest = new(nothrow) MessageRequest(
		message, messenger, message->ReturnAddress(), originalMessageID);
	if (!localRequest) {
PRINT(("  failed to allocate local request\n"));
		_SendDeliveryFailedReply(message, B_NO_MEMORY);
		delete message;
		return;
	}
	// push the request
	status_t error = PushLocalRequest(localRequest);
	if (error != B_OK) {
PRINT(("  failed to push local request\n"));
		_SendDeliveryFailedReply(message, error);
		delete localRequest;
	}
PRINT(("RemoteServer::ProxyMessageReceived() done\n"));
}

// HandleRequest
void
RemoteServer::HandleRequest(Connection* connection, Request* request)
{
	FUNCTION(("request: %lx\n", request->GetRequestMessage()->what));
	RequestReference requestReference(request);
	if (fPendingRequests.Push(request) == B_OK) {
		request->AddReference();
	} else {
		// error
		FATAL(("Failed to enqueue request!\n"));
	}
}

// ErrorOccurred
bool
RemoteServer::ErrorOccurred(Connection* connection, status_t error)
{
	LocalRequest* request = new ConnectionBrokenRequest(connection);
	if (request) {
		if (PushLocalRequest(request) != B_OK)
			delete request;
	}
	return false;
}

// SendRequest
status_t
RemoteServer::SendRequest(BMessage* request, Message* parameter)
{
	Connection* connection = NULL;
	status_t error = _GetFreeConnection(&connection);
	if (error != B_OK)
		return error;
	error = connection->SendRequest(request, parameter);
	_ReleaseConnection(connection);
	return error;
}

// SendMessage
status_t
RemoteServer::SendMessage(Message* message, BMessenger target,
	BMessenger replyMessenger, int32 messageID, int32 originalMessageID)
{
	BMessage request(MSG_REQUEST_MESSAGE);
	status_t error = request.AddMessenger("target", target);
	if (error != B_OK)
		return error;
	error = request.AddMessenger("reply target", replyMessenger);
	if (error != B_OK)
		return error;
	error = request.AddInt32("id", messageID);
	if (error != B_OK)
		return error;
	error = request.AddInt32("original id", originalMessageID);
	if (error != B_OK)
		return error;
	return SendRequest(&request, message);
}

// _HandleGetAppListRequest
void
RemoteServer::_HandleGetAppListRequest(Request* request)
{
PRINT(("RemoteServer::_HandleGetAppListRequest()\n"));
	// get parameters
	BMessage* requestMessage = request->GetRequestMessage();
	const char* signature = NULL;
	if (requestMessage->FindString("signature", &signature) != B_OK)
		signature = NULL;
	// get app list
	BList teams;
	if (signature)
		be_roster->GetAppList(signature, &teams);
	else
		be_roster->GetAppList(&teams);
	// prepare reply
	status_t error = B_OK;
	BMessage reply(MSG_REQUEST_REPLY);
	int32 id;
	if (requestMessage->FindInt32("request id", &id) == B_OK)
		reply.AddInt32("request id", id);
	int32 count = teams.CountItems();
	if (error == B_OK) {
		for (int32 i = 0; i < count; i++)
			reply.AddInt32("teams", (int32)teams.ItemAt(i));
	} else
		reply.AddInt32("error", error);
	// send reply
	SendRequest(&reply);
PRINT(("RemoteServer::_HandleGetAppListRequest() done\n"));
}

// _HandleGetAppInfoRequest
void
RemoteServer::_HandleGetAppInfoRequest(Request* request)
{
PRINT(("RemoteServer::_HandleGetAppInfoRequest()\n"));
	// get parameters and the respective app info
	BMessage* requestMessage = request->GetRequestMessage();
	status_t error = B_OK;
	app_info info;
	const char* signature = NULL;
	team_id team;
	if (requestMessage->FindString("signature", &signature) == B_OK)
		error = be_roster->GetAppInfo(signature, &info);
	else if (requestMessage->FindInt32("team", &team) == B_OK)
		error = be_roster->GetRunningAppInfo(team, &info);
	else
		error = be_roster->GetActiveAppInfo(&info);
	// prepare reply
	BMessage reply(MSG_REQUEST_REPLY);
	int32 id;
	if (requestMessage->FindInt32("request id", &id) == B_OK)
		reply.AddInt32("request id", id);
	if (error == B_OK) {
		reply.AddInt32("thread", info.thread);
		reply.AddInt32("team", info.team);
		reply.AddInt32("port", info.port);
		reply.AddInt32("flags", (int32)info.flags);
		reply.AddRef("ref", &info.ref);
		reply.AddString("signature", info.signature);
	} else
		reply.AddInt32("error", error);
	// send reply
	SendRequest(&reply);
PRINT(("RemoteServer::_HandleGetAppInfoRequest() done\n"));
}

// _HandleGetMessengerRequest
void
RemoteServer::_HandleGetMessengerRequest(Request* request)
{
PRINT(("RemoteServer::_HandleGetMessengerRequest()\n"));
	// get parameters and the respective app info
	BMessage* requestMessage = request->GetRequestMessage();
	app_info info;
	const char* signature = NULL;
	team_id team = -1;
	if (requestMessage->FindString("signature", &signature) != B_OK)
		signature = NULL;
	if (requestMessage->FindInt32("team", &team) != B_OK)
		team = -1;
	// get messenger
	status_t error = B_OK;
	BMessenger messenger(signature, team, &error);
	bool orName;
	if (error != B_OK && requestMessage->FindBool("or name", &orName) == B_OK
		&& orName) {
		// Initializing the messenger failed, but we are asked also to check
		// for matching file names. So, traverse the list of running apps...
		BList teams;
		be_roster->GetAppList(&teams);
		int32 count = teams.CountItems();
		for (int32 i = 0; i < count; i++) {
			team_id team = (team_id)teams.ItemAt(i);
			app_info info;
			if (be_roster->GetRunningAppInfo(team, &info) == B_OK
				&& !strcmp(info.ref.name, signature)) {
				// found a match
				messenger = BMessenger(NULL, team, &error);
				break;
			}
		}
	}
	// prepare reply
	BMessage reply(MSG_REQUEST_REPLY);
	int32 id;
	if (requestMessage->FindInt32("request id", &id) == B_OK)
		reply.AddInt32("request id", id);
	if (error == B_OK)
		reply.AddMessenger("messenger", messenger);
	else
		reply.AddInt32("error", error);
	// send reply
	SendRequest(&reply);
PRINT(("RemoteServer::_HandleGetMessengerRequest() done\n"));
}

// _HandleRequestReply
void
RemoteServer::_HandleRequestReply(Request* request)
{
PRINT(("RemoteServer::_HandleRequestReply()\n"));
	// get parameters
	BMessage* requestMessage = request->GetRequestMessage();
	// get request id
	int32 id;
	if (requestMessage->FindInt32("request id", &id) != B_OK) {
PRINT(("RemoteServer::_HandleRequestReply() done 1\n"));
		return;
	}
	// get original request, this is the reply for
	UserRequest* originalRequest = fUnrepliedRequests.Remove(id);
	if (!originalRequest) {
PRINT(("RemoteServer::_HandleRequestReply() done 2\n"));
		return;
	}
	originalRequest->ProcessReply(requestMessage);
	if (originalRequest->IsDone())
		delete originalRequest;
PRINT(("RemoteServer::_HandleRequestReply() done\n"));
}

// _HandleMessageRequest
void
RemoteServer::_HandleMessageRequest(Request* request)
{
PRINT(("RemoteServer::_HandleMessageRequest()\n"));
	BMessage* requestMessage = request->GetRequestMessage();
	BMessenger target;
	BMessenger replyTarget;
	// get message ID
	int32 messageID;
	if (requestMessage->FindInt32("id", &messageID) != B_OK)
		messageID = -1;
PRINT(("  message ID: %ld\n", messageID));
	// get reply target
	if (requestMessage->FindMessenger("reply target", &replyTarget) != B_OK) {
		// no reply target: we can't even send a RMESSAGE_DELIVERY_FAILED
		// message
		PRINT(("  no reply target\n"));
		return;
	}
	// get the message
	Message* parameter = request->GetParameter();
	if (!parameter || parameter->IsEmpty()) {
PRINT(("RemoteServer::_HandleMessageRequest() done 2\n"));
		_SendMessageRequestFailedReply(NULL, messageID, replyTarget, B_ERROR);
		return;
	}
	BMessage* message = parameter->GetMessage();
	// get target
	if (requestMessage->FindMessenger("target", &target) != B_OK) {
		PRINT(("  no target\n"));
		_SendMessageRequestFailedReply(message, messageID, replyTarget,
			B_ERROR);
		return;
	}
char targetString[128];
TOUCH(targetString);
PRINT(("  target: %s\n", messenger_to_string(target, targetString)));
PRINT(("  reply target: %s\n", messenger_to_string(replyTarget, targetString)));
	// target valid?
	if (!target.IsValid()) {
		PRINT(("  invalid target\n"));
		_SendMessageRequestFailedReply(message, messageID, replyTarget,
			B_BAD_TEAM_ID);
		return;
	}
	// prepare the message
	// translate the messengers inside the message
	if (message)
		_TranslateRemoteTargets(message);
	// prepare a delivery event
	MessageDeliveryRequest* deliveryRequest
		= new(nothrow) MessageDeliveryRequest(this, target, replyTarget,
			messageID);
	if (!deliveryRequest) {
PRINT(("RemoteServer::_HandleMessageRequest() done 3\n"));
		_SendMessageRequestFailedReply(message, messageID, replyTarget,
			B_NO_MEMORY);
		return;
	}
	ObjectDeleter<MessageDeliveryRequest> deleter(deliveryRequest);
	// get original message ID
	int32 originalMessageID;
	if (requestMessage->FindInt32("original id", &originalMessageID) != B_OK)
		originalMessageID = -1;
PRINT(("  original message ID: %ld\n", originalMessageID));
	// get the original message, if any
	status_t error;
	BMessage* originalMessage = fUnrepliedMessages.Remove(originalMessageID);
	if (originalMessage) {
		// got the message, there should be a timeout event, too
		DeliveryTimeoutEvent* timeoutEvent = fDeliveryTimeoutEvents.Remove(
			originalMessageID);
		if (timeoutEvent) {
			// ... yep, remove and delete it
			if (fEventQueue->RemoveEvent(timeoutEvent))
				delete timeoutEvent;
		}
		error = deliveryRequest->Init(parameter, originalMessage);
	} else
		error = deliveryRequest->Init(parameter, NULL);
	if (error != B_OK) {
PRINT(("RemoteServer::_HandleMessageRequest() done 4\n"));
		_SendMessageRequestFailedReply(message, messageID, replyTarget,
			error);
		return;
	}
	// push the delivery request
	if (fMessageDeliverer->PushRequest(deliveryRequest) != B_OK) {
PRINT(("RemoteServer::_HandleMessageRequest() done 5\n"));
		_SendMessageRequestFailedReply(message, messageID, replyTarget,
			B_NO_MEMORY);
		return;
	}
	deleter.Detach();
PRINT(("RemoteServer::_HandleMessageRequest() done\n"));
}

// _HandleMessageLocalRequest
void
RemoteServer::_HandleMessageLocalRequest(BMessage* message, BMessenger target,
	BMessenger replyTarget, int32 originalMessageID)
{
PRINT(("RemoteServer::_HandleMessageLocalRequest()\n"));
D(
	char targetString[128];
	char replyTargetString[128];
	messenger_to_string(target, targetString);
	messenger_to_string(replyTarget, replyTargetString);
	PRINT(("  target: %s, reply target: %s\n", targetString,
		replyTargetString));
)
	ObjectDeleter<BMessage> deleter(message);
	status_t error = B_OK;
	Message msg;
	if (message->what == RMESSAGE_DATA) {
		// a data message: extract the data
		// allocate parameters message
		BMessage* parameters = new BMessage;
		if (!parameters) {
			_SendDeliveryFailedReply(message, B_NO_MEMORY);
			return;
		}
		ObjectDeleter<BMessage> parametersDeleter(parameters);
		// extract data and parameters from data message
		const void* data;
		size_t size;
		bool hasParameters;
		error = RemoteMessage::GetData(message, &data, &size,
			parameters, &hasParameters);
		if (error != B_OK) {
			_SendDeliveryFailedReply(message, error);
			return;
		}
		// allocate remote data
		RemoteData* remoteData
			= new(nothrow) RemoteData((uint8*)data, size);
		if (!remoteData) {
			RemoteMessage::ReleaseData(data, size);
			_SendDeliveryFailedReply(message, B_NO_MEMORY);
			return;
		}
		DataReference dataReference(remoteData, false);
		// set up message
		if (hasParameters)
			parametersDeleter.Detach();
		else
			parameters = NULL;
		msg.SetTo(parameters, remoteData);
	} else {
		// a normal message
		msg.SetTo(message);
	}
	// msg owns the message now
	deleter.Detach();
	// if the message needs a reply, add it to the unreplied messages
	bool needsReply = _NeedsReply(message);
	int32 messageID = -1;
	if (needsReply) {
		messageID = _GetNextMessageID();
		error = fUnrepliedMessages.Put(messageID, message);
		if (error != B_OK) {
			_SendDeliveryFailedReply(message, error);
			return;
		}
	}
	// send the message
PRINT(("  sending message (id: %ld) to remote:\n", messageID));
D(message->PrintToStream());
	error = SendMessage(&msg, target, replyTarget, messageID,
		originalMessageID);
	if (error == B_OK) {
		if (needsReply) {
			// sending went fine: schedule a timeout event
			// if that fails, there's little we can do
PRINT(("  scheduling timeout event...\n"));
			DeliveryTimeoutEvent* timeoutEvent
				= new(nothrow) DeliveryTimeoutEvent(this, messageID);
			if (timeoutEvent) {
				if (fDeliveryTimeoutEvents.Put(messageID, timeoutEvent)
					== B_OK) {
					timeoutEvent->SetTime(system_time()
						+ kDefaultDeliveryTimeout);
					if (!fEventQueue->AddEvent(timeoutEvent)) {
PRINT(("  failed to add timeout event to event queue\n"));
						if (fDeliveryTimeoutEvents.Remove(messageID))
							delete timeoutEvent;
					}
				} else 
					delete timeoutEvent;
			}
			// don't delete the message as long as in is in the unreplied
			// messages map
			msg.DetachMessage();
			// strip the message off all data to save memory
			fUnrepliedMessages.Lock();
			if (fUnrepliedMessages.Get(messageID) == message)
				message->MakeEmpty();
			fUnrepliedMessages.Unlock();
		}
	} else {
PRINT(("  sending the message failed: %s\n", strerror(error)));
		// sending failed: remove the message from the unreplied messages
		if (needsReply) {
			if (!fUnrepliedMessages.Remove(messageID))
				msg.DetachMessage();	// can't possibly happen
			_SendDeliveryFailedReply(message, error);
		}
	}
	// If something went wrong or the message doesn't need a reply,
	// msg will still own it, so that it is deleted here.
PRINT(("RemoteServer::_HandleMessageLocalRequest() done\n"));
}

// _TranslateRemoteTarget
status_t
RemoteServer::_TranslateRemoteTarget(BMessenger target,
	BMessenger* translatedTarget, ProxyHandler** _handler)
{
	fProxyHandlers.Lock();
	ProxyHandler* handler = fProxyHandlers.Get(target);
	if (!handler) {
		// create a new handler
		handler = new(nothrow) ProxyHandler(target);
		if (!handler)
			RETURN_ERROR(B_NO_MEMORY);
		// put it into the map
		status_t error = fProxyHandlers.Put(target, handler);
		if (error != B_OK) {
			delete handler;
			RETURN_ERROR(error);
		}
		// add the handler to the looper
		Lock();
		AddHandler(handler);
		Unlock();
	}
	if (_handler)
		*_handler = handler;
	status_t error = B_OK;
	if (translatedTarget)
		*translatedTarget = BMessenger(handler, NULL, &error);
	fProxyHandlers.Unlock();
D(
if (error == B_OK) {
	char originalString[128];
	char translatedString[128];
	messenger_to_string(target, originalString);
	messenger_to_string(BMessenger(handler, NULL, &error), translatedString);
	PRINT(("RemoteServer::_TranslateRemoteTarget(): %s -> %s\n",
		originalString, translatedString));
}
)
	RETURN_ERROR(error);
}

// _TranslateRemoteTargets
status_t
RemoteServer::_TranslateRemoteTargets(BMessage* message)
{
	if (!message)
		return B_BAD_VALUE;
	char* field;
//	const char* field;	// Zeta
	uint32 type;
	int32 count;
	for (int32 i = 0;
		 message->GetInfo(B_MESSENGER_TYPE, i, &field, &type, &count) == B_OK;
		 i++) {
		for (int32 k = 0; k < count; k++) {
			// get messenger
			BMessenger target;
			status_t error = message->FindMessenger(field, k, &target);
			if (error != B_OK) {
FATAL(("Failed to get messenger: field: %s, index: %ld\n", field, k));
				return error;
			}
			// translate messenger
			BMessenger translatedTarget;
			error = _TranslateRemoteTarget(target, &translatedTarget);
			if (error != B_OK) {
FATAL(("Failed to translate messenger: field: %s, index: %ld\n", field, k));
				return error;
			}
			// replace the original messenger
			error = message->ReplaceMessenger(field, k, translatedTarget);
			if (error != B_OK) {
FATAL(("Failed to replace messenger: field: %s, index: %ld\n", field, k));
				return error;
			}
		}
	}
	return B_OK;
}

// _SendUserRequestMessage
status_t
RemoteServer::_SendUserRequestMessage(UserRequest* request,
	BMessage* requestMessage)
{
	// get a fresh ID for the request
	int32 id = _GetNextRequestID();
	request->SetID(id);
	status_t error = requestMessage->AddInt32("request id", id);
	if (error != B_OK)
		return error;
	// remember the request as being unreplied
	error = fUnrepliedRequests.Put(id, request);
	if (error == B_OK) {
		// send the request
		error = SendRequest(requestMessage);
		// on error, remove the request from the unreplied request
		// map
		if (error != B_OK)
			fUnrepliedRequests.Remove(id);
	}
	return error;
}

// _SendDeliveryFailedReply
void
RemoteServer::_SendDeliveryFailedReply(BMessage* message,
	status_t deliveryError)
{
	// check parameters
	if (!message) {
		PRINT(("RemoteServer::_SendDeliveryFailedReply(): NULL message!\n"));
		return;
	}
	if (deliveryError == B_OK) {
		PRINT(("RemoteServer::_SendDeliveryFailedReply(): error is B_OK? "
			"Setting to B_ERROR.\n"));
		deliveryError = B_ERROR;
	}
	// prepare and send the reply
	BMessage reply(RMESSAGE_DELIVERY_FAILED);
	reply.AddInt32("error", deliveryError);
	status_t error = message->SendReply(&reply);
	if (error != B_OK) {
		PRINT(("RemoteServer::_SendDeliveryFailedReply(): sending reply "
			"failed: %s\n", strerror(error)));
	}
}

// _SendMessageRequestFailedReply
void
RemoteServer::_SendMessageRequestFailedReply(BMessage* originalMessage,
	int32 messageID, BMessenger replyTarget, status_t deliveryError,
	bool synchronous)
{
	// check parameters
	if (replyTarget.Team() < 0) {
		PRINT(("RemoteServer::_SendMessageRequestFailedReply(): reply target "
			"is invalid.\n"));
		return;
	}
	if (messageID < 0) {
		PRINT(("RemoteServer::_SendMessageRequestFailedReply(): invalid "
			"message ID -- no reply needed.\n"));
		return;
	}
	if (originalMessage && !_NeedsReply(originalMessage)) {
		// don't send an error message as a reply to a B_NO_REPLY or an
		// error message
		char buffer[10];
		TOUCH(buffer);
		PRINT(("RemoteServer::_SendMessageRequestFailedReply(): message "
			"doesn't need a reply (%s).\n",
			message_what(originalMessage, buffer)));
		return;
	}
	if (deliveryError == B_OK) {
		PRINT(("RemoteServer::_SendMessageRequestFailedReply(): error is B_OK? "
			"Setting to B_ERROR.\n"));
		deliveryError = B_ERROR;
	}
	BMessage* reply = new BMessage(RMESSAGE_DELIVERY_FAILED);
	if (!reply) {
		PRINT(("RemoteServer::_SendMessageRequestFailedReply(): failed to "
			"allocate reply message.\n"));
		return;
	}
	reply->AddInt32("error", deliveryError);
	if (synchronous) {
		_HandleMessageLocalRequest(reply, replyTarget, BMessenger(),
			messageID);
	} else
		ProxyMessageReceived(replyTarget, reply, messageID);
}

// _NeedsReply
bool
RemoteServer::_NeedsReply(BMessage* message)
{
	if (!message)
		return false;
	switch (message->what) {
		case B_NO_REPLY:
		case RMESSAGE_DELIVERY_FAILED:
		case RMESSAGE_DELIVERY_TIMED_OUT:
			return false;
		default:
			return true;
	}
}

// _GetFreeConnection
status_t
RemoteServer::_GetFreeConnection(Connection** _connection)
{
	// check the connection count
	status_t error = B_OK;
	Connection* connection = NULL;
	if (fConnectionCount == 0) {
		// no connections yet: create one
		error = _CreateConnection(&connection);
		if (error == B_OK)
			atomic_add(&fConnectionCount, 1);
	} else {
		// there are connections: get one
		error = fFreeConnections.Pop(&connection);
	}
	*_connection = connection;
	return error;
}

// _ReleaseConnection
void
RemoteServer::_ReleaseConnection(Connection* connection)
{
	if (!connection->IsValid() || fFreeConnections.Push(connection) != B_OK)
		_DeleteConnection(connection);
}

// _CreateConnection
status_t
RemoteServer::_CreateConnection(Connection** connection)
{
	status_t error = fConnectionFactory->CreateConnection(fAddress, connection);
//	if (error == B_OK)
//		(*connection)->SetRequestHandler(this);
	return error;
}

// _DeleteConnection
void
RemoteServer::_DeleteConnection(Connection* connection)
{
	delete connection;
	atomic_add(&fConnectionCount, -1);
	// TODO: If connection count == 0, notify the RMessageServer?
}

// _GetNextRequestID
int32
RemoteServer::_GetNextRequestID()
{
	return atomic_add(&fNextRequestID, 1);
}

// _GetNextMessageID
int32
RemoteServer::_GetNextMessageID()
{
	return atomic_add(&fNextMessageID, 1);
}

// _RequestProcessorEntry
int32
RemoteServer::_RequestProcessorEntry(void* data)
{
	return ((RemoteServer*)data)->_RequestProcessor();
}

// _RequestProcessor
int32
RemoteServer::_RequestProcessor()
{
	while (!fTerminating) {
		Request* request = NULL;
		status_t error = fPendingRequests.Pop(&request);
		if (error != B_OK || !request)
			continue;
		// handle request
PRINT(("got request: %lx\n", request->GetRequestMessage()->what));
		switch (request->GetRequestMessage()->what) {
			case MSG_GET_APP_LIST:
				_HandleGetAppListRequest(request);
				break;
			case MSG_GET_APP_INFO:
				_HandleGetAppInfoRequest(request);
				break;
			case MSG_GET_MESSENGER:
				_HandleGetMessengerRequest(request);
				break;
			case MSG_REQUEST_REPLY:
				_HandleRequestReply(request);
				break;
			case MSG_REQUEST_MESSAGE:
				_HandleMessageRequest(request);
				break;
			default:
				break;
		}
		request->RemoveReference();
	}
	return B_OK;
}

// _LocalRequestProcessorEntry
int32
RemoteServer::_LocalRequestProcessorEntry(void* data)
{
	return ((RemoteServer*)data)->_LocalRequestProcessor();
}

// _LocalRequestProcessor
int32
RemoteServer::_LocalRequestProcessor()
{
	while (!fTerminating) {
		// fetch request
		LocalRequest* request = NULL;
		status_t error = fPendingLocalRequests.Pop(&request);
		if (error != B_OK || !request)
			continue;
		// process request
FUNCTION(("got local request: %p\n", request));
		request->Process();
		if (request->IsDone())
			delete request;
	}
	return B_OK;
}


// LocalRequest

// constructor
RemoteServer::LocalRequest::LocalRequest()
	: fServer(NULL),
	  fDone(false)
{
}

// destructor
RemoteServer::LocalRequest::~LocalRequest()
{
PRINT(("%p->RemoteServer::LocalRequest::~LocalRequest()\n", this));
}

// SetServer
void
RemoteServer::LocalRequest::SetServer(RemoteServer* server)
{
	fServer = server;
}

// IsDone
bool
RemoteServer::LocalRequest::IsDone()
{
	return fDone;
}


// UserRequest

// constructor
RemoteServer::UserRequest::UserRequest(BMessage* requestMessage)
	: LocalRequest(),
	  fID(-1),
	  fRequestMessage(requestMessage)
{
}

// destructor
RemoteServer::UserRequest::~UserRequest()
{
	delete fRequestMessage;
}

// SetID
void
RemoteServer::UserRequest::SetID(int32 id)
{
	fID = id;
}

// GetID
int32
RemoteServer::UserRequest::GetID() const
{
	return fID;
}

// Process
void
RemoteServer::UserRequest::Process()
{
	status_t error = fServer->_SendUserRequestMessage(this, fRequestMessage);
	fDone = (error != B_OK);
}

// ProcessReply
void
RemoteServer::UserRequest::ProcessReply(BMessage* reply)
{
	reply->what = MSG_REPLY;
	fRequestMessage->SendReply(reply, (BHandler*)NULL, 0LL);
	fDone = true;
}


// GetMessengerRequest

// ProcessReply
void
RemoteServer::GetMessengerRequest::ProcessReply(BMessage* reply)
{
	// translate the returned messenger
	BMessenger messenger;
	if (reply->FindMessenger("messenger", &messenger) == B_OK) {
		BMessenger translatedMessenger;
		status_t error
			= fServer->_TranslateRemoteTarget(messenger, &translatedMessenger);
		if (error == B_OK)
			error = reply->ReplaceMessenger("messenger", translatedMessenger);
		if (error == B_OK)
			UserRequest::ProcessReply(reply);
	} else {
		PRINT(("GetMessengerRequest::ProcessReply(): No messenger found in "
			"reply.\n"));
		D(reply->PrintToStream());
	}
	fDone = true;
}


// ConnectionBrokenRequest

// Process
void
RemoteServer::ConnectionBrokenRequest::Process()
{
	if (fServer->fFreeConnections.Remove(fConnection) == B_OK)
		fServer->_DeleteConnection(fConnection);
	fDone = true;
}


// MessageRequest

// destructor
RemoteServer::MessageRequest::~MessageRequest()
{
	if (fMessage) {
		_SendDeliveryFailedReply(fMessage, B_NO_MEMORY);
		delete fMessage;
	}
}

// Process
void
RemoteServer::MessageRequest::Process()
{
	fServer->_HandleMessageLocalRequest(fMessage, fTarget, fReplyTarget,
		fOriginalMessageID);
	fMessage = NULL;
	fDone = true;
}

