// RemoteServer.h
// 
// Copyright (c) 2004, Ingo Weinhold (bonefish@cs.tu-berlin.de)
// 
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation 
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
// THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
// DEALINGS IN THE SOFTWARE.
// 
// Except as contained in this notice, the name of a copyright holder shall
// not be used in advertising or otherwise to promote the sale, use or other
// dealings in this Software without prior written authorization of the
// copyright holder.

#ifndef RMESSAGE_REMOTE_SERVER_H
#define RMESSAGE_REMOTE_SERVER_H

#include <List.h>
#include <Locker.h>
#include <Looper.h>
#include <OS.h>

#include "BlockingQueue.h"
#include "HashMap.h"
#include "HostAddress.h"
#include "Messenger.h"
#include "RequestHandler.h"

class ConnectionFactory;
class EventQueue;
class LocalMessageDeliverer;
class Message;

class RemoteServer : public BLooper, private RequestHandler {
public:
								RemoteServer(const HostAddress& address,
									ConnectionFactory* connectionFactory,
									LocalMessageDeliverer* messageDeliverer,
									EventQueue* eventQueue);
								~RemoteServer();

			status_t			Init(Connection* receivingConnection,
									Connection* sendingConnection);
			void				Uninit();

			status_t			AttachReceivingConnection(
									Connection* connection);
			void				AttachSendingConnection(Connection* connection);

			void				MessageReceived(BMessage* message);

			EventQueue*			GetEventQueue() const;

			void				Dump();

	// private

			class LocalRequest;
			class UserRequest;
			class GetMessengerRequest;
			class MessageDeliveryRequest;
			class MessageRequest;
			class ConnectionBrokenRequest;
			class DeliveryTimeoutEvent;
			class ProxyHandler;

			friend class LocalRequest;
			friend class UserRequest;
			friend class GetMessengerRequest;
			friend class MessageDeliveryRequest;
			friend class MessageRequest;
			friend class ConnectionBrokenRequest;
			friend class DeliveryTimeoutEvent;
			friend class ProxyHandler;

			status_t			PushLocalRequest(LocalRequest* request);

			void				ProxyMessageReceived(BMessenger messenger,
									BMessage* message, int32 originalMessageID);

private:
	virtual	void				HandleRequest(Connection* connection,
									Request* request);
	virtual	bool				ErrorOccurred(Connection* connection,
									status_t error);

			status_t			SendRequest(BMessage* request,
									Message* parameter = NULL);
			status_t			SendMessage(Message* message,
									BMessenger target,
									BMessenger replyMessenger,
									int32 messageID, int32 originalMessageID);

			void				_HandleGetAppListRequest(Request* request);
			void				_HandleGetAppInfoRequest(Request* request);
			void				_HandleGetMessengerRequest(Request* request);
			void				_HandleRequestReply(Request* request);
			void				_HandleMessageRequest(Request* request);
			void				_HandleMessageLocalRequest(BMessage* message,
									BMessenger target, BMessenger replyTarget,
									int32 originalMessageID);
			void				_HandleScriptingRequest(BMessage* request);

			status_t			_TranslateRemoteTarget(BMessenger target,
									BMessenger* translatedTarget,
									ProxyHandler** handler = NULL);
			status_t			_TranslateRemoteTargets(BMessage* message);

			status_t			_SendUserRequestMessage(UserRequest* request,
									BMessage* requestMessage);

	static	void				_SendDeliveryFailedReply(BMessage* message,
									status_t error);
			void				_SendMessageRequestFailedReply(
									BMessage* originalMessage, int32 messageID,
									BMessenger replyTarget, status_t error,
									bool synchronous = true);

	static	bool				_NeedsReply(BMessage* message);

			status_t			_GetFreeConnection(Connection** connection);
			void				_ReleaseConnection(Connection* connection);
			status_t			_CreateConnection(Connection** connection);
			void				_DeleteConnection(Connection* connection);

			int32				_GetNextRequestID();
			int32				_GetNextMessageID();

	static	int32				_RequestProcessorEntry(void* data);
			int32				_RequestProcessor();

	static	int32				_LocalRequestProcessorEntry(void* data);
			int32				_LocalRequestProcessor();

private:
			typedef BlockingQueue<Request> RequestQueue;
			typedef BlockingQueue<LocalRequest> LocalRequestQueue;
			typedef BlockingQueue<Connection> ConnectionQueue;
			typedef SynchronizedHashMap<HashKey32<int32>, UserRequest>
				IDRequestMap;
			typedef SynchronizedHashMap<HashKey32<int32>, BMessage>
				UnrepliedMessageMap;
			typedef SynchronizedHashMap<Messenger, ProxyHandler>
				MessengerProxyMap;
			typedef SynchronizedHashMap<HashKey32<int32>, DeliveryTimeoutEvent>
				DeliveryTimeoutEventMap;

			HostAddress			fAddress;
			ConnectionFactory*	fConnectionFactory;
			LocalMessageDeliverer* fMessageDeliverer;
			EventQueue*			fEventQueue;
			MessengerProxyMap	fProxyHandlers;
			IDRequestMap		fUnrepliedRequests;
			UnrepliedMessageMap	fUnrepliedMessages;
			DeliveryTimeoutEventMap	fDeliveryTimeoutEvents;
			ConnectionQueue		fReceivingConnections;
			ConnectionQueue		fFreeConnections;
			vint32				fConnectionCount;
			RequestQueue		fPendingRequests;
			LocalRequestQueue	fPendingLocalRequests;
			thread_id			fRequestProcessor;
			thread_id			fLocalRequestProcessor;
			vint32				fNextRequestID;
			vint32				fNextMessageID;
			bool				fTerminating;
};

#endif	// RMESSAGE_REMOTE_SERVER_H
