// Request.cpp
// 
// Copyright (c) 2004, Ingo Weinhold (bonefish@cs.tu-berlin.de)
// 
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation 
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
// THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
// DEALINGS IN THE SOFTWARE.
// 
// Except as contained in this notice, the name of a copyright holder shall
// not be used in advertising or otherwise to promote the sale, use or other
// dealings in this Software without prior written authorization of the
// copyright holder.

#include <Message.h>
#include <NetBuffer.h>

#include "Compatibility.h"
#include "Debug.h"
#include "Message.h"
#include "Request.h"
#include "RMessageServerDefs.h"

// sanity limits
static const int32 kMaxSaneRequestSize			= 1024;
static const int32 kMaxSaneMessageParameterSize	= 1024 * 1024;
static const int32 kMaxSaneDataParameterSize	= 10 * 1024 * 1024;

static const uint32 kRequestHeaderMagic	= 0x473e2951UL;

static vint32 sNextRequestHeaderID = 0;

// RequestHeader

// constructor
RequestHeader::RequestHeader()
	: requestSize(0),
	  parameterType(REQUEST_NO_PARAMETER),
	  messageSize(0),
	  dataSize(0),
	  id(atomic_add(&sNextRequestHeaderID, 1))
{
}

// GetOverallSize
int32
RequestHeader::GetOverallSize() const
{
	return GetHeaderSize() + requestSize + messageSize + dataSize;
}

// Write
status_t
RequestHeader::Write(BNetBuffer& buffer) const
{
	status_t error = buffer.AppendUint32(kRequestHeaderMagic);
	if (error != B_OK)
		return error;
	error = buffer.AppendInt32(requestSize);
	if (error != B_OK)
		return error;
	error = buffer.AppendUint32(parameterType);
	if (error != B_OK)
		return error;
	error = buffer.AppendInt32(messageSize);
	if (error != B_OK)
		return error;
	error = buffer.AppendInt32(dataSize);
	if (error != B_OK)
		return error;
PRINT(("RequestHeader::Write(): id: %ld\n", id));
	error = buffer.AppendInt32(id);
//	if (error != B_OK)
		return error;
}

// Read
status_t
RequestHeader::Read(BNetBuffer& buffer)
{
	if ((int32)buffer.Size() < GetHeaderSize())
		RETURN_ERROR(B_BAD_VALUE);
	// get the data
	uint32 magic;
	status_t error = buffer.RemoveUint32(magic);
	if (error != B_OK)
		RETURN_ERROR(error);
	if (magic != kRequestHeaderMagic)
		RETURN_ERROR(B_BAD_DATA);
	error = buffer.RemoveInt32(requestSize);
	if (error != B_OK)
		RETURN_ERROR(error);
	error = buffer.RemoveUint32(parameterType);
	if (error != B_OK)
		RETURN_ERROR(error);
	error = buffer.RemoveInt32(messageSize);
	if (error != B_OK)
		RETURN_ERROR(error);
	error = buffer.RemoveInt32(dataSize);
	if (error != B_OK)
		RETURN_ERROR(error);
	error = buffer.RemoveInt32(id);
	if (error != B_OK)
		RETURN_ERROR(error);
PRINT(("RequestHeader::Read(): id: %ld\n", id));
	// sanity check data
	if (requestSize <= 0 || requestSize > kMaxSaneRequestSize
		|| messageSize < 0 || dataSize < 0) {
		RETURN_ERROR(B_BAD_DATA);
	}
	switch (parameterType) {
		case REQUEST_NO_PARAMETER:
			if (messageSize != 0)
				RETURN_ERROR(B_BAD_DATA);
			break;
		case REQUEST_MESSAGE_PARAMETER:
			if (messageSize == 0
				|| messageSize > kMaxSaneMessageParameterSize
				|| dataSize != 0) {
				RETURN_ERROR(B_BAD_DATA);
			}
			break;
		case REQUEST_DATA_PARAMETER:
			if (messageSize == 0
				|| messageSize > kMaxSaneMessageParameterSize
				|| dataSize > kMaxSaneDataParameterSize)
				RETURN_ERROR(B_BAD_DATA);
			break;
		default:
			RETURN_ERROR(B_BAD_DATA);
	}
	return B_OK;
}

// GetHeaderSize
int32
RequestHeader::GetHeaderSize()
{
	return 24;
}


// Request

// constructor
Request::Request()
	: fRequestMessage(NULL),
	  fParameter()
{
}

// ocnstructor
Request::Request(BMessage* request, Message* parameter)
	: fRequestMessage(request),
	  fParameter(parameter)
{
}

// destructor
Request::~Request()
{
	if (fRequestMessage)
		delete fRequestMessage;
}

// GetHeader
void
Request::GetHeader(RequestHeader* header) const
{
	header->requestSize = fRequestMessage->FlattenedSize();
	if (!fParameter.Get() || fParameter->IsEmpty()) {
		// no parameter
		header->parameterType = REQUEST_NO_PARAMETER;
		header->messageSize = 0;
		header->dataSize = 0;
	} else if (fParameter->IsDataMessage()) {
		// data message
		header->parameterType = REQUEST_DATA_PARAMETER;
		if (fParameter->GetMessage())
			header->messageSize = fParameter->GetMessage()->FlattenedSize();
		else
			header->messageSize = 0;
		header->dataSize = fParameter->GetDataSize();
	} else {
		// normal message
		header->parameterType = REQUEST_MESSAGE_PARAMETER;
		header->messageSize = fParameter->GetMessage()->FlattenedSize();
		header->dataSize = 0;
	}
}

// SetRequestMessage
void
Request::SetRequestMessage(BMessage* request)
{
	fRequestMessage = request;
}

// GetRequestMessage
BMessage*
Request::GetRequestMessage() const
{
	return fRequestMessage;
}

// SetParameter
void
Request::SetParameter(Message* parameter)
{
	fParameter.SetTo(parameter);
}

// GetParameter
Message*
Request::GetParameter() const
{
	return &*fParameter;
}

