// InputDeviceConnection.cpp
// 
// Copyright (c) 2004, Ingo Weinhold (bonefish@cs.tu-berlin.de)
// 
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation 
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
// THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
// DEALINGS IN THE SOFTWARE.
// 
// Except as contained in this notice, the name of a copyright holder shall
// not be used in advertising or otherwise to promote the sale, use or other
// dealings in this Software without prior written authorization of the
// copyright holder.

#include <new>

#include <Message.h>

#include <RemoteRoster.h>

#include "AutoDeleter.h"
#include "Debug.h"
#include "InputDeviceConnection.h"
#include "RDesktopServerDefs.h"

// constructor
InputDeviceConnection::InputDeviceConnection()
	: fInputDeviceTimeOffset(0),
	  fMessageSender(-1),
	  fMessageQueue(),
	  fTerminating(false)
{
}

// destructor
InputDeviceConnection::~InputDeviceConnection()
{
	fTerminating = true;
	fMessageQueue.Close(true);
	if (fMessageSender >= 0) {
		int32 result;
		wait_for_thread(fMessageSender, &result);
	}
}

// Init
status_t
InputDeviceConnection::Init(RemoteRoster* roster)
{
	// spawn the message sender thread
	fMessageSender = spawn_thread(_MessageSenderEntry,
		"input device messsage sender", B_NORMAL_PRIORITY, this);
	if (fMessageSender < 0) {
		FATAL(("Failed to spawn message sender thread: %s\n",
			strerror(fMessageSender)));
		return fMessageSender;
	}
	resume_thread(fMessageSender);
	// get a messenger for the remote input server
	BMessenger messenger;
	status_t error = roster->GetMessenger("application/x-vnd.Be-input_server",
		&messenger);
	if (error != B_OK) {
		FATAL(("Failed to get input server messenger: %s\n", strerror(error)));
		return error;
	}
	// get the device looper messenger
	error = InputServerConnection::Init(messenger,
		kRDesktopInputDeviceLooperName);
	if (error != B_OK) {
		FATAL(("Failed to get input device looper messenger: %s\n",
			strerror(error)));
		return error;
	}
	// send a connect request to the intput device
	BMessage connectMessage(MSG_INPUT_DEVICE_CONNECT);
	error = connectMessage.AddInt64("when", real_time_clock());
	if (error != B_OK)
		return error;
	BMessage reply;
	error = fTarget.SendMessage(&connectMessage, &reply);
	if (error != B_OK) {
		FATAL(("Failed to connect to remote input server device: %s\n",
			strerror(error)));
		return error;
	}
	if (reply.FindInt64("time offset", &fInputDeviceTimeOffset) != B_OK)
		fInputDeviceTimeOffset = 0;
	return error;
}

// PushMessage
status_t
InputDeviceConnection::PushMessage(BMessage* message)
{
	return fMessageQueue.Push(message);
}

// _MessageSenderEntry
int32
InputDeviceConnection::_MessageSenderEntry(void* data)
{
	return ((InputDeviceConnection*)data)->_MessageSender();
}

// _MessageSender
int32
InputDeviceConnection::_MessageSender()
{
	while (!fTerminating) {
		// fetch next message
		BMessage* message = NULL;
		status_t error = fMessageQueue.Pop(&message);
		if (error != B_OK || !message)
			continue;
		// if it is a mouse moved message check for subsequent mouse moved
		// message
		if (message->what == B_MOUSE_MOVED) {
			BAutolock _(fMessageQueue);
			BMessage* nextMessage = NULL;
			while (fMessageQueue.Peek(&nextMessage) == B_OK
				&& nextMessage && nextMessage->what == B_MOUSE_MOVED) {
				if (fMessageQueue.Pop(&nextMessage) != B_OK)
					break;
PRINT(("InputDeviceConnection::_MessageSender(): skipping B_MOUSE_MOVED message\n"));
				delete message;
				message = nextMessage;
			}
		}
		// update the "when" field in the message
		int64 when;
		if (message->FindInt64("when", &when) == B_OK)
			message->ReplaceInt64("when", when + fInputDeviceTimeOffset);
		// send message
		BMessage reply;
//PRINT(("Sending message: \n"));
//D(message->PrintToStream());
		error = fTarget.SendMessage(message, &reply);
		if (error != B_OK) {
			PRINT(("InputDeviceConnection::_MessageSender(): Failed to send "
			"message to input server: %s\n", strerror(error)));
		}
		delete message;
	}
	return B_OK;
}

