// InputFilterConnection.cpp
// 
// Copyright (c) 2004, Ingo Weinhold (bonefish@cs.tu-berlin.de)
// 
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation 
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
// THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
// DEALINGS IN THE SOFTWARE.
// 
// Except as contained in this notice, the name of a copyright holder shall
// not be used in advertising or otherwise to promote the sale, use or other
// dealings in this Software without prior written authorization of the
// copyright holder.

#include <new>

#include <InterfaceDefs.h>
#include <Message.h>
#include <MessageRunner.h>

#include "AutoDeleter.h"
#include "Debug.h"
#include "InputFilterConnection.h"
#include "InputMessageHandler.h"
#include "RDesktopServerDefs.h"

static const bigtime_t kInputFilterPingInterval = 1000000;	// 1s

// constructor
InputFilterConnection::InputFilterConnection(
	InputMessageHandler* inputMessageHandler)
	: fInputMessageHandler(inputMessageHandler),
	  fFilterPingRunner(NULL)
{
}

// destructor
InputFilterConnection::~InputFilterConnection()
{
	Disconnect();
}

// Init
status_t
InputFilterConnection::Init()
{
	// get a messenger for the local input server
	status_t error = B_OK;
	BMessenger messenger("application/x-vnd.Be-input_server", -1, &error);
	if (error != B_OK) {
		FATAL(("Failed to get local input server messenger: %s\n",
			strerror(error)));
		return error;
	}
	// get the filter looper messenger
	error = InputServerConnection::Init(messenger,
		kRDesktopInputFilterLooperName);
	if (error != B_OK) {
		FATAL(("Failed to get input filter looper messenger: %s\n",
			strerror(error)));
		return error;
	}
	return error;
}

// Connect
status_t
InputFilterConnection::Connect()
{
	// create a message runner that pings the input server filter
	Disconnect();
	BMessage pingerMessage(MSG_INPUT_FILTER_PING);
	fFilterPingRunner = new(nothrow) BMessageRunner(BMessenger(this),
		&pingerMessage, kInputFilterPingInterval);
	status_t error = B_OK;
	if (!fFilterPingRunner)
		error = B_NO_MEMORY;
	if (error == B_OK)
		error = fFilterPingRunner->InitCheck();
	// connect to the input filter
	BMessage connectMessage(MSG_INPUT_FILTER_CONNECT);
	if (error == B_OK)
		error = connectMessage.AddMessenger("listener", BMessenger(this));
	if (error == B_OK)
		error = fTarget.SendMessage(&connectMessage, this);
	if (error != B_OK)
		Disconnect();
	return error;
}

// Disconnect
void
InputFilterConnection::Disconnect()
{
	if (fFilterPingRunner) {
		delete fFilterPingRunner;
		fFilterPingRunner = NULL;
		fTarget.SendMessage(MSG_INPUT_FILTER_DISCONNECT);
	}
}

// MessageReceived
void
InputFilterConnection::MessageReceived(BMessage* message)
{
	switch (message->what) {
		case MSG_INPUT_FILTER_PING:
			fTarget.SendMessage(MSG_INPUT_FILTER_PING, this);
			break;
		case MSG_INPUT_FILTER_MESSAGE:
		{
			if (!fInputMessageHandler)
				break;
			// allocate a filter message
			BMessage* filterMessage = new BMessage;
			if (!filterMessage) {
				PRINT(("InputFilterConnection::MessageReceived("
					"MSG_INPUT_FILTER_MESSAGE): failed to allocate filter "
					"message:\n"));
			}
			// get the filter message
			status_t error = message->FindMessage("message", filterMessage);
			if (error != B_OK) {
				PRINT(("InputFilterConnection::MessageReceived("
					"MSG_INPUT_FILTER_MESSAGE): Failed to get message.\n"));
				delete filterMessage;
				break;
			}
			// check, if the message is an "escape" message
			bool inputEscape = false;
			if (filterMessage->what == B_KEY_DOWN) {
				int32 modifiers;
				const char* bytes;
				const int32 modifierPattern = B_COMMAND_KEY | B_CONTROL_KEY;
				if (filterMessage->FindString("bytes", &bytes) == B_OK
					&& filterMessage->FindInt32("modifiers", &modifiers) == B_OK
					&& bytes[0] == B_ESCAPE
					&& !(~modifiers & modifierPattern)) {
					inputEscape = true;
				}
			}
			// deliver the message
			if (inputEscape) {
				fInputMessageHandler->InputEscape();
				delete filterMessage;
			} else
				fInputMessageHandler->HandleInputMessage(filterMessage);
			break;
		}
		default:
			BHandler::MessageReceived(message);
			break;
	}
}

