// RDesktopClient.cpp
// 
// Copyright (c) 2004, Ingo Weinhold (bonefish@cs.tu-berlin.de)
// 
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation 
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
// THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
// DEALINGS IN THE SOFTWARE.
// 
// Except as contained in this notice, the name of a copyright holder shall
// not be used in advertising or otherwise to promote the sale, use or other
// dealings in this Software without prior written authorization of the
// copyright holder.

#include <new>

#include <Application.h>
#include <Message.h>
#include <MessageFilter.h>
#include <Window.h>

#include <RemoteMessage.h>
#include <RemoteRoster.h>

#include "AutoDeleter.h"
#include "Debug.h"
#include "DebugSupport.h"
#include "InputDeviceConnection.h"
#include "InputFilterConnection.h"
#include "RDesktopClient.h"
#include "RDesktopServerConnection.h"
#include "RDesktopServerDefs.h"
#include "RDesktopWindow.h"

// application signature
static const char* kRDesktopClientSignature
	= "application/x-vnd.bonefish.rdesktop-client";

// usage
const char* kUsage =
"Usage: RDesktopClient <options> <host>\n"
"Options:\n"
"  -h, --help     - print this text\n"
"  --no-gfx       - the screen images won't be transmitted to the client\n"
"  --no-mouse     - the client's mouse input won't be sent to the server\n"
"  --no-keyboard  - the client's keyboard input won't be sent to the server\n"
"\n";

// InputEscapeFilter
class InputEscapeFilter : public BMessageFilter {
public:
	InputEscapeFilter(InputMessageHandler* inputHandler)
		: BMessageFilter(B_KEY_DOWN),
		  fInputHandler(inputHandler)
	{
	}

	virtual filter_result Filter(BMessage* message, BHandler** target)
	{
		bool inputEscape = false;
		if (message->what == B_KEY_DOWN) {
			int32 modifiers;
			const char* bytes;
			const int32 modifierPattern = B_COMMAND_KEY | B_CONTROL_KEY;
			if (message->FindString("bytes", &bytes) == B_OK
				&& message->FindInt32("modifiers", &modifiers) == B_OK
				&& bytes[0] == B_ESCAPE
				&& !(~modifiers & modifierPattern)) {
				inputEscape = true;
			}
		}
		PRINT(("InputEscapeFilter::Filter(): input escape: %d\n", inputEscape));
		if (inputEscape) {
			fInputHandler->InputEscape();
			return B_SKIP_MESSAGE;
		}
		return B_DISPATCH_MESSAGE;
	}

private:
	InputMessageHandler*	fInputHandler;
};


// RDesktopClient

// constructor
RDesktopClient::RDesktopClient(const char* host, bool gfxEnabled,
	bool mouseEnabled, bool keyboardEnabled)
	: BApplication(kRDesktopClientSignature),
	  fHost(host),
	  fDesktopModel(gfxEnabled, mouseEnabled, keyboardEnabled),
	  fServerConnection(NULL),
	  fInputFilterConnection(NULL),
	  fInputDeviceConnection(NULL),
	  fInputActive(false),
	  fWindow(NULL)
{
}

// destructor
RDesktopClient::~RDesktopClient()
{
	Lock();
	if (fServerConnection) {
		RemoveHandler(fServerConnection);
		delete fServerConnection;
	}
	if (fInputDeviceConnection) {
		RemoveHandler(fInputDeviceConnection);
		delete fInputDeviceConnection;
	}
	if (fInputFilterConnection) {
		RemoveHandler(fInputFilterConnection);
		delete fInputFilterConnection;
	}
}

// Init
status_t
RDesktopClient::Init()
{
	// setup rmessaging to host
	RemoteRoster roster;
	status_t error = roster.SetTo(fHost.String());
	if (error != B_OK) {
		fprintf(stderr, "Failed to connect to host `%s': %s\n", fHost.String(),
			strerror(error));
		return error;
	}
	// setup connection to rmessage server
	if (fDesktopModel.IsGfxEnabled()) {
		fServerConnection = new(nothrow) RDesktopServerConnection(
			&fDesktopModel);
		if (!fServerConnection)
			return B_NO_MEMORY;
		AddHandler(fServerConnection);
		error = fServerConnection->Init(&roster);
		if (error != B_OK) {
			fprintf(stderr, "Failed to connect to RDesktopServer: %s\n",
				strerror(error));
			return error;
		}
	}
	// setup connection to the remote input device
	fInputDeviceConnection = new InputDeviceConnection;
	if (!fInputDeviceConnection)
		return B_NO_MEMORY;
	AddHandler(fInputDeviceConnection);
	error = fInputDeviceConnection->Init(&roster);
	if (error != B_OK) {
		fprintf(stderr, "Failed to connect to remote input device: %s\n",
			strerror(error));
		return error;
	}
	// setup connection to the local input filter
	if (fDesktopModel.IsKeyboardEnabled()) {
		fInputFilterConnection = new InputFilterConnection(this);
		if (!fInputFilterConnection)
			return B_NO_MEMORY;
		AddHandler(fInputFilterConnection);
		error = fInputFilterConnection->Init();
		if (error != B_OK) {
			fprintf(stderr, "Failed to connect to local input filter: %s\n",
				strerror(error));
			return error;
		}
	}
	// create window
	fWindow = new(nothrow) RDesktopWindow(fHost.String());
	if (!fWindow)
		return B_NO_MEMORY;
	fWindow->AddCommonFilter(new InputEscapeFilter(this));
	error = fWindow->Init(&fDesktopModel, this);
	if (error != B_OK) {
		fprintf(stderr, "Failed to initialize window: %s\n", strerror(error));
		delete fWindow;
		fWindow = NULL;
		return error;
	}
	fWindow->Show();
	return error;
}

// MessageReceived
void
RDesktopClient::MessageReceived(BMessage* message)
{
	switch (message->what) {
		default:
			BApplication::MessageReceived(message);
			break;
	}
}

// HandleInputMessage
void
RDesktopClient::HandleInputMessage(BMessage* message)
{
	if (fInputActive && fInputDeviceConnection
		&& fInputDeviceConnection->PushMessage(message) == B_OK) {
		// fine
	} else
		delete message;
}

// InputActivated
void
RDesktopClient::InputActivated(bool activated)
{
	if (activated) {
		if (fInputFilterConnection) {
			status_t error = fInputFilterConnection->Connect();
			if (error == B_OK) {
				fInputActive = true;
			} else {
				PRINT(("Failed to connect to input filter: %s\n",
					strerror(error)));
			}
		} else
			fInputActive = true;
	} else {
		if (fInputFilterConnection)
			fInputFilterConnection->Disconnect();
		fInputActive = false;
	}
}

// InputEscape
void
RDesktopClient::InputEscape()
{
	fWindow->PostMessage(B_ZOOM);
}

// print_usage
static
void
print_usage(bool error)
{
	fprintf((error ? stderr : stdout), kUsage);
}

// main
int
main(int argc, char** argv)
{
	init_debugging();
	PRINT(("main()\n"));
	// parse parameters
	bool gfxEnabled = true;
	bool mouseEnabled = true;
	bool keyboardEnabled = true;
	const char* host = NULL;
	int argi = 1;
	for (;argi < argc; argi++) {
		const char* arg = argv[argi];
		int argLen = strlen(arg);
		if (argLen > 0 && arg[0] == '-') {
			// option
			const char* option = arg + 1;
			// help
			if (strcmp(option, "h") == 0 || strcmp(option, "-help") == 0) {
				print_usage(false);
				return 0;
			// --no-gfx
			} else if (strcmp(option, "-no-gfx") == 0) {
				gfxEnabled = false;
			// --no-mouse
			} else if (strcmp(option, "-no-mouse") == 0) {
				mouseEnabled = false;
			// --no-keyboard
			} else if (strcmp(option, "-no-keyboard") == 0) {
				keyboardEnabled = false;
			// unknown
			} else {
				fprintf(stderr, "Illegal option: `-%s'\n", option);
				print_usage(true);
				return 1;
			}
		} else
			break;
	}
	if (argi >= argc) {
		print_usage(true);
		return 1;
	}
	// get host
	host = argv[argi];
	argi++;
	if (argi != argc) {
		print_usage(true);
		return 1;
	}
	// start application
	RDesktopClient app(host, gfxEnabled, mouseEnabled, keyboardEnabled);
	status_t error = app.Init();
	if (error == B_OK)
		app.Run();
	else
		fprintf(stderr, "Initialization failed: %s\n", strerror(error));
	PRINT(("main() done\n"));
	exit_debugging();
	return 0;
}
