// RDesktopServerConnection.cpp
// 
// Copyright (c) 2004, Ingo Weinhold (bonefish@cs.tu-berlin.de)
// 
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation 
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
// THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
// DEALINGS IN THE SOFTWARE.
// 
// Except as contained in this notice, the name of a copyright holder shall
// not be used in advertising or otherwise to promote the sale, use or other
// dealings in this Software without prior written authorization of the
// copyright holder.

#include <new>

#include <zlib.h>

#include <Bitmap.h>
#include <Messenger.h>

#include <RemoteMessage.h>
#include <RemoteRoster.h>

#include "Debug.h"
#include "DebugSupport.h"
#include "RDesktopServerDefs.h"
#include "RDesktopServerConnection.h"
#include "RDesktopServerModel.h"

// constructor
RDesktopServerConnection::RDesktopServerConnection(RDesktopServerModel* model)
	: BHandler("rdesktop server connection"),
	  fModel(model),
	  fToken(-1),
	  fRequestMessenger(),
	  fScreenFrame(),
	  fScreenColorSpace(B_RGB32),
	  fScreenShot(NULL),
	  fDecompressionBuffer(NULL),
	  fDecompressionBufferSize(0)
{
}

// constructor
RDesktopServerConnection::~RDesktopServerConnection()
{
	delete[] fDecompressionBuffer;
}

// Init
status_t
RDesktopServerConnection::Init(RemoteRoster* roster)
{
	// get server messenger
	BMessenger messenger;
	status_t error = roster->GetMessenger(kRDesktopServerSignature, &messenger);
	if (error != B_OK) {
		FATAL(("Failed to get server messenger: %s\n", strerror(error)));
		return error;
	}
	// send connect request
	BMessage message(MSG_CONNECT);
	error = message.AddMessenger("messenger", BMessenger(this));
	if (error != B_OK) {
		FATAL(("Failed to prepare connect message: %s\n", strerror(error)));
		return error;
	}
	BMessage reply;
	error = messenger.SendMessage(&message, &reply);
	if (error != B_OK) {
		FATAL(("Failed to send connect message: %s\n", strerror(error)));
		return error;
	}
	// inspect request
	// check what
	if (reply.what != MSG_REPLY) {
		FATAL(("Incorrect reply\n"));
		return B_ERROR;
	}
	// check error
	status_t replyError;
	if (reply.FindInt32("error", &replyError) == B_OK && replyError != B_OK) {
		FATAL(("Reply error: %s\n", strerror(replyError)));
		return replyError;
	}
	// get token
	error = reply.FindInt32("token", &fToken);
	if (error != B_OK) {
		FATAL(("No token in connect reply: %s\n", strerror(error)));
		return error;
	}
	// get request messenger
	error = reply.FindMessenger("request messenger", &fRequestMessenger);
	if (error != B_OK) {
		FATAL(("No request messenger in connect reply: %s\n", strerror(error)));
		return error;
	}
	return error;
}

// MessageReceived
void
RDesktopServerConnection::MessageReceived(BMessage* message)
{
	switch (message->what) {
		case MSG_DESKTOP_CONFIG_UPDATE:
			_HandleConfigUpdate(message);
			break;
		case RMESSAGE_DATA:
		{
char buffer[1024];
TOUCH(buffer);
PRINT(("RMESSAGE_DATA: reply target: %s\n",
messenger_to_string(message->ReturnAddress(), buffer)));
			_HandleGraphicsUpdate(message);
			break;
		}
		default:
			BHandler::MessageReceived(message);
			break;
	}
}

// _HandleConfigUpdate
void
RDesktopServerConnection::_HandleConfigUpdate(BMessage* message)
{
	// screen frame
	BRect screenFrame;
	status_t error = message->FindRect("screen frame", &screenFrame);
	if (error != B_OK) {
		PRINT(("Failed to get screen frame: %s\n", strerror(error)));
		return;
	}
	// color space
	color_space colorSpace;
	error = message->FindInt32("color space", (int32*)&colorSpace);
	if (error != B_OK) {
		PRINT(("Failed to get color space: %s\n", strerror(error)));
		return;
	}
	if (fScreenFrame != screenFrame || colorSpace != fScreenColorSpace)
		_DesktopConfigChanged(screenFrame, colorSpace);
}

// _HandleGraphicsUpdate
void
RDesktopServerConnection::_HandleGraphicsUpdate(BMessage* message)
{
	// get data from message
	const void* data;
	size_t size;
	BMessage parameters;
	bool hasParameters = false;
	status_t error = RemoteMessage::GetData(message, &data, &size,
		&parameters, &hasParameters);
	if (error != B_OK) {
		PRINT(("Failed to get data: %s\n", strerror(error)));
		return;
	}
	// release data when returning
	struct DataReleaser {
		const void*	fData;
		size_t		fSize;

		DataReleaser(const void* data, size_t size)
			: fData(data), fSize(size) {}

		~DataReleaser()
		{
			RemoteMessage::ReleaseData(fData, fSize);
		}
	} dataReleaser(data, size);
	// get parameters
	// check, if there are any, first
	if (!hasParameters || parameters.what != MSG_DESKTOP_GRAPHICS_UPDATE) {
		PRINT(("Wrong parameters: %s\n", strerror(error)));
		return;
	}
	// update frame
	BRect updateFrame;
	error = parameters.FindRect("update frame", &updateFrame);
	if (error != B_OK) {
		PRINT(("Failed to get update frame: %s\n", strerror(error)));
		return;
	}
	// color space
	color_space colorSpace;
	error = parameters.FindInt32("color space", (int32*)&colorSpace);
	if (error != B_OK) {
		PRINT(("Failed to get color space: %s\n", strerror(error)));
		return;
	}
	// bytes per row
	int32 bpr;
	error = parameters.FindInt32("bytes per row", &bpr);
	if (error != B_OK) {
		PRINT(("Failed to get BPR: %s\n", strerror(error)));
		return;
	}
	// data offset
	int32 dataOffset;
	error = parameters.FindInt32("data offset", &dataOffset);
	if (error != B_OK) {
		PRINT(("Failed to get data offset: %s\n", strerror(error)));
		return;
	}
	if (dataOffset < 0) {
		PRINT(("Bad data offset: %ld\n", dataOffset));
		return;
	}
	// delta
	bool delta;
	error = parameters.FindBool("delta", &delta);
	if (error != B_OK) {
		PRINT(("Failed to get delta: %s\n", strerror(error)));
		return;
	}
	// compressed
	bool compressed;
	error = parameters.FindBool("compressed", &compressed);
	if (error != B_OK) {
		PRINT(("Failed to get compressed: %s\n", strerror(error)));
		return;
	}
	// if compressed decompress
	if (compressed) {
		int32 decompressedSize;
		error = _Decompress(data, size, &decompressedSize);
		if (error != B_OK) {
			PRINT(("Failed to decompress data: %s\n", strerror(error)));
			return;
		}
		data = fDecompressionBuffer;
		size = decompressedSize;
	}
	_UpdateScreen(updateFrame, colorSpace, data, size, bpr, dataOffset, delta);
}

// _DesktopConfigChanged
void
RDesktopServerConnection::_DesktopConfigChanged(BRect screenFrame,
	color_space colorSpace)
{
	RDesktopServerModelLocker _(fModel);
	// unset old
	fScreenShot = NULL;
	fScreenColorSpace = colorSpace;
	fScreenFrame = screenFrame;
	delete[] fDecompressionBuffer;
	fDecompressionBufferSize = 0;
	// set new
	if (fModel->SetScreenConfig(screenFrame, colorSpace) != B_OK)
		return;
	fScreenShot = fModel->GetScreen();
	fDecompressionBufferSize = fScreenShot->BitsLength();
	fDecompressionBuffer = new(nothrow) uint8[fDecompressionBufferSize];
	if (!fDecompressionBuffer) {
		fDecompressionBufferSize = 0;
		PRINT(("Allocating the decompression buffer failed\n"));
		return;
	}
}

// _UpdateScreen
void
RDesktopServerConnection::_UpdateScreen(BRect updateFrame,
	color_space colorSpace, const void* data, size_t size, int32 bpr,
	int32 dataOffset, bool delta)
{
// TODO: ...
	if (!fScreenShot || fScreenShot->Bounds() != updateFrame
		|| fScreenShot->ColorSpace() != colorSpace
		|| fScreenShot->BytesPerRow() != bpr
		|| fScreenShot->BitsLength() < (int32)size + dataOffset) {
		PRINT(("Incompatible or currently unsupported update data.\n"));
		return;
	}
	uint8* bits = (uint8*)fScreenShot->Bits() + dataOffset;
	if (delta) {
		const uint8* deltaBits = (const uint8*)data;
		const uint8* deltaBitsEnd = deltaBits + size;
		for (; deltaBits != deltaBitsEnd; deltaBits++, bits++)
			*bits += *deltaBits;
	} else
		memcpy(bits, data, size);
	fModel->ScreenUpdated();
}

// _Decompress
status_t
RDesktopServerConnection::_Decompress(const void* data, int32 dataSize,
	int32* decompressedSize)
{
	if (!data || dataSize <= 0 || !fDecompressionBuffer
		|| fDecompressionBufferSize <= 0 || !decompressedSize) {
		return B_BAD_VALUE;
	}
	// prepare stream
	z_stream zStream = {
		(Bytef*)data,					// next_in
		dataSize,						// avail_in
		0,								// total_in
		(Bytef*)fDecompressionBuffer,	// next_out
		fDecompressionBufferSize,		// avail_out
		0,								// total_out
		0,								// msg
		0,								// state;
		Z_NULL,							// zalloc
		Z_NULL,							// zfree
		Z_NULL,							// opaque
		0,								// data_type
		0,								// adler
		0,								// reserved
	};
	int error = inflateInit(&zStream);
	if (error != Z_OK) {
PRINT(("RDesktopServerConnection::_Decompress(): failed: %d\n", error));
		return B_ERROR;
	}
	// inflate
	status_t result = B_OK;
	error = inflate(&zStream, Z_FINISH);
	if (error != Z_STREAM_END) {
PRINT(("RDesktopServerConnection::_Decompress(): failed: %d\n", error));
		result = B_ERROR;
	}
	if (inflateEnd(&zStream) != Z_OK && result == B_OK) {
PRINT(("RDesktopServerConnection::_Decompress(): failed: %d\n", error));
		result = B_ERROR;
	}
	if (result == B_OK)
		*decompressedSize = zStream.total_out;
	return result;
}
