// RDesktopServerModel.cpp
// 
// Copyright (c) 2004, Ingo Weinhold (bonefish@cs.tu-berlin.de)
// 
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation 
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
// THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
// DEALINGS IN THE SOFTWARE.
// 
// Except as contained in this notice, the name of a copyright holder shall
// not be used in advertising or otherwise to promote the sale, use or other
// dealings in this Software without prior written authorization of the
// copyright holder.

#include <string.h>

#include <Bitmap.h>

#include "Debug.h"
#include "RDesktopServerDefs.h"
#include "RDesktopServerModel.h"

// constructor
RDesktopServerModel::RDesktopServerModel(bool gfxEnabled, bool mouseEnabled,
	bool keyboardEnabled)
	: fLock(),
	  fGfxEnabled(gfxEnabled),
	  fMouseEnabled(mouseEnabled),
	  fKeyboardEnabled(keyboardEnabled),
	  fScreenFrame(),
	  fScreenColorSpace(B_RGB32),
	  fScreen(NULL),
	  fObservers(),
	  fModified(false)
{
}

// destructor
RDesktopServerModel::~RDesktopServerModel()
{
	delete fScreen;
	for (ObserverMap::Iterator it = fObservers.GetIterator(); it.HasNext();) {
		ObserverMap::Entry entry = it.Next();
		delete entry.value;
	}
}

// Lock
bool
RDesktopServerModel::Lock()
{
	return fLock.Lock();
}

// Unlock
void
RDesktopServerModel::Unlock()
{
	if (fLock.IsLocked()) {
		if (fModified && fLock.CountLocks() == 1) {
			// notify the observers that something has changed
			BMessage message(RDESKTOP_SERVER_MODEL_CHANGED);
			for (ObserverMap::Iterator it = fObservers.GetIterator();
				 it.HasNext();) {
				ObserverMap::Entry entry = it.Next();
				entry.key.SendMessage(&message);
			}
			fModified = false;
		}
		fLock.Unlock();
	}
}

// IsGfxEnabled
bool
RDesktopServerModel::IsGfxEnabled() const
{
	return fGfxEnabled;
}

// IsMouseEnabled
bool
RDesktopServerModel::IsMouseEnabled() const
{
	return fMouseEnabled;
}

// IsKeyboardEnabled
bool
RDesktopServerModel::IsKeyboardEnabled() const
{
	return fKeyboardEnabled;
}

// SetScreenConfig
status_t
RDesktopServerModel::SetScreenConfig(BRect frame, color_space colorSpace)
{
	RDesktopServerModelLocker _(this);
	// delete old screen bitmap
	delete fScreen;
	fScreen = NULL;
	// set new config values
	fScreenFrame = frame;
	fScreenColorSpace = colorSpace;
	_AddToChangeMasks(RDESKTOP_SCREEN_CONFIG);
	// allocate new screen bitmap
	fScreen = new(nothrow) BBitmap(frame, colorSpace);
	if (!fScreen) {
		PRINT(("Allocating the screen bitmap failed.\n"));
		return B_NO_MEMORY;
	}
	status_t error = fScreen->InitCheck();
	if (error != B_OK) {
		PRINT(("Initializing the screen bitmap failed: %s\n",
			strerror(fScreen->InitCheck())));
		delete fScreen;
		fScreen = NULL;
		return error;
	}
	return B_OK;
}

// ScreenUpdated
void
RDesktopServerModel::ScreenUpdated()
{
	RDesktopServerModelLocker _(this);
	_AddToChangeMasks(RDESKTOP_SCREEN_GRAPHICS);
}

// GetScreenFrame
BRect
RDesktopServerModel::GetScreenFrame() const
{
	return fScreenFrame;
}

// GetColorSpace
color_space
RDesktopServerModel::GetColorSpace() const
{
	return fScreenColorSpace;
}

// GetScreen
BBitmap*
RDesktopServerModel::GetScreen() const
{
	return fScreen;
}

// AddObserver
status_t
RDesktopServerModel::AddObserver(BMessenger observer)
{
	RDesktopServerModelLocker _(this);
	uint32* mask = new(nothrow) uint32(0);
	if (!mask)
		return B_NO_MEMORY;
	status_t error = fObservers.Put(observer, mask);
	if (error != B_OK)
		delete mask;
	return error;
}

// RemoveObserver
void
RDesktopServerModel::RemoveObserver(BMessenger observer)
{
	RDesktopServerModelLocker _(this);
	delete fObservers.Remove(observer);
}

// ClearChangeMask
uint32
RDesktopServerModel::ClearChangeMask(BMessenger observer)
{
	RDesktopServerModelLocker _(this);
	uint32 changes = 0;
	if (uint32* mask = fObservers.Get(observer)) {
		changes = *mask;
		*mask = 0;
	}
	return changes;
}

// _AddToChangeMasks
void
RDesktopServerModel::_AddToChangeMasks(uint32 changes)
{
	fModified = true;
	for (ObserverMap::Iterator it = fObservers.GetIterator(); it.HasNext();) {
		ObserverMap::Entry entry = it.Next();
		*entry.value |= changes;
	}
}

