// RDesktopView.cpp
// 
// Copyright (c) 2004, Ingo Weinhold (bonefish@cs.tu-berlin.de)
// 
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation 
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
// THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
// DEALINGS IN THE SOFTWARE.
// 
// Except as contained in this notice, the name of a copyright holder shall
// not be used in advertising or otherwise to promote the sale, use or other
// dealings in this Software without prior written authorization of the
// copyright holder.

#include <algorithm>
#include <new>

#include <Bitmap.h>
#include <Screen.h>
#include <ScrollBar.h>
#include <Window.h>

#include "AutoDeleter.h"
#include "Debug.h"
#include "DebugSupport.h"
#include "InputMessageHandler.h"
#include "RDesktopServerDefs.h"
#include "RDesktopServerModel.h"
#include "RDesktopView.h"

static const bigtime_t kInputFilterPingInterval = 1000000;	// 1s

static const BRect kDefaultScreenViewFrame(0, 0, 799, 599);

// constructor
RDesktopView::RDesktopView(RDesktopServerModel* model,
	InputMessageHandler* inputHandler, BRect frame, uint32 resizeMask,
	uint32 flags)
	: BView(frame, "desktop view", resizeMask, flags | B_FRAME_EVENTS),
	  fModel(model),
	  fInputHandler(inputHandler),
	  fScreenFrame(),
	  fScreenViewFrame(kDefaultScreenViewFrame),
	  fIsFullScreen(false)
{
	SetViewColor(B_TRANSPARENT_32_BIT);
}

// destructor
RDesktopView::~RDesktopView()
{
}

// MessageReceived
void
RDesktopView::MessageReceived(BMessage* message)
{
char buffer[10];
TOUCH(buffer);
PRINT(("RDesktopView::MessageReceived(%s)\n", message_what(message, buffer)));
	switch (message->what) {
		case RDESKTOP_SERVER_MODEL_CHANGED:
		{
			RDesktopServerModelLocker _(fModel);
			uint32 mask = fModel->ClearChangeMask(BMessenger(this));
			if (mask && RDESKTOP_SCREEN_CONFIG)
				_ScreenConfigChanged();
			if (mask && RDESKTOP_SCREEN_GRAPHICS)
				Invalidate();
			break;
		}
		default:
			BView::MessageReceived(message);
			break;
	}
}

// Draw
void
RDesktopView::Draw(BRect updateRect)
{
	RDesktopServerModelLocker _(fModel);
	if (fModel->GetScreenFrame() != fScreenFrame)
		_ScreenConfigChanged();
	if (BBitmap* screen = fModel->GetScreen()) {
		DrawBitmap(screen, screen->Bounds(), fScreenViewFrame);
	} else {
		SetHighColor(0, 0, 0);
		FillRect(updateRect);
	}
	// fill the space right of and below the bitmap
	SetHighColor(ui_color(B_PANEL_BACKGROUND_COLOR));
	BRect rightRect(updateRect);
	rightRect.left = fScreenViewFrame.right + 1;
	if (rightRect.IsValid())
		FillRect(rightRect);
	BRect bottomRect(updateRect);
	bottomRect.top = fScreenViewFrame.bottom + 1;
	if (bottomRect.IsValid())
		FillRect(bottomRect);
}

// FrameResized
void
RDesktopView::FrameResized(float width, float height)
{
	BView::FrameResized(width, height);
	_UpdateScrollers();
}

// KeyDown
void
RDesktopView::KeyDown(const char *bytes, int32 numBytes)
{
PRINT(("RDesktopView::KeyDown()\n"));
	if (numBytes > 0 && bytes && bytes[0] == B_ESCAPE) {
PRINT(("  B_ESCAPE\n"));
		BMessage* message = Window()->CurrentMessage();
		const int32 modifierPattern = B_COMMAND_KEY | B_CONTROL_KEY;
		int32 modifiers;
		if (message->FindInt32("modifiers", &modifiers) == B_OK
			&& !(~modifiers & modifierPattern)) {
PRINT(("  fInputHandler->InputEscape()\n"));
			fInputHandler->InputEscape();
		}
	}
}

// MouseMoved
void
RDesktopView::MouseMoved(BPoint where, uint32 code, const BMessage *message)
{
	_MouseEvent(B_MOUSE_MOVED, where);
	BView::MouseMoved(where, code, message);
}

// MouseDown
void
RDesktopView::MouseDown(BPoint where)
{
	_MouseEvent(B_MOUSE_DOWN, where);
	BView::MouseDown(where);
}

// MouseUp
void
RDesktopView::MouseUp(BPoint where)
{
	_MouseEvent(B_MOUSE_UP, where);
	BView::MouseUp(where);
}

// AttachedToWindow
void
RDesktopView::AttachedToWindow()
{
	fModel->AddObserver(BMessenger(this));
	_UpdateScrollers();
//	MakeFocus(true);
}

// DetachedFromWindow
void
RDesktopView::DetachedFromWindow()
{
	fModel->RemoveObserver(BMessenger(this));
}


// SetFullScreen
void
RDesktopView::SetFullScreen(bool fullScreen)
{
	fIsFullScreen = fullScreen;
	_UpdateScreenViewFrame();
}

// _MouseEvent
void
RDesktopView::_MouseEvent(uint32 what, BPoint where)
{
	if (!fInputHandler || !fModel->IsMouseEnabled())
		return;
	float x = where.x / (fScreenViewFrame.Width() + 1);
	float y = where.y / (fScreenViewFrame.Height() + 1);
	x = max(0.0f, min(x, 1.0f));
	y = max(0.0f, min(y, 1.0f));
	BMessage* remoteMessage = new BMessage(what);
	if (!remoteMessage)
		return;
	remoteMessage->AddFloat("x", x);
	remoteMessage->AddFloat("y", y);
	if (BMessage* currentMessage = Window()->CurrentMessage()) {
		int64 when;
		if (currentMessage->FindInt64("when", &when) == B_OK)
			remoteMessage->AddInt64("when", when); 
		int32 buttons;
		if (currentMessage->FindInt32("buttons", &buttons) == B_OK)
			remoteMessage->AddInt32("buttons", buttons); 
		int32 modifiers;
		if (currentMessage->FindInt32("modifiers", &modifiers) == B_OK)
			remoteMessage->AddInt32("modifiers", modifiers);
		int32 clicks;
		if (currentMessage->FindInt32("clicks", &clicks) == B_OK)
			remoteMessage->AddInt32("clicks", clicks);
	}
	// pass the message to the handler
	fInputHandler->HandleInputMessage(remoteMessage);
}

// _UpdateScrollers
void
RDesktopView::_UpdateScrollers()
{
	float hRange = 0;
	float vRange = 0;
	float hProportion = 1;
	float vProportion = 1;
	BRect frame(Frame());
	hRange = max(fScreenViewFrame.Width() - frame.Width(), 0.0f);
	vRange = max(fScreenViewFrame.Height() - frame.Height(), 0.0f);
	hProportion
		= min((frame.Width() + 1) / (fScreenViewFrame.Width() + 1), 1.0f);
	vProportion
		= min((frame.Height() + 1) / (fScreenViewFrame.Height() + 1), 1.0f);
	if (BScrollBar* hScrollBar = ScrollBar(B_HORIZONTAL))
{
PRINT(("hScrollBar->SetRange(0, %f)\n", hRange));
		hScrollBar->SetRange(0, hRange);
		hScrollBar->SetProportion(hProportion);
}
	if (BScrollBar* vScrollBar = ScrollBar(B_VERTICAL))
{
PRINT(("vScrollBar->SetRange(0, %f)\n", vRange));
		vScrollBar->SetRange(0, vRange);
		vScrollBar->SetProportion(vProportion);
}
BRect rect(Bounds());
PRINT(("bounds: (%f, %f) - (%f, %f)\n", rect.left, rect.top, rect.right,
rect.bottom));
rect = Frame();
PRINT(("frame: (%f, %f) - (%f, %f)\n", rect.left, rect.top, rect.right,
rect.bottom));
}

// _ScreenConfigChanged
void
RDesktopView::_ScreenConfigChanged()
{
	RDesktopServerModelLocker _(fModel);
	BRect screenFrame = fModel->GetScreenFrame();
	if (screenFrame != fScreenFrame) {
		fScreenFrame = screenFrame;
		_UpdateScreenViewFrame();
	}
}

// _UpdateScreenViewFrame
void
RDesktopView::_UpdateScreenViewFrame()
{
	if (fIsFullScreen) {
		BScreen screen;
		BRect screenFrame(screen.Frame());
		if (screenFrame.IsValid())
			fScreenViewFrame = screenFrame;
	} else {
		if (fScreenFrame.IsValid())
			fScreenViewFrame = fScreenFrame;
		else
			fScreenViewFrame = kDefaultScreenViewFrame;
	}
	_UpdateScrollers();
	Invalidate();
}

