// RDesktopWindow.cpp
// 
// Copyright (c) 2004, Ingo Weinhold (bonefish@cs.tu-berlin.de)
// 
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation 
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
// THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
// DEALINGS IN THE SOFTWARE.
// 
// Except as contained in this notice, the name of a copyright holder shall
// not be used in advertising or otherwise to promote the sale, use or other
// dealings in this Software without prior written authorization of the
// copyright holder.

#include <new>

#include <zlib.h>

#include <Application.h>
#include <Bitmap.h>
#include <Messenger.h>
#include <Screen.h>
#include <ScrollView.h>

#include "Debug.h"
#include "DebugSupport.h"
#include "InputMessageHandler.h"
#include "RDesktopServerDefs.h"
#include "RDesktopView.h"
#include "RDesktopWindow.h"

// constructor
RDesktopWindow::RDesktopWindow(const char* title)
	: BWindow(BRect(100, 100, 1000, 600), title, B_DOCUMENT_WINDOW,
		B_ASYNCHRONOUS_CONTROLS),
	  fView(NULL),
	  fInputHandler(NULL),
	  fIsFullScreen(false),
	  fOldFrame()
{
}

// Init
status_t
RDesktopWindow::Init(RDesktopServerModel* model,
	InputMessageHandler* inputHandler)
{
	if (fView)
		return B_ERROR;	// Init() called twice
	fInputHandler = inputHandler;
	// create view
	BRect viewFrame(Bounds());
	viewFrame.right -=  B_V_SCROLL_BAR_WIDTH;
	viewFrame.bottom -= B_H_SCROLL_BAR_HEIGHT; 
	fView = new(nothrow) RDesktopView(model, inputHandler, viewFrame,
		B_FOLLOW_ALL_SIDES, B_WILL_DRAW);
	if (!fView)
		return B_NO_MEMORY;
	// create scroll view
	BScrollView* scrollView = new(nothrow) BScrollView(NULL, fView,
		B_FOLLOW_ALL_SIDES, 0, true, true, B_PLAIN_BORDER);
	if (!scrollView) {
		delete fView;
		fView = NULL;
		return B_NO_MEMORY;
	}
	AddChild(scrollView);
	return B_OK;
}

// MessageReceived
void
RDesktopWindow::MessageReceived(BMessage* message)
{
	switch (message->what) {
		default:
			BWindow::MessageReceived(message);
			break;
	}
}

// QuitRequested
bool
RDesktopWindow::QuitRequested()
{
	be_app->PostMessage(B_QUIT_REQUESTED);
	return true;
}

// WindowActivated
void
RDesktopWindow::WindowActivated(bool activated)
{
	BWindow::WindowActivated(activated);
	if (fInputHandler)
		fInputHandler->InputActivated(activated);
	if (fView)
		fView->MakeFocus(true);
}

// Zoom
void
RDesktopWindow::Zoom(BPoint location, float width, float height)
{
	if (fIsFullScreen) {
		location = fOldFrame.LeftTop();
		width = fOldFrame.Width();
		height = fOldFrame.Height();
	} else {
		fOldFrame = Frame();
		// location
		BPoint viewLocation = fView->ConvertToScreen(fView->Bounds().LeftTop());
		location = Frame().LeftTop() - viewLocation;
		// size
		float viewWidth = fView->Frame().Width();
		float viewHeight = fView->Frame().Height();
		float widthDelta = Frame().Width() - viewWidth;
		float heightDelta = Frame().Height() - viewHeight;
		BScreen screen;
		BRect screenFrame(screen.Frame());
		if (screenFrame.IsValid()) {
			width = screenFrame.Width() + widthDelta;
			height = screenFrame.Height() + heightDelta;
		}
	}
	BWindow::Zoom(location, width, height);
	fIsFullScreen = !fIsFullScreen;
	fView->SetFullScreen(fIsFullScreen);
}

