// RDesktopServerDefs.h
// 
// Copyright (c) 2004, Ingo Weinhold (bonefish@cs.tu-berlin.de)
// 
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation 
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
// THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
// DEALINGS IN THE SOFTWARE.
// 
// Except as contained in this notice, the name of a copyright holder shall
// not be used in advertising or otherwise to promote the sale, use or other
// dealings in this Software without prior written authorization of the
// copyright holder.

#ifndef RDESKTOP_SERVER_DEFS_H
#define RDESKTOP_SERVER_DEFS_H

#include "RDesktopBuild.h"

// message what types
enum {
	// client-server messages
	MSG_REPLY							= 'RDre',
	MSG_CONNECT							= 'RDco',
	MSG_DESKTOP_CONFIG_UPDATE			= 'RDcu',
	MSG_DESKTOP_GRAPHICS_UPDATE			= 'RDgu',

	// server internal messages
	MSG_UPDATE_TICK						= 'RDtk',

	// client-input server device messages
	MSG_INPUT_DEVICE_CONNECT			= 'RDdc',

	// client-input server filter messages
	MSG_INPUT_FILTER_CONNECT			= 'RDfc',
	MSG_INPUT_FILTER_DISCONNECT			= 'RDfd',
	MSG_INPUT_FILTER_PING				= 'RDpi',
	MSG_INPUT_FILTER_MESSAGE			= 'RDfm',

	// input server filter internal messages
	MSG_INPUT_FILTER_CHECK_PING			= 'RDcp',

	// client internal messages
	RDESKTOP_SERVER_MODEL_CHANGED		= 'RDmc',
};

// the server signature
extern _IMPEXP_RDESKTOP const char* kRDesktopServerSignature;

// the name of the looper the input server device creates
extern _IMPEXP_RDESKTOP const char* kRDesktopInputDeviceLooperName;

// the name of the looper the input server filter creates
extern _IMPEXP_RDESKTOP const char* kRDesktopInputFilterLooperName;


/*
Client - Server Protocols
-------------------------

standard reply message

reply:		MSG_REPLY
on error:	"error": B_INT32_TYPE
			[ "error_description": B_STRING_TYPE ]
			[ <additional fields> ]

fields:
- "error": The error code (a status_t).
- "error_description": Optional human readable description.
- <additional fields>: Request-specific fields.

-----------------------------------------------------------------------

connection request

Connects with a remote desktop server and returns a token identifying the
client.

target:		server app looper
message:	MSG_CONNECT
			"messenger": 			B_MESSENGER_TYPE
reply:		standard success
			"token":				B_INT32_TYPE
			"request messenger":	B_MESSENGER_TYPE
on error:	- B_NO_REPLY (fatal)
			- standard error

message fields:
- "messenger": The messenger referring to the client's handler the server
  shall send update messages to.

reply fields:
- "token": A token assigned to the client. It must be passed with subsequent
  requests.
- "request_messenger": Messenger to which to send further requests.

error reply fields:
- "error":
  - B_BAD_VALUE: A request message field is missing or contains an
    invalid value.
  - ...

-----------------------------------------------------------------------

desktop config update

Sent by the server to notify the client of a configuration change. The
message is sent after the client connected to the server, and each time
the desktop resolution or depth changes.

target:		client messenger
message:	MSG_DESKTOP_CONFIG_UPDATE
			"screen frame": 		B_RECT_TYPE
			"color space":			B_INT32_TYPE
reply:		none

message fields:
- "screen frame": The screen frame.
- "color space": The color space the used for the screen (color_space).

-----------------------------------------------------------------------

desktop graphics update

Sent by the server to notify the client of a change of the screen graphics.

target:		client messenger
message:	data message
parameter:	MSG_DESKTOP_GRAPHICS_UPDATE
			"update frame":			B_RECT_TYPE
			"color space":			B_INT32_TYPE
			"bytes per row":		B_INT32_TYPE
			"delta":				B_BOOL_TYPE
reply:		none

parameter fields:
- "update frame": The update region.
- "color space": The color space used for the data.
- "bytes per row": The number of bytes per bitmap row.
- "delta": If true, the bytes in the data are differences to the last state.

-----------------------------------------------------------------------

input device connection request

Sent by the client to initiate a connection. Currently only the difference
between client and server time is computed and sent back. So there's no need
to disconnect actively later.

target:		(server side) input server device looper
message:	MSG_INPUT_DEVICE_CONNECT
			"when":					B_INT64_TYPE
reply:		MSG_REPLY
			"time offset":			B_INT64_TYPE

message fields:
- "when": The client time at which the client sends the connection request.

reply fields:
- "time offset": The difference between the server time at which the connection
  request arrived and the transmitted client time. The client shall add this
  time difference to all timestamps in messages it sends to the device looper.

-----------------------------------------------------------------------

input filter connection request

Sent by the client to initiate a connection. After receiving the connection
request, the filter will eat all keyboard event messages and send them via
MSG_INPUT_FILTER_MESSAGE messages to the client, until the client
sends a MSG_INPUT_FILTER_DISCONNECT message, sending a message to the client
failed, or the client failed to send a MSG_INPUT_FILTER_PING for a certain
time.

target:		(client side) input server filter looper
message:	MSG_INPUT_FILTER_CONNECT
			"listener":				B_MESSENGER_TYPE
reply:		none

message fields:
- "listener": The address at which the client desires to get the event
  messages delivered.

-----------------------------------------------------------------------

input filter disconnection request

Sent by the client to terminate a connection. The filter will immediately
stop to eat keyboard event messages and forward them to the client.

target:		(client side) input server filter looper
message:	MSG_INPUT_FILTER_DISCONNECT
reply:		none

-----------------------------------------------------------------------

input filter event message

Sent by the input server device to the client. Contains an keyboard
event message.

target:		client
message:	MSG_INPUT_FILTER_MESSAGE
			"message":				B_MESSAGE_TYPE
reply:		none

*/

#endif	// RDESKTOP_SERVER_DEFS_H
