// RDesktopInputFilter.cpp
// 
// Copyright (c) 2004, Ingo Weinhold (bonefish@cs.tu-berlin.de)
// 
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation 
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
// THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
// DEALINGS IN THE SOFTWARE.
// 
// Except as contained in this notice, the name of a copyright holder shall
// not be used in advertising or otherwise to promote the sale, use or other
// dealings in this Software without prior written authorization of the
// copyright holder.

#include <new>

#include <Autolock.h>
#include <Looper.h>
#include <Message.h>
#include <MessageRunner.h>
#include <Messenger.h>
#include <add-ons/input_server/InputServerFilter.h>

#include "Debug.h"
#include "DebugSupport.h"
#include "RDesktopServerDefs.h"

class RDesktopInputFilter;

static const bigtime_t kInputFilterPingCheckInterval = 5000000;	// 5s

// RDesktopInputFilterLooper
class RDesktopInputFilterLooper : public BLooper {
public:
								RDesktopInputFilterLooper(
									RDesktopInputFilter* device);
	virtual						~RDesktopInputFilterLooper();

	virtual	void				MessageReceived(BMessage* message);

			BMessenger			GetListener();
			void				SetListener(BMessenger listener);

private:
			RDesktopInputFilter*	fFilter;
			BMessenger				fListener;
			BMessageRunner*			fPingCheckRunner;
			bool					fPingReceived;
};

// RDesktopInputFilter
class RDesktopInputFilter : public BInputServerFilter {
public:
								RDesktopInputFilter();
	virtual						~RDesktopInputFilter();

	virtual	status_t			InitCheck();

	virtual	filter_result		Filter(BMessage* message, BList* outList);

private:
			RDesktopInputFilterLooper*	fLooper;

};

// instantiate_input_filter
extern "C" _EXPORT
BInputServerFilter*
instantiate_input_filter()
{
	return new(nothrow) RDesktopInputFilter;
}

// RDesktopInputFilterLooper

// constructor
RDesktopInputFilterLooper::RDesktopInputFilterLooper(
	RDesktopInputFilter* filter)
	: BLooper(kRDesktopInputFilterLooperName),
	  fFilter(filter),
	  fListener(),
	  fPingCheckRunner(NULL)
{
	BMessage pingerMessage(MSG_INPUT_FILTER_CHECK_PING);
	fPingCheckRunner = new BMessageRunner(BMessenger(this), &pingerMessage,
		kInputFilterPingCheckInterval);
}

// destructor
RDesktopInputFilterLooper::~RDesktopInputFilterLooper()
{
	delete fPingCheckRunner;
}

// MessageReceived
void
RDesktopInputFilterLooper::MessageReceived(BMessage* message)
{
	switch (message->what) {
		case MSG_INPUT_FILTER_CONNECT:
		{
PRINT(("RDesktopInputFilterLooper::MessageReceived(MSG_INPUT_FILTER_CONNECT)\n"));
			BMessenger listener;
			if (message->FindMessenger("listener", &listener) == B_OK
				&& listener.IsValid()) {
PRINT(("  connected\n"));
				fListener = listener;
				fPingReceived = true;
			}
			break;
		}
		case MSG_INPUT_FILTER_DISCONNECT:
		{
			fListener = BMessenger();
			fPingReceived = false;
		}
		case MSG_INPUT_FILTER_PING:
PRINT(("RDesktopInputFilterLooper::MessageReceived(MSG_INPUT_FILTER_PING)\n"));
			fPingReceived = true;
			break;
		case MSG_INPUT_FILTER_CHECK_PING:
		{
PRINT(("RDesktopInputFilterLooper::MessageReceived(MSG_INPUT_FILTER_CHECK_PING)\n"));
			if (!fPingReceived)
				fListener = BMessenger();
			fPingReceived = false;
			break;
		}
		default:
			BLooper::MessageReceived(message);
			break;
	}
}

// SetListener
void
RDesktopInputFilterLooper::SetListener(BMessenger listener)
{
	BAutolock _(this);
	fListener = listener;
}

// GetListener
BMessenger
RDesktopInputFilterLooper::GetListener()
{
	BAutolock _(this);
	return fListener;
}


// RDesktopInputFilter

// constructor
RDesktopInputFilter::RDesktopInputFilter()
	: fLooper(NULL)
{
init_debugging();
	fLooper = new(nothrow) RDesktopInputFilterLooper(this);
	if (fLooper)
		fLooper->Run();
}

// destructor
RDesktopInputFilter::~RDesktopInputFilter()
{
	if (fLooper) {
		fLooper->Lock();
		fLooper->Quit();
		fLooper = NULL;
	}
exit_debugging();
}

// InitCheck
status_t
RDesktopInputFilter::InitCheck()
{
	return (fLooper ? B_OK : B_NO_MEMORY);
}

// Filter
filter_result
RDesktopInputFilter::Filter(BMessage* message, BList* outList)
{
char buffer[10];
TOUCH(buffer);
PRINT(("RDesktopInputFilter::Filter(%s)\n", message_what(message, buffer)));
	filter_result result = B_DISPATCH_MESSAGE;
	BMessenger listener = fLooper->GetListener();
	if (listener.IsValid()) {
		switch (message->what) {
			case B_KEY_DOWN:
			case B_KEY_UP:
			case B_UNMAPPED_KEY_DOWN:
			case B_UNMAPPED_KEY_UP:
			case B_MODIFIERS_CHANGED:
			{
				BMessage filterMessage(MSG_INPUT_FILTER_MESSAGE);
				filterMessage.AddMessage("message", message);
				status_t error = listener.SendMessage(&filterMessage);
				if (error == B_OK)
					result = B_SKIP_MESSAGE;
				else
					fLooper->SetListener(BMessenger());
			}
		}
	}
	return result;
}
