// RDesktopServer.cpp
// 
// Copyright (c) 2004, Ingo Weinhold (bonefish@cs.tu-berlin.de)
// 
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation 
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
// THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
// DEALINGS IN THE SOFTWARE.
// 
// Except as contained in this notice, the name of a copyright holder shall
// not be used in advertising or otherwise to promote the sale, use or other
// dealings in this Software without prior written authorization of the
// copyright holder.

#include <new>

#include <stdlib.h>

#include <zlib.h>

#include <Application.h>
#include <Bitmap.h>
#include <Message.h>
#include <MessageRunner.h>
#include <Messenger.h>
#include <Screen.h>

#include <RemoteMessage.h>

#include "BlockingQueue.h"
#include "Debug.h"
#include "DebugSupport.h"
#include "RDesktopServerDefs.h"

// usage
const char* kUsage =
"Usage: RDesktopServer <options>\n"
"Options:\n"
"  -h, --help         - print this text\n"
"  --frames <frames>  - number of frames per second, the server shall try to\n"
"                       sent"
"\n";

// number of frames per second the server shall send (default, maximal)
static const int32 kDefaultFrameRate = 3;
static const int32 kMaxFrameRate = 25;

class RDesktopServer : public BApplication {
public:
								RDesktopServer(int32 frameRate);
	virtual						~RDesktopServer();

			status_t			Init();

	virtual	void				MessageReceived(BMessage* message);

private:
			void				_HandleConnectRequest(BMessage* request);
			void				_HandleUpdateTick();
			void				_HandleDesktopConfigChanged(BScreen& screen,
									BRect screenFrame, color_space colorSpace);
			void				_UnsetDesktopConfig();
			void				_SendFullSnapshot(BScreen& screen);

			status_t			_SendMessageToClient(BMessage* message,
									BMessage* originalMessage = NULL);

			void				_Disconnected();

			bool				_Compress(const void* data, int32 dataSize,
									int32* compressedSize);

	static	int32				_RequestProcessorEntry(void* data);
			int32				_RequestProcessor();

private:
			typedef BlockingQueue<BMessage> RequestQueue;

			int32				fFrameRate;
			RequestQueue		fPendingRequests;
			thread_id			fRequestProcessor;
			bool				fTerminating;
			BMessenger			fClient;
			int32				fClientToken;
			int32				fNextClientToken;
			BMessageRunner*		fUpdateTickRunner;
	volatile	bool			fUpdateTickPending;
			BRect				fScreenFrame;
			color_space			fScreenColorSpace;
			BBitmap*			fCurrentScreenShot;
			BBitmap*			fLastScreenShot;
			uint8*				fCompressionBuffer;
			int32				fCompressionBufferSize;
			uint8*				fDeltaBuffer;
			bool				fLastScreenShotValid;
};

// constructor
RDesktopServer::RDesktopServer(int32 frameRate)
	: BApplication(kRDesktopServerSignature),
	  fFrameRate(frameRate),
	  fPendingRequests(),
	  fRequestProcessor(-1),
	  fTerminating(false),
	  fClient(),
	  fClientToken(-1),
	  fNextClientToken(0),
	  fUpdateTickRunner(NULL),
	  fUpdateTickPending(false),
	  fScreenFrame(),
	  fScreenColorSpace(B_RGB32),
	  fCurrentScreenShot(NULL),
	  fLastScreenShot(NULL),
	  fCompressionBuffer(NULL),
	  fCompressionBufferSize(0),
	  fDeltaBuffer(NULL),
	  fLastScreenShotValid(false)
{
}

// destructor
RDesktopServer::~RDesktopServer()
{
	fTerminating = true;
	delete fUpdateTickRunner;
	// unblock the request processor, and wait for it to terminate
	fPendingRequests.Close(true);
 	if (fRequestProcessor >= 0) {
		int32 result;
		wait_for_thread(fRequestProcessor, &result);
	}
}

// Init
status_t
RDesktopServer::Init()
{
	status_t error = fPendingRequests.InitCheck();
	if (error != B_OK)
		return error;
	// create update tick runner
	BMessage updateTickMessage(MSG_UPDATE_TICK);
	fUpdateTickRunner = new(nothrow) BMessageRunner(
		BMessenger(this), &updateTickMessage, 1000000 / fFrameRate);
	if (!fUpdateTickRunner)
		return B_NO_MEMORY;
	error = fUpdateTickRunner->InitCheck();
	if (error != B_OK)
		return error;
	// spawn request processor thread
	fRequestProcessor = spawn_thread(_RequestProcessorEntry,
		"main request processor", B_NORMAL_PRIORITY, this);
	if (fRequestProcessor < 0)
		return fRequestProcessor;
	// resume the thread
	resume_thread(fRequestProcessor);
	return B_OK;
}

// MessageReceived
void
RDesktopServer::MessageReceived(BMessage* message)
{
	switch (message->what) {
		case MSG_UPDATE_TICK:
			if (fUpdateTickPending)
				break;
			DetachCurrentMessage();
			fUpdateTickPending = true;
			if (fPendingRequests.Push(message) != B_OK) {
				fUpdateTickPending = false;
				delete message;
			}
			break;
		case MSG_CONNECT:
			DetachCurrentMessage();
			if (fPendingRequests.Push(message) != B_OK)
				delete message;
			break;
		default:
			BApplication::MessageReceived(message);
			break;
	}
}

// _HandleConnectRequest
void
RDesktopServer::_HandleConnectRequest(BMessage* request)
{
PRINT(("RDesktopServer::_HandleConnectRequest()\n"));
	// currently only one connection is allowed
	status_t error = B_OK;
	if (fClientToken >= 0)
		SET_ERROR(error, B_ERROR);
	// get client messenger
	if (error == B_OK)
		error = request->FindMessenger("messenger", &fClient);
	// send reply
	BMessage reply(MSG_REPLY);
	if (error == B_OK) {
		fClientToken = fNextClientToken++;
		reply.AddMessenger("request messenger", BMessenger(this));
		reply.AddInt32("token", fClientToken);
	} else {
		reply.AddInt32("error", error);
	}
	request->SendReply(&reply);
PRINT(("RDesktopServer::_HandleConnectRequest() done\n"));
}

// _HandleUpdateTick
void
RDesktopServer::_HandleUpdateTick()
{
	// if no client is connected, there's nothing to do
	if (fClientToken < 0)
		return;
	// check whether the screen config has changed
	BScreen screen;
	color_space colorSpace = screen.ColorSpace();
	BRect screenFrame = screen.Frame();
	if (colorSpace != fScreenColorSpace || screenFrame != fScreenFrame) {
		// ...
		_HandleDesktopConfigChanged(screen, screenFrame, colorSpace);
		return;
	}
	// screen config is the same: take a snapshot and compare it with the last
	// one
// TODO: implement properly
	_SendFullSnapshot(screen);
}

// _HandleDesktopConfigChanged
void
RDesktopServer::_HandleDesktopConfigChanged(BScreen& screen,
	BRect screenFrame, color_space colorSpace)
{
	// unset old
	_UnsetDesktopConfig();
	// set new
	fCurrentScreenShot = new(nothrow) BBitmap(screenFrame, colorSpace);
//fCurrentScreenShot = new(nothrow) BBitmap(screenFrame, B_BITMAP_IS_CONTIGUOUS, colorSpace);
	if (!fCurrentScreenShot) {
		PRINT(("Allocating the current snapshot bitmap failed.\n"));
		return;
	}
	if (fCurrentScreenShot->InitCheck() != B_OK) {
		PRINT(("Initializing the current snapshot bitmap failed: %s\n",
			strerror(fCurrentScreenShot->InitCheck())));
		return;
	}
	fLastScreenShot = new(nothrow) BBitmap(screenFrame, colorSpace);
//fLastScreenShot = new(nothrow) BBitmap(screenFrame, B_BITMAP_IS_CONTIGUOUS, colorSpace);
	if (!fLastScreenShot) {
		PRINT(("Allocating the last snapshot bitmap failed.\n"));
		return;
	}
	if (fLastScreenShot->InitCheck() != B_OK) {
		PRINT(("Initializing the last snapshot bitmap failed: %s\n",
			strerror(fLastScreenShot->InitCheck())));
		return;
	}
	fCompressionBufferSize = fCurrentScreenShot->BitsLength();
	fCompressionBuffer = new(nothrow) uint8[fCompressionBufferSize];
	if (!fCompressionBuffer) {
		fCompressionBufferSize = 0;
		PRINT(("Allocating the compression buffer failed.\n"));
		return;
	}
	fDeltaBuffer = new(nothrow) uint8[fCompressionBufferSize];
	if (!fDeltaBuffer) {
		PRINT(("Allocating the delta buffer failed.\n"));
		return;
	}
	fScreenFrame = screenFrame;
	fScreenColorSpace = colorSpace;
	// notify the client
// TODO: Error checking!
	BMessage message(MSG_DESKTOP_CONFIG_UPDATE);
	message.AddRect("screen frame", fScreenFrame);
	message.AddInt32("color space", fScreenColorSpace);
	status_t error = _SendMessageToClient(&message);
	if (error != B_OK)
		return;
	// send a full screen
	_SendFullSnapshot(screen);
}

// _UnsetDesktopConfig
void
RDesktopServer::_UnsetDesktopConfig()
{
	delete fCurrentScreenShot;
	fCurrentScreenShot = NULL;
	delete fLastScreenShot;
	fLastScreenShot = NULL;
	delete[] fCompressionBuffer;
	fCompressionBuffer = NULL;
	fCompressionBufferSize = 0;
	delete[] fDeltaBuffer;
	fDeltaBuffer = NULL;
	fScreenFrame = BRect();
	fScreenColorSpace = B_RGB32;
	fLastScreenShotValid = false;
}

// _SendFullSnapshot
void
RDesktopServer::_SendFullSnapshot(BScreen& screen)
{
PRINT(("RDesktopServer::_SendFullSnapshot()\n"))
bigtime_t now = system_time();
bigtime_t lastTime = now;
	// take a snapshot
	BRect bounds(fScreenFrame);
	status_t error = screen.ReadBitmap(fCurrentScreenShot, true, &bounds);
now = system_time();
PRINT(("  BScreen::ReadBitmap() took: %Ld us\n", now - lastTime));
lastTime = now;
	if (error != B_OK) {
		PRINT(("BScreen::ReadBitmap() failed: %s\n", strerror(error)));
		PRINT(("  current screen shot bitmap: %p\n", fCurrentScreenShot));
		if (fCurrentScreenShot) {
			PRINT(("  InitCheck(): %lx\n", fCurrentScreenShot->InitCheck()));
			PRINT(("  color space: %lx\n",
				(int32)fCurrentScreenShot->ColorSpace()));
			BRect bounds(fCurrentScreenShot->Bounds());
			TOUCH(bounds);
			PRINT(("  size: (%f, %f)\n", bounds.Width(), bounds.Height()));
		}
		return;
	}
	if (bounds != fScreenFrame) {
		PRINT(("BScreen::ReadBitmap() reported different screen bounds\n"));
		return;
	}
	void* data = fCurrentScreenShot->Bits();
	int32 dataSize = fCurrentScreenShot->BitsLength();
	bool delta = false;
	int32 dataOffset = 0;
	// if last screen shot is valid, send a delta
	if (fLastScreenShotValid) {
		uint8* oldBitsStart = (uint8*)fLastScreenShot->Bits();
		uint8* oldBits = oldBitsStart;
		uint8* newBits = (uint8*)fCurrentScreenShot->Bits();
		uint8* deltaBits = fDeltaBuffer;
		const uint8* oldBitsEnd = oldBits + dataSize;
		// find first difference
		for (; oldBits != oldBitsEnd; oldBits++, newBits++) {
			if (*newBits != *oldBits)
				break;
		}
		dataOffset = oldBits - oldBitsStart;
		if (dataOffset < dataSize) {
			// create the delta
			const uint8* lastOldBitsDifference = NULL;
			for (; oldBits != oldBitsEnd; deltaBits++, oldBits++, newBits++) {
				if ((*deltaBits = *newBits - *oldBits))
					lastOldBitsDifference = oldBits;
			}
			dataSize = lastOldBitsDifference - oldBitsStart + 1 - dataOffset;
		} else {
			// no differences
			PRINT(("  no differences\n"));
			return;
		}
		data = fDeltaBuffer;
		delta = true;
now = system_time();
PRINT(("  computing delta took: %Ld us\n", now - lastTime));
lastTime = now;
	}
	// try to compress the data
	int32 compressedSize;
	bool compressed = _Compress(data, dataSize, &compressedSize);
	if (compressed) {
		if (compressedSize < dataSize) {
			data = fCompressionBuffer;
			dataSize = compressedSize;
		} else
			compressed = false;
	}
now = system_time();
PRINT(("  compressing took: %Ld us\n", now - lastTime));
lastTime = now;
	// prepare data message
// TODO: Error checking!
	BMessage parameter(MSG_DESKTOP_GRAPHICS_UPDATE);
	parameter.AddRect("update frame", fCurrentScreenShot->Bounds());
	parameter.AddInt32("color space", fCurrentScreenShot->ColorSpace());
	parameter.AddInt32("bytes per row", fCurrentScreenShot->BytesPerRow());
	parameter.AddInt32("data offset", dataOffset);
	parameter.AddBool("delta", delta);
	parameter.AddBool("compressed", compressed);
	BMessage envelope;
now = system_time();
PRINT(("  preparing parameter message took: %Ld us\n", now - lastTime));
lastTime = now;
	error = RemoteMessage::SetData(&envelope, data, dataSize, &parameter);
	if (error != B_OK) {
		PRINT(("RemoteMessage::SetData() failed: %s\n", strerror(error)));
		return;
	}
now = system_time();
PRINT(("  preparing data message took: %Ld us\n", now - lastTime));
lastTime = now;
	// send message
	error = _SendMessageToClient(&envelope);
	if (error != B_OK) {
		PRINT(("Sending the data message failed: %s\n", strerror(error)));
		return;
	}
now = system_time();
PRINT(("  sending message took: %Ld us\n", now - lastTime));
lastTime = now;
	BBitmap* last = fLastScreenShot;
	fLastScreenShot = fCurrentScreenShot;
	fCurrentScreenShot = last;
	fLastScreenShotValid = true;
}

// _SendMessageToClient
status_t
RDesktopServer::_SendMessageToClient(BMessage* message,
	BMessage* originalMessage)
{
	BMessage reply;
	status_t error;
	if (originalMessage)
		error = originalMessage->SendReply(message, &reply);
	else
		error = fClient.SendMessage(message, &reply);
	if (error != B_OK) {
		PRINT(("Sending the data message failed: %s\n", strerror(error)));
		if (error == B_BAD_TEAM_ID) {
			// the rmessage server is gone?
			_Disconnected();
		}
		return error;
	}
	if (reply.what == RMESSAGE_DELIVERY_FAILED) {
		// delivery failed
		status_t deliveryError;
		if (reply.FindInt32("error", &deliveryError) == B_OK)
			error = deliveryError;
		else
			error = B_ERROR;
		PRINT(("Delivering the data message failed: %s\n", strerror(error)));
		_Disconnected();
	} else if (reply.what == RMESSAGE_DELIVERY_TIMED_OUT) {
		PRINT(("Delivering the data message timed out\n"));
		_Disconnected();
	}
	return error;
}

// _Disconnected
void
RDesktopServer::_Disconnected()
{
	fClientToken = -1;
	_UnsetDesktopConfig();
}

// _Compress
bool
RDesktopServer::_Compress(const void* data, int32 dataSize,
	int32* compressedSize)
{
	if (!data || dataSize <= 0 || !fCompressionBuffer
		|| fCompressionBufferSize <= 0 || !compressedSize) {
		return false;
	}
	// prepare stream
	z_stream zStream = {
		(Bytef*)data,				// next_in
		dataSize,					// avail_in
		0,							// total_in
		(Bytef*)fCompressionBuffer,	// next_out
		fCompressionBufferSize,		// avail_out
		0,							// total_out
		0,							// msg
		0,							// state;
		Z_NULL,						// zalloc
		Z_NULL,						// zfree
		Z_NULL,						// opaque
		0,							// data_type
		0,							// adler
		0,							// reserved
	};
	int error = deflateInit(&zStream, Z_BEST_SPEED);
	if (error != Z_OK) {
PRINT(("RDesktopServer::_Compress(): init failed: %d\n", error));
		return false;
	}
	// deflate
	error = deflate(&zStream, Z_FINISH);
	bool result = true;
	if (error != Z_STREAM_END) {
PRINT(("RDesktopServer::_Compress(): deflate() failed: %d\n", error));
		result = false;
	}
	error = deflateEnd(&zStream);
	if (error != Z_OK && result) {
PRINT(("RDesktopServer::_Compress(): deflateEnd() failed: %d\n", error));
		result = false;
	}
	if (result) {
		*compressedSize = zStream.total_out;
PRINT(("RDesktopServer::_Compress(): successful: %ld -> %ld\n", dataSize,
*compressedSize));
	}
	return result;
}

// _RequestProcessorEntry
int32
RDesktopServer::_RequestProcessorEntry(void* data)
{
	return ((RDesktopServer*)data)->_RequestProcessor();
}

// _RequestProcessor
int32
RDesktopServer::_RequestProcessor()
{
	while (!fTerminating) {
		BMessage* request = NULL;
		status_t error = fPendingRequests.Pop(&request);
		if (error != B_OK || !request)
			continue;
		PRINT(("RDesktopServer::_RequestProcessor(): Got request\n"));
		// handle request
		switch (request->what) {
			case MSG_CONNECT:
				_HandleConnectRequest(request);
				break;
			case MSG_UPDATE_TICK:
				_HandleUpdateTick();
				fUpdateTickPending = false;
				break;
			default:
			{
				char dbgBuffer[5];
				TOUCH(dbgBuffer);
				FATAL(("Unknown request: %s\n",
					message_what(request, dbgBuffer)));
				break;
			}
		}
		delete request;
	}
	return B_OK;
}

// print_usage
static
void
print_usage(bool error)
{
	fprintf((error ? stderr : stdout), kUsage);
}

// main
int
main(int argc, char** argv)
{
	init_debugging();
	FUNCTION_START();
	// parse parameters
	const char* framesString = NULL;
	int32 frames = kDefaultFrameRate;
	int argi = 1;
	for (;argi < argc; argi++) {
		const char* arg = argv[argi];
		int argLen = strlen(arg);
		if (argLen > 0 && arg[0] == '-') {
			// option
			const char* option = arg + 1;
			// help
			if (strcmp(option, "h") == 0 || strcmp(option, "-help") == 0) {
				print_usage(false);
				return 0;
			// frames
			} else if (strcmp(option, "-frames") == 0) {
				argi++;
				if (argi >= argc) {
					fprintf(stderr, "--frames requires a parameter\n");
					print_usage(true);
					return 1;
				}
				framesString = argv[argi];
			// unknown
			} else {
				fprintf(stderr, "Illegal option: `-%s'\n", option);
				print_usage(true);
				return 1;
			}
		} else
			break;
	}
	if (argi != argc) {
		print_usage(true);
		return 1;
	}
	// if a frame rate was given, analyze the string
	if (framesString) {
		if (sscanf(framesString, "%ld", &frames) < 1 || frames < 1
			|| frames > kMaxFrameRate) {
			fprintf(stderr, "`%s' does not specify a valid frame rate. "
				" Should be a number between 1 and %ld.\n",
				framesString, kMaxFrameRate);
			return 1;
		}
	}
	// start application
	RDesktopServer app(frames);
	status_t error = app.Init();
	if (error == B_OK)
		app.Run();
	else
		fprintf(stderr, "Initialization failed: %s\n", strerror(error));
	FUNCTION_END();
	exit_debugging();
	return 0;
}
