// simpleTest3.cpp
// 
// Copyright (c) 2004, Ingo Weinhold (bonefish@cs.tu-berlin.de)
// 
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation 
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
// THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
// DEALINGS IN THE SOFTWARE.
// 
// Except as contained in this notice, the name of a copyright holder shall
// not be used in advertising or otherwise to promote the sale, use or other
// dealings in this Software without prior written authorization of the
// copyright holder.

#include <stdio.h>
#include <string.h>

#include <Application.h>
#include <NetAddress.h>

#include <RemoteMessage.h>
#include <RemoteRoster.h>

static const char* kAppSignature = "application/x-vnd.bonefish.rmessage-test3";

// message what values
enum {
	MSG_MESSAGE1	= 1,
	MSG_MESSAGE2	= 2,
	MSG_MESSAGE3	= 3,
};

enum {
	MESSAGE1_VALUE	= 435,
	MESSAGE2_VALUE	= 1498,
	MESSAGE3_VALUE	= 947,
};

enum {
	DATA_SIZE1 = 1024,
};

enum {
	DATA_OFFSET1 = 2,
};

static bool gIsMaster = false;

// prepare_data
void*
prepare_data(int32 size, int32 offset)
{
	uint8* data = new uint8[size];
	for (int i = 0; i < size; i++)
		data[i] = uint8(offset + i);
	return data;
}

// check_data
bool
check_data(void* _data, int32 size, int32 offset)
{
	uint8* data = (uint8*)_data;
	for (int i = 0; i < size; i++) {
		if (data[i] != uint8(offset + i))
			return false;
	}
	return true;
}

class MyHandler : public BHandler {
public:
	virtual void MessageReceived(BMessage* message)
	{
		switch (message->what) {
			case MSG_MESSAGE3:
			{
				printf("MyHandler: MSG_MESSAGE3 received\n");
				// examine the message
				int32 value;
				if (message->FindInt32("value", &value) == B_OK) {
					if (value != MESSAGE3_VALUE)
						printf("message contains wrong value\n");
				} else
					printf("message does not contain a value\n");
				// send a reply
				if (!gIsMaster) {
					BMessage reply(MSG_MESSAGE3);
					reply.AddInt32("value", MESSAGE3_VALUE);
					status_t error = message->SendReply(&reply);
					if (error != B_OK) {
						printf("Sending reply to message %lu failed: %s\n",
							message->what, strerror(error));
					}
				}
				be_app->PostMessage(B_QUIT_REQUESTED);
				break;
			}
			case MSG_MESSAGE1:
			case MSG_MESSAGE2:
				printf("MyLooper: received erroneous message: %lu\n",
					message->what);
				break;
		}
	}
};

class MyLooper : public BLooper {
public:
	MyLooper()
		: BLooper(),
		  fHandler()
	{
		AddHandler(&fHandler);
	}

	virtual void MessageReceived(BMessage* message)
	{
		switch (message->what) {
			case MSG_MESSAGE2:
			{
				printf("MyLooper: MSG_MESSAGE2 received\n");
				// examine the message
				int32 value;
				if (message->FindInt32("value", &value) == B_OK) {
					if (value != MESSAGE2_VALUE)
						printf("message contains wrong value\n");
				} else
					printf("message does not contain a value\n");
				// send a reply
				BMessage reply;
				if (gIsMaster) {
					reply.what = MSG_MESSAGE3;
					reply.AddInt32("value", MESSAGE3_VALUE);
				} else {
					reply.what = MSG_MESSAGE2;
					reply.AddInt32("value", MESSAGE2_VALUE);
				}
				status_t error = message->SendReply(&reply, &fHandler);
				if (error != B_OK) {
					printf("Sending reply to message %lu failed: %s\n",
						message->what, strerror(error));
				}
				break;
			}
			case MSG_MESSAGE1:
			case MSG_MESSAGE3:
				printf("MyLooper: received erroneous message: %lu\n",
					message->what);
				break;
		}
	}

private:
	MyHandler	fHandler;
};

class MyApplication : public BApplication {
public:
	MyApplication()
		: BApplication(kAppSignature),
		  fLooper(new MyLooper)
	{
		fLooper->Run();
	}

	virtual void MessageReceived(BMessage* message)
	{
		switch (message->what) {
			case RMESSAGE_DATA:
			{
				printf("MyApplication: REMOTE_DATA_MESSAGE received\n");
				const void* data;
				size_t size;
				BMessage parameters;
				bool hasParameters = false;
				status_t error = RemoteMessage::GetData(message,
					&data, &size, &parameters, &hasParameters);
				if (error != B_OK) {
					printf("Failed to get remote data: %s\n", strerror(error));
					break;
				}
				if (size == DATA_SIZE1) {
					if (check_data((void*)data, size, DATA_OFFSET1))
						printf("data arrived well\n");
					else
						printf("data mismatch\n");
				} else {
					printf("data size doesn't match: %lu vs. %d\n", size,
						DATA_SIZE1);
				}
				error = RemoteMessage::ReleaseData(data, size);
				if (error != B_OK) {
					printf("Failed to release remote data: %s\n",
						strerror(error));
				}
				break;
			}
			case MSG_MESSAGE1:
			{
				printf("MyApplication: MSG_MESSAGE1 received\n");
				// examine the message
				int32 value;
				if (message->FindInt32("value", &value) == B_OK) {
					if (value != MESSAGE1_VALUE)
						printf("message contains wrong value\n");
				} else
					printf("message does not contain a value\n");
				// send a reply
				BMessage reply;
				if (gIsMaster) {
					reply.what = MSG_MESSAGE2;
					reply.AddInt32("value", MESSAGE2_VALUE);
				} else {
					reply.what = MSG_MESSAGE1;
					reply.AddInt32("value", MESSAGE1_VALUE);
				}
				status_t error = message->SendReply(&reply, fLooper);
				if (error != B_OK) {
					printf("Sending reply to message %lu failed: %s\n",
						message->what, strerror(error));
				}
				break;
			}
			case MSG_MESSAGE2:
			case MSG_MESSAGE3:
				printf("MyApplication: received erroneous message: %lu\n",
					message->what);
				break;
		}
	}

private:
	MyLooper*	fLooper;
};


status_t
send_initial_message(const char* host)
{
	// get address
	BNetAddress address(host);
	status_t error = address.InitCheck();
	if (error != B_OK) {
		printf("Failed to get address: %s\n", strerror(error));
		return error;
	}
	// init remote roster
	printf("initializing roster...\n");
	RemoteRoster roster(&address);
	error = roster.InitCheck();
	if (error != B_OK) {
		printf("Failed to init remote roster: %s\n", strerror(error));
		return error;
	}
	// get the remote messenger
	printf("getting remote messenger...\n");
	BMessenger messenger;
	error = roster.GetMessenger(kAppSignature, &messenger);
	if (error != B_OK) {
		printf("Failed to get remote messenger: %s\n", strerror(error));
		return error;
	}
	// prepare data message
	printf("sending message...\n");
	BMessage message(MSG_MESSAGE1);
	message.AddInt32("value", MESSAGE1_VALUE);
	void* data = prepare_data(DATA_SIZE1, DATA_OFFSET1);
	BMessage envelope;
	error = RemoteMessage::SetData(&envelope, data, DATA_SIZE1, &message);
	if (error != B_OK) {
		printf("Failed to set up data message: %s\n", strerror(error));
		return error;
	}
	// send message
	error = messenger.SendMessage(&envelope, be_app_messenger);
	if (error != B_OK) {
		printf("Failed to send initial message: %s\n", strerror(error));
		return error;
	}
	return error;
}

// main
int
main(int argc, char** argv)
{
	MyApplication app;
	if (argc > 1) {
		gIsMaster = true;
		status_t error = send_initial_message(argv[1]);
		if (error != B_OK)
			return 1;
	}
	app.Run();
	return 0;
}
